/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "nthw_drv.h"

#include "flow_nthw_info.h"
#include "flow_nthw_ifr.h"
#include "flow_nthw_cat.h"
#include "flow_nthw_csu.h"
#include "flow_nthw_km.h"
#include "flow_nthw_flm.h"
#include "flow_nthw_hfu.h"
#include "flow_nthw_hsh.h"
#include "flow_nthw_qsl.h"
#include "flow_nthw_slc.h"
#include "flow_nthw_slc_lr.h"
#include "flow_nthw_pdb.h"
#include "flow_nthw_ioa.h"
#include "flow_nthw_rpp_lr.h"
#include "flow_nthw_roa.h"
#include "flow_nthw_rmc.h"
#include "flow_nthw_tx_cpy.h"
#include "flow_nthw_tx_ins.h"
#include "flow_nthw_tx_rpl.h"
#include "flow_backend.h"
#include "flow_api_backend.h"

#include <stdio.h> /* printf */

#if !defined(MAX_PHYS_ADAPTERS)
#define MAX_PHYS_ADAPTERS (8)
#endif

/*
 * Binary Flow API backend implementation into ntservice driver
 *
 * General note on this backend implementation:
 * Maybe use shadow class to combine multiple writes. However, this backend is only for dev/testing
 */

static struct  backend_dev_s {
  uint8_t adapter_no;
  enum debug_mode_e dmode;
  struct InfoNthw *pInfoNthw;
  struct CatNthw  *pCatNthw;
  struct KmNthw   *pKmNthw;
  struct FlmNthw  *pFlmNthw;
  struct HshNthw  *pHshNthw;
  struct QslNthw  *pQslNthw;
  struct SlcNthw  *pSlcNthw;
  struct SlcLrNthw  *pSlcLrNthw;
  struct PdbNthw  *pPdbNthw;
  struct IoaNthw  *pIoaNthw;
  struct RoaNthw  *pRoaNthw;
  struct RmcNthw  *pRmcNthw;
  struct HfuNthw   *pHfuNthw;   /* TPE module */
  struct RppLrNthw *pRppLrNthw; /* TPE module */
  struct TxCpyNthw *pTxCpyNthw; /* TPE module */
  struct TxInsNthw *pTxInsNthw; /* TPE module */
  struct TxRplNthw *pTxRplNthw; /* TPE module */
  struct CsuNthw   *pCsuNthw;   /* TPE module */
  struct IfrNthw   *pIfrNthw;   /* TPE module */
} be_devs[MAX_PHYS_ADAPTERS];

#if (1)
#define _CHECK_DEBUG_ON(be, mod, inst)                           \
  int __debug__ = 0;                                             \
  if ((be->dmode & FLOW_BACKEND_DEBUG_MODE_WRITE) || mod->debug) \
  do {  \
    mod ## Nthw_setDebugMode(inst, 0xFF);                      \
    __debug__ = 1;                                               \
  } while (0)

#define _CHECK_DEBUG_OFF(mod, inst) do {\
    if (__debug__) mod ## Nthw_setDebugMode(inst, 0);\
    } while(0)
#else
#define _CHECK_DEBUG_ON(be, mod, inst)
#define _CHECK_DEBUG_OFF(mod, inst)
#endif

static int set_debug_mode(void *be_dev, enum debug_mode_e mode)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  be->dmode = mode;
  return 0;
}

/*
 *  *****************  INFO  *******************
 */

static int get_nb_phy_ports(void *be_dev)
{
  struct backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbPhyPorts(be->pInfoNthw);
}

static int get_nb_rx_ports(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbRxPorts(be->pInfoNthw);
}

static int get_ltx_avail(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getLtxAvail(be->pInfoNthw);
}

static int get_nb_cat_funcs(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbCatFuncs(be->pInfoNthw);
}

static int get_nb_categories(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbCategories(be->pInfoNthw);
}

static int get_nb_cat_km_if_cnt(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbCatKmIfCnt(be->pInfoNthw);
}

static int get_nb_cat_km_if_m0(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbCatKmIfM0(be->pInfoNthw);
}

static int get_nb_cat_km_if_m1(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbCatKmIfM1(be->pInfoNthw);
}

static int get_nb_queues(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbQueues(be->pInfoNthw);
}

static int get_nb_km_flow_types(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbKMFlowTypes(be->pInfoNthw);
}

static int get_nb_pm_ext(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbPmExt(be->pInfoNthw);
}

static int get_nb_len(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbLen(be->pInfoNthw);
}

static int get_kcc_size(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getKccSize(be->pInfoNthw);
}

static int get_kcc_banks(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getKccBanks(be->pInfoNthw);
}

static int get_nb_km_categories(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbKmCategories(be->pInfoNthw);
}

static int get_nb_km_cam_banks(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbKmCamBanks(be->pInfoNthw);
}

static int get_nb_km_cam_record_words(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbKmCamRecordWords(be->pInfoNthw);
}

static int get_nb_km_cam_records(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbKmCamRecords(be->pInfoNthw);
}

static int get_nb_km_tcam_banks(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbKmTcamBanks(be->pInfoNthw);
}

static int get_nb_km_tcam_bank_width(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbKmTcamBankWidth(be->pInfoNthw);
}

static int get_nb_flm_categories(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbFlmCategories(be->pInfoNthw);
}

static int get_nb_flm_size_mb(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbFlmSizeMb(be->pInfoNthw);
}

static int get_nb_flm_entry_size(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbFlmEntrySize(be->pInfoNthw);
}

static int get_nb_flm_variant(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbFlmVariant(be->pInfoNthw);
}

static int get_nb_flm_prios(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbFlmPrios(be->pInfoNthw);
}

static int get_nb_flm_pst_profiles(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbFlmPstProfiles(be->pInfoNthw);
}

static int get_nb_qsl_categories(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbQslCategories(be->pInfoNthw);
}

static int get_nb_qsl_qst_entries(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbQslQstEntries(be->pInfoNthw);
}

static int get_nb_pdb_categories(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbPdbCategories(be->pInfoNthw);
}

static int get_nb_ioa_categories(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbIoaCategories(be->pInfoNthw);
}

static int get_nb_roa_categories(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbRoaCategories(be->pInfoNthw);
}

static int get_nb_tpe_categories(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbTpeCategories(be->pInfoNthw);
}

static int get_nb_tx_cpy_writers(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbTxCpyWriters(be->pInfoNthw);
}

static int get_nb_tx_cpy_mask_mem(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbTxCpyMaskMem(be->pInfoNthw);
}

static int get_nb_tx_rpl_depth(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbTxRplDepth(be->pInfoNthw);
}

static int get_nb_tx_rpl_ext_categories(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbTxRplExtCategories(be->pInfoNthw);
}

static int get_nb_tpe_ifr_categories(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return InfoNthw_getNbTpeIfrCategories(be->pInfoNthw);
}

/*
 *  *****************  CAT  *******************
 */

static bool cat_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pCatNthw != NULL;
}

static uint32_t cat_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pCatNthw->mCat) << 16) |
                    (nthw_module_get_minor_version(be->pCatNthw->mCat) & 0xffff));
}

static int cat_cfn_flush(void *be_dev, const struct cat_func_s *cat, int cat_func, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;

  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18) {
    CatNthw_CfnCnt(be->pCatNthw, 1U);
    for (int i = 0; i < cnt; i++) {
      CatNthw_CfnSelect(be->pCatNthw, cat_func);
      CatNthw_CfnEnable(be->pCatNthw, cat->v18.cfn[cat_func].enable);
      CatNthw_CfnInv         (be->pCatNthw, cat->v18.cfn[cat_func].inv);
      CatNthw_CfnPtcInv      (be->pCatNthw, cat->v18.cfn[cat_func].ptc_inv);
      CatNthw_CfnPtcIsl      (be->pCatNthw, cat->v18.cfn[cat_func].ptc_isl);
      CatNthw_CfnPtcCfp      (be->pCatNthw, cat->v18.cfn[cat_func].ptc_cfp);
      CatNthw_CfnPtcMac      (be->pCatNthw, cat->v18.cfn[cat_func].ptc_mac);
      CatNthw_CfnPtcL2       (be->pCatNthw, cat->v18.cfn[cat_func].ptc_l2);
      CatNthw_CfnPtcVnTag    (be->pCatNthw, cat->v18.cfn[cat_func].ptc_vntag);
      CatNthw_CfnPtcVlan     (be->pCatNthw, cat->v18.cfn[cat_func].ptc_vlan);
      CatNthw_CfnPtcMpls     (be->pCatNthw, cat->v18.cfn[cat_func].ptc_mpls);
      CatNthw_CfnPtcL3       (be->pCatNthw, cat->v18.cfn[cat_func].ptc_l3);
      CatNthw_CfnPtcFrag     (be->pCatNthw, cat->v18.cfn[cat_func].ptc_frag);
      CatNthw_CfnPtcIpProt   (be->pCatNthw, cat->v18.cfn[cat_func].ptc_ip_prot);
      CatNthw_CfnPtcL4       (be->pCatNthw, cat->v18.cfn[cat_func].ptc_l4);
      CatNthw_CfnPtcTunnel   (be->pCatNthw, cat->v18.cfn[cat_func].ptc_tunnel);
      CatNthw_CfnPtcTnlL2    (be->pCatNthw, cat->v18.cfn[cat_func].ptc_tnl_l2);
      CatNthw_CfnPtcTnlVlan  (be->pCatNthw, cat->v18.cfn[cat_func].ptc_tnl_vlan);
      CatNthw_CfnPtcTnlMpls  (be->pCatNthw, cat->v18.cfn[cat_func].ptc_tnl_mpls);
      CatNthw_CfnPtcTnlL3    (be->pCatNthw, cat->v18.cfn[cat_func].ptc_tnl_l3);
      CatNthw_CfnPtcTnlFrag  (be->pCatNthw, cat->v18.cfn[cat_func].ptc_tnl_frag);
      CatNthw_CfnPtcTnlIpProt(be->pCatNthw, cat->v18.cfn[cat_func].ptc_tnl_ip_prot);
      CatNthw_CfnPtcTnlL4    (be->pCatNthw, cat->v18.cfn[cat_func].ptc_tnl_l4);

      CatNthw_CfnErrInv      (be->pCatNthw, cat->v18.cfn[cat_func].err_inv);
      CatNthw_CfnErrCv       (be->pCatNthw, cat->v18.cfn[cat_func].err_cv);
      CatNthw_CfnErrFcs      (be->pCatNthw, cat->v18.cfn[cat_func].err_fcs);
      CatNthw_CfnErrTrunc    (be->pCatNthw, cat->v18.cfn[cat_func].err_trunc);
      CatNthw_CfnErrL3Cs     (be->pCatNthw, cat->v18.cfn[cat_func].err_l3_cs);
      CatNthw_CfnErrL4Cs     (be->pCatNthw, cat->v18.cfn[cat_func].err_l4_cs);

      CatNthw_CfnMacPort     (be->pCatNthw, cat->v18.cfn[cat_func].mac_port);

      CatNthw_CfnPmCmp       (be->pCatNthw, cat->v18.cfn[cat_func].pm_cmp);
      CatNthw_CfnPmDct       (be->pCatNthw, cat->v18.cfn[cat_func].pm_dct);
      CatNthw_CfnPmExtInv    (be->pCatNthw, cat->v18.cfn[cat_func].pm_ext_inv);
      CatNthw_CfnPmCmb       (be->pCatNthw, cat->v18.cfn[cat_func].pm_cmb);
      CatNthw_CfnPmAndInv    (be->pCatNthw, cat->v18.cfn[cat_func].pm_and_inv);
      CatNthw_CfnPmOrInv     (be->pCatNthw, cat->v18.cfn[cat_func].pm_or_inv);
      CatNthw_CfnPmInv       (be->pCatNthw, cat->v18.cfn[cat_func].pm_inv);

      CatNthw_CfnLc          (be->pCatNthw, cat->v18.cfn[cat_func].lc);
      CatNthw_CfnLcInv       (be->pCatNthw, cat->v18.cfn[cat_func].lc_inv);
      CatNthw_CfnKm0Or       (be->pCatNthw, cat->v18.cfn[cat_func].km_or);
      CatNthw_CfnFlush(be->pCatNthw);
      cat_func++;
    }
  }
  else if (cat->ver == 21 || cat->ver == 22) {
    CatNthw_CfnCnt(be->pCatNthw, 1U);
    for (int i = 0; i < cnt; i++) {
      CatNthw_CfnSelect(be->pCatNthw, cat_func);
      CatNthw_CfnEnable(be->pCatNthw, cat->v21.cfn[cat_func].enable);
      CatNthw_CfnInv         (be->pCatNthw, cat->v21.cfn[cat_func].inv);
      CatNthw_CfnPtcInv      (be->pCatNthw, cat->v21.cfn[cat_func].ptc_inv);
      CatNthw_CfnPtcIsl      (be->pCatNthw, cat->v21.cfn[cat_func].ptc_isl);
      CatNthw_CfnPtcCfp      (be->pCatNthw, cat->v21.cfn[cat_func].ptc_cfp);
      CatNthw_CfnPtcMac      (be->pCatNthw, cat->v21.cfn[cat_func].ptc_mac);
      CatNthw_CfnPtcL2       (be->pCatNthw, cat->v21.cfn[cat_func].ptc_l2);
      CatNthw_CfnPtcVnTag    (be->pCatNthw, cat->v21.cfn[cat_func].ptc_vntag);
      CatNthw_CfnPtcVlan     (be->pCatNthw, cat->v21.cfn[cat_func].ptc_vlan);
      CatNthw_CfnPtcMpls     (be->pCatNthw, cat->v21.cfn[cat_func].ptc_mpls);
      CatNthw_CfnPtcL3       (be->pCatNthw, cat->v21.cfn[cat_func].ptc_l3);
      CatNthw_CfnPtcFrag     (be->pCatNthw, cat->v21.cfn[cat_func].ptc_frag);
      CatNthw_CfnPtcIpProt   (be->pCatNthw, cat->v21.cfn[cat_func].ptc_ip_prot);
      CatNthw_CfnPtcL4       (be->pCatNthw, cat->v21.cfn[cat_func].ptc_l4);
      CatNthw_CfnPtcTunnel   (be->pCatNthw, cat->v21.cfn[cat_func].ptc_tunnel);
      CatNthw_CfnPtcTnlL2    (be->pCatNthw, cat->v21.cfn[cat_func].ptc_tnl_l2);
      CatNthw_CfnPtcTnlVlan  (be->pCatNthw, cat->v21.cfn[cat_func].ptc_tnl_vlan);
      CatNthw_CfnPtcTnlMpls  (be->pCatNthw, cat->v21.cfn[cat_func].ptc_tnl_mpls);
      CatNthw_CfnPtcTnlL3    (be->pCatNthw, cat->v21.cfn[cat_func].ptc_tnl_l3);
      CatNthw_CfnPtcTnlFrag  (be->pCatNthw, cat->v21.cfn[cat_func].ptc_tnl_frag);
      CatNthw_CfnPtcTnlIpProt(be->pCatNthw, cat->v21.cfn[cat_func].ptc_tnl_ip_prot);
      CatNthw_CfnPtcTnlL4    (be->pCatNthw, cat->v21.cfn[cat_func].ptc_tnl_l4);

      CatNthw_CfnErrInv      (be->pCatNthw, cat->v21.cfn[cat_func].err_inv);
      CatNthw_CfnErrCv       (be->pCatNthw, cat->v21.cfn[cat_func].err_cv);
      CatNthw_CfnErrFcs      (be->pCatNthw, cat->v21.cfn[cat_func].err_fcs);
      CatNthw_CfnErrTrunc    (be->pCatNthw, cat->v21.cfn[cat_func].err_trunc);
      CatNthw_CfnErrL3Cs     (be->pCatNthw, cat->v21.cfn[cat_func].err_l3_cs);
      CatNthw_CfnErrL4Cs     (be->pCatNthw, cat->v21.cfn[cat_func].err_l4_cs);
      CatNthw_CfnErrTnlL3Cs  (be->pCatNthw, cat->v21.cfn[cat_func].err_tnl_l3_cs);
      CatNthw_CfnErrTnlL4Cs  (be->pCatNthw, cat->v21.cfn[cat_func].err_tnl_l4_cs);
      CatNthw_CfnErrTtlExp   (be->pCatNthw, cat->v21.cfn[cat_func].err_ttl_exp);
      CatNthw_CfnErrTnlTtlExp(be->pCatNthw, cat->v21.cfn[cat_func].err_tnl_ttl_exp);

      CatNthw_CfnMacPort     (be->pCatNthw, cat->v21.cfn[cat_func].mac_port);

      CatNthw_CfnPmCmp       (be->pCatNthw, cat->v21.cfn[cat_func].pm_cmp);
      CatNthw_CfnPmDct       (be->pCatNthw, cat->v21.cfn[cat_func].pm_dct);
      CatNthw_CfnPmExtInv    (be->pCatNthw, cat->v21.cfn[cat_func].pm_ext_inv);
      CatNthw_CfnPmCmb       (be->pCatNthw, cat->v21.cfn[cat_func].pm_cmb);
      CatNthw_CfnPmAndInv    (be->pCatNthw, cat->v21.cfn[cat_func].pm_and_inv);
      CatNthw_CfnPmOrInv     (be->pCatNthw, cat->v21.cfn[cat_func].pm_or_inv);
      CatNthw_CfnPmInv       (be->pCatNthw, cat->v21.cfn[cat_func].pm_inv);

      CatNthw_CfnLc          (be->pCatNthw, cat->v21.cfn[cat_func].lc);
      CatNthw_CfnLcInv       (be->pCatNthw, cat->v21.cfn[cat_func].lc_inv);
      CatNthw_CfnKm0Or       (be->pCatNthw, cat->v21.cfn[cat_func].km0_or);
      if (be->pCatNthw->mKmIfCnt > 1) {
        CatNthw_CfnKm1Or     (be->pCatNthw, cat->v21.cfn[cat_func].km1_or);
      }
      CatNthw_CfnFlush(be->pCatNthw);
      cat_func++;
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}



static int cat_kce_flush(void *be_dev, const struct cat_func_s *cat, int km_if_idx, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18) {
    CatNthw_KceCnt(be->pCatNthw, 0, 1U);
    for (int i = 0; i < cnt; i++) {
      CatNthw_KceSelect(be->pCatNthw, 0, index + i);
      CatNthw_KceEnable(be->pCatNthw, 0, cat->v18.kce[index + i].enable_bm);
      CatNthw_KceFlush(be->pCatNthw, 0);
    }
  } else
  if (cat->ver == 21 || cat->ver == 22) {
    CatNthw_KceCnt(be->pCatNthw, km_if_idx, 1U);
    for (int i = 0; i < cnt; i++) {
      CatNthw_KceSelect(be->pCatNthw, km_if_idx, index + i);
      CatNthw_KceEnable(be->pCatNthw, km_if_idx, cat->v21.kce[index + i].enable_bm[km_if_idx]);
      CatNthw_KceFlush(be->pCatNthw, km_if_idx);
    }

  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}

static int cat_kcs_flush(void *be_dev, const struct cat_func_s *cat, int km_if_idx, int cat_func, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18) {
    CatNthw_KcsCnt(be->pCatNthw, 0, 1U);
    for (int i = 0; i < cnt; i++) {
      CatNthw_KcsSelect(be->pCatNthw, 0, cat_func);
      CatNthw_KcsCategory(be->pCatNthw, 0, cat->v18.kcs[cat_func].category);
      CatNthw_KcsFlush(be->pCatNthw, 0);
      cat_func++;
    }
  } else
  if (cat->ver == 21 || cat->ver == 22) {
    CatNthw_KcsCnt(be->pCatNthw, km_if_idx, 1U);
    for (int i = 0; i < cnt; i++) {
      CatNthw_KcsSelect(be->pCatNthw, km_if_idx, cat_func);
      CatNthw_KcsCategory(be->pCatNthw, km_if_idx, cat->v21.kcs[cat_func].category[km_if_idx]);
      CatNthw_KcsFlush(be->pCatNthw, km_if_idx);
      cat_func++;
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}

static int cat_fte_flush(void *be_dev, const struct cat_func_s *cat, int km_if_idx, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18) {
    CatNthw_FteCnt(be->pCatNthw, 0, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_FteSelect(be->pCatNthw, 0, index + i);
      CatNthw_FteEnable(be->pCatNthw, 0, cat->v18.fte[index + i].enable_bm);
      CatNthw_FteFlush(be->pCatNthw, 0);
    }
  } else
  if (cat->ver == 21 || cat->ver == 22) {
    CatNthw_FteCnt(be->pCatNthw, km_if_idx, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_FteSelect(be->pCatNthw, km_if_idx, index + i);
      CatNthw_FteEnable(be->pCatNthw, km_if_idx, cat->v21.fte[index + i].enable_bm[km_if_idx]);
      CatNthw_FteFlush(be->pCatNthw, km_if_idx);
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}



static int cat_cte_flush(void *be_dev, const struct cat_func_s *cat, int cat_func, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18 ||cat->ver == 21) {
    CatNthw_CteCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_CteSelect(be->pCatNthw, cat_func);
      CatNthw_CteEnableCol(be->pCatNthw, cat->v18.cte[cat_func].b.col);
      CatNthw_CteEnableCor(be->pCatNthw, cat->v18.cte[cat_func].b.cor);
      CatNthw_CteEnableHsh(be->pCatNthw, cat->v18.cte[cat_func].b.hsh);
      CatNthw_CteEnableQsl(be->pCatNthw, cat->v18.cte[cat_func].b.qsl);
      CatNthw_CteEnableIpf(be->pCatNthw, cat->v18.cte[cat_func].b.ipf);
      CatNthw_CteEnableSlc(be->pCatNthw, cat->v18.cte[cat_func].b.slc);
      CatNthw_CteEnablePdb(be->pCatNthw, cat->v18.cte[cat_func].b.pdb);
      CatNthw_CteEnableMsk(be->pCatNthw, cat->v18.cte[cat_func].b.msk);
      CatNthw_CteEnableHst(be->pCatNthw, cat->v18.cte[cat_func].b.hst);
      CatNthw_CteEnableEpp(be->pCatNthw, cat->v18.cte[cat_func].b.epp);
      CatNthw_CteEnableTpe(be->pCatNthw, cat->v18.cte[cat_func].b.tpe);

      CatNthw_CteFlush(be->pCatNthw);
      cat_func++;
    }
  } else
  if (cat->ver == 22) {
    CatNthw_CteCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_CteSelect(be->pCatNthw, cat_func);
      CatNthw_CteEnableCol(be->pCatNthw, cat->v22.cte[cat_func].b.col);
      CatNthw_CteEnableCor(be->pCatNthw, cat->v22.cte[cat_func].b.cor);
      CatNthw_CteEnableHsh(be->pCatNthw, cat->v22.cte[cat_func].b.hsh);
      CatNthw_CteEnableQsl(be->pCatNthw, cat->v22.cte[cat_func].b.qsl);
      CatNthw_CteEnableIpf(be->pCatNthw, cat->v22.cte[cat_func].b.ipf);
      CatNthw_CteEnableSlc(be->pCatNthw, cat->v22.cte[cat_func].b.slc);
      CatNthw_CteEnablePdb(be->pCatNthw, cat->v22.cte[cat_func].b.pdb);
      CatNthw_CteEnableMsk(be->pCatNthw, cat->v22.cte[cat_func].b.msk);
      CatNthw_CteEnableHst(be->pCatNthw, cat->v22.cte[cat_func].b.hst);
      CatNthw_CteEnableEpp(be->pCatNthw, cat->v22.cte[cat_func].b.epp);
      CatNthw_CteEnableTpe(be->pCatNthw, cat->v22.cte[cat_func].b.tpe);
      CatNthw_CteEnableTpe(be->pCatNthw, cat->v22.cte[cat_func].b.rrb);

      CatNthw_CteFlush(be->pCatNthw);
      cat_func++;
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}

static int cat_cts_flush(void *be_dev, const struct cat_func_s *cat, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18 || cat->ver == 21 || cat->ver == 22) {
    CatNthw_CtsCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_CtsSelect(be->pCatNthw, index + i);
      CatNthw_CtsCatA(be->pCatNthw, cat->v18.cts[index + i].cat_a);
      CatNthw_CtsCatB(be->pCatNthw, cat->v18.cts[index + i].cat_b);
      CatNthw_CtsFlush(be->pCatNthw);
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}

static int cat_cot_flush(void *be_dev, const struct cat_func_s *cat, int cat_func, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18 || cat->ver == 21 || cat->ver == 22) {
    CatNthw_CotCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_CotSelect(be->pCatNthw, cat_func + i);
      CatNthw_CotColor(be->pCatNthw, cat->v18.cot[cat_func + i].color);
      CatNthw_CotKm(be->pCatNthw, cat->v18.cot[cat_func + i].km);
      CatNthw_CotFlush(be->pCatNthw);
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}

static int cat_cct_flush(void *be_dev, const struct cat_func_s *cat, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18 || cat->ver == 21 || cat->ver == 22) {
    CatNthw_CctCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_CctSelect(be->pCatNthw, index + i);
      CatNthw_CctColor(be->pCatNthw, cat->v18.cct[index + i].color);
      CatNthw_CctKm(be->pCatNthw, cat->v18.cct[index + i].km);
      CatNthw_CctFlush(be->pCatNthw);
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}

static int cat_exo_flush(void *be_dev, const struct cat_func_s *cat, int ext_index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18 || cat->ver == 21 || cat->ver == 22) {
    CatNthw_ExoCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_ExoSelect(be->pCatNthw, ext_index + i);
      CatNthw_ExoDyn(be->pCatNthw, cat->v18.exo[ext_index + i].dyn);
      CatNthw_ExoOfs(be->pCatNthw, cat->v18.exo[ext_index + i].ofs);
      CatNthw_ExoFlush(be->pCatNthw);
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}

static int cat_rck_flush(void *be_dev, const struct cat_func_s *cat, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18 || cat->ver == 21 || cat->ver == 22) {
    CatNthw_RckCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_RckSelect(be->pCatNthw, index + i);
      CatNthw_RckData(be->pCatNthw, cat->v18.rck[index + i].rck_data);
      CatNthw_RckFlush(be->pCatNthw);
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}

static int cat_len_flush(void *be_dev, const struct cat_func_s *cat, int len_index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 18 || cat->ver == 21 || cat->ver == 22) {
    CatNthw_LenCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_LenSelect(be->pCatNthw, len_index + i);
      CatNthw_LenLower(be->pCatNthw, cat->v18.len[len_index + i].lower);
      CatNthw_LenUpper(be->pCatNthw, cat->v18.len[len_index + i].upper);
      CatNthw_LenDyn1(be->pCatNthw, cat->v18.len[len_index + i].dyn1);
      CatNthw_LenDyn2(be->pCatNthw, cat->v18.len[len_index + i].dyn2);
      CatNthw_LenInv(be->pCatNthw, cat->v18.len[len_index + i].inv);
      CatNthw_LenFlush(be->pCatNthw);
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}


static int cat_kcc_flush(void *be_dev, const struct cat_func_s *cat, int len_index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

if (cat->ver == 18 || cat->ver == 21 || cat->ver == 22) {
    CatNthw_KccCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_KccSelect(be->pCatNthw, len_index + i);
      CatNthw_KccKey(be->pCatNthw, cat->v18.kcc_cam[len_index + i].key);
      CatNthw_KccCategory(be->pCatNthw, cat->v18.kcc_cam[len_index + i].category);
      CatNthw_KccId(be->pCatNthw, cat->v18.kcc_cam[len_index + i].id);
      CatNthw_KccFlush(be->pCatNthw);
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}


static int cat_cce_flush(void *be_dev, const struct cat_func_s *cat, int len_index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 22) {
    CatNthw_CceCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_CceSelect(be->pCatNthw, len_index + i);
      CatNthw_CceDataImm(be->pCatNthw, cat->v22.cce[len_index + i].imm);
      CatNthw_CceDataInd(be->pCatNthw, cat->v22.cce[len_index + i].ind);
      CatNthw_CceFlush(be->pCatNthw);
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}


static int cat_ccs_flush(void *be_dev, const struct cat_func_s *cat, int len_index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, cat, be->pCatNthw);

  if (cat->ver == 22) {
    CatNthw_CcsCnt(be->pCatNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CatNthw_CcsSelect(be->pCatNthw, len_index + i);
      CatNthw_CcsDataCorEn(be->pCatNthw,   cat->v22.ccs[len_index + i].cor_en);
      CatNthw_CcsDataCor(be->pCatNthw,     cat->v22.ccs[len_index + i].cor);
      CatNthw_CcsDataHshEn(be->pCatNthw,   cat->v22.ccs[len_index + i].hsh_en);
      CatNthw_CcsDataHsh(be->pCatNthw,     cat->v22.ccs[len_index + i].hsh);
      CatNthw_CcsDataQslEn(be->pCatNthw,   cat->v22.ccs[len_index + i].qsl_en);
      CatNthw_CcsDataQsl(be->pCatNthw,     cat->v22.ccs[len_index + i].qsl);
      CatNthw_CcsDataIpfEn(be->pCatNthw,   cat->v22.ccs[len_index + i].ipf_en);
      CatNthw_CcsDataIpf(be->pCatNthw,     cat->v22.ccs[len_index + i].ipf);
      CatNthw_CcsDataSlcEn(be->pCatNthw,   cat->v22.ccs[len_index + i].slc_en);
      CatNthw_CcsDataSlc(be->pCatNthw,     cat->v22.ccs[len_index + i].slc);
      CatNthw_CcsDataPdbEn(be->pCatNthw,   cat->v22.ccs[len_index + i].pdb_en);
      CatNthw_CcsDataPdb(be->pCatNthw,     cat->v22.ccs[len_index + i].pdb);
      CatNthw_CcsDataMskEn(be->pCatNthw,   cat->v22.ccs[len_index + i].msk_en);
      CatNthw_CcsDataMsk(be->pCatNthw,     cat->v22.ccs[len_index + i].msk);
      CatNthw_CcsDataHstEn(be->pCatNthw,   cat->v22.ccs[len_index + i].hst_en);
      CatNthw_CcsDataHst(be->pCatNthw,     cat->v22.ccs[len_index + i].hst);
      CatNthw_CcsDataEppEn(be->pCatNthw,   cat->v22.ccs[len_index + i].epp_en);
      CatNthw_CcsDataEpp(be->pCatNthw,     cat->v22.ccs[len_index + i].epp);
      CatNthw_CcsDataTpeEn(be->pCatNthw,   cat->v22.ccs[len_index + i].tpe_en);
      CatNthw_CcsDataTpe(be->pCatNthw,     cat->v22.ccs[len_index + i].tpe);
      CatNthw_CcsDataRrbEn(be->pCatNthw,   cat->v22.ccs[len_index + i].rrb_en);
      CatNthw_CcsDataRrb(be->pCatNthw,     cat->v22.ccs[len_index + i].rrb);
      CatNthw_CcsDataSb0Type(be->pCatNthw, cat->v22.ccs[len_index + i].sb0_type);
      CatNthw_CcsDataSb0Data(be->pCatNthw, cat->v22.ccs[len_index + i].sb0_data);
      CatNthw_CcsDataSb1Type(be->pCatNthw, cat->v22.ccs[len_index + i].sb1_type);
      CatNthw_CcsDataSb1Data(be->pCatNthw, cat->v22.ccs[len_index + i].sb1_data);
      CatNthw_CcsDataSb2Type(be->pCatNthw, cat->v22.ccs[len_index + i].sb2_type);
      CatNthw_CcsDataSb2Data(be->pCatNthw, cat->v22.ccs[len_index + i].sb2_data);
      CatNthw_CcsFlush(be->pCatNthw);
    }
  }

  _CHECK_DEBUG_OFF(cat, be->pCatNthw);
  return 0;
}



/*
 *  *****************  KM  *******************
 */

static bool km_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pKmNthw != NULL;
}

static uint32_t km_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pKmNthw->mKm) << 16) |
                    (nthw_module_get_minor_version(be->pKmNthw->mKm) & 0xffff));
}

static int km_rcp_flush(void *be_dev, const struct km_func_s *km, int category, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;

  _CHECK_DEBUG_ON(be, km, be->pKmNthw);

  if (km->ver == 7) {
    KmNthw_RcpCnt(be->pKmNthw, 1);
    for (int i = 0; i < cnt; i++) {
      KmNthw_RcpSelect(be->pKmNthw, category + i);
      KmNthw_RcpQw0Dyn       (be->pKmNthw, km->v7.rcp[category + i].qw0_dyn);
      KmNthw_RcpQw0Ofs       (be->pKmNthw, km->v7.rcp[category + i].qw0_ofs);
      KmNthw_RcpQw0SelA      (be->pKmNthw, km->v7.rcp[category + i].qw0_sel_a);
      KmNthw_RcpQw0SelB      (be->pKmNthw, km->v7.rcp[category + i].qw0_sel_b);
      KmNthw_RcpQw4Dyn       (be->pKmNthw, km->v7.rcp[category + i].qw4_dyn);
      KmNthw_RcpQw4Ofs       (be->pKmNthw, km->v7.rcp[category + i].qw4_ofs);
      KmNthw_RcpQw4SelA      (be->pKmNthw, km->v7.rcp[category + i].qw4_sel_a);
      KmNthw_RcpQw4SelB      (be->pKmNthw, km->v7.rcp[category + i].qw4_sel_b);
      KmNthw_RcpDw8Dyn       (be->pKmNthw, km->v7.rcp[category + i].dw8_dyn);
      KmNthw_RcpDw8Ofs       (be->pKmNthw, km->v7.rcp[category + i].dw8_ofs);
      KmNthw_RcpDw8SelA      (be->pKmNthw, km->v7.rcp[category + i].dw8_sel_a);
      KmNthw_RcpDw8SelB      (be->pKmNthw, km->v7.rcp[category + i].dw8_sel_b);
      KmNthw_RcpDw10Dyn      (be->pKmNthw, km->v7.rcp[category + i].dw10_dyn);
      KmNthw_RcpDw10Ofs      (be->pKmNthw, km->v7.rcp[category + i].dw10_ofs);
      KmNthw_RcpDw10SelA     (be->pKmNthw, km->v7.rcp[category + i].dw10_sel_a);
      KmNthw_RcpDw10SelB     (be->pKmNthw, km->v7.rcp[category + i].dw10_sel_b);
      KmNthw_RcpSwxCch       (be->pKmNthw, km->v7.rcp[category + i].swx_cch);
      KmNthw_RcpSwxSelA      (be->pKmNthw, km->v7.rcp[category + i].swx_sel_a);
      KmNthw_RcpSwxSelB      (be->pKmNthw, km->v7.rcp[category + i].swx_sel_b);
      KmNthw_RcpMaskDA       (be->pKmNthw, km->v7.rcp[category + i].mask_d_a);
      KmNthw_RcpMaskB        (be->pKmNthw, km->v7.rcp[category + i].mask_b);
      KmNthw_RcpDual         (be->pKmNthw, km->v7.rcp[category + i].dual);
      KmNthw_RcpPaired       (be->pKmNthw, km->v7.rcp[category + i].paired);
      KmNthw_RcpElA          (be->pKmNthw, km->v7.rcp[category + i].el_a);
      KmNthw_RcpElB          (be->pKmNthw, km->v7.rcp[category + i].el_b);
      KmNthw_RcpInfoA        (be->pKmNthw, km->v7.rcp[category + i].info_a);
      KmNthw_RcpInfoB        (be->pKmNthw, km->v7.rcp[category + i].info_b);
      KmNthw_RcpFtmA         (be->pKmNthw, km->v7.rcp[category + i].ftm_a);
      KmNthw_RcpFtmB         (be->pKmNthw, km->v7.rcp[category + i].ftm_b);
      KmNthw_RcpBankA        (be->pKmNthw, km->v7.rcp[category + i].bank_a);
      KmNthw_RcpBankB        (be->pKmNthw, km->v7.rcp[category + i].bank_b);
      KmNthw_RcpKlA          (be->pKmNthw, km->v7.rcp[category + i].kl_a);
      KmNthw_RcpKlB          (be->pKmNthw, km->v7.rcp[category + i].kl_b);
      KmNthw_RcpKeywayA      (be->pKmNthw, km->v7.rcp[category + i].keyway_a);
      KmNthw_RcpKeywayB      (be->pKmNthw, km->v7.rcp[category + i].keyway_b);
      KmNthw_RcpSynergyMode  (be->pKmNthw, km->v7.rcp[category + i].synergy_mode);
      KmNthw_RcpDw0BDyn      (be->pKmNthw, km->v7.rcp[category + i].dw0_b_dyn);
      KmNthw_RcpDw0BOfs      (be->pKmNthw, km->v7.rcp[category + i].dw0_b_ofs);
      KmNthw_RcpDw2BDyn      (be->pKmNthw, km->v7.rcp[category + i].dw2_b_dyn);
      KmNthw_RcpDw2BOfs      (be->pKmNthw, km->v7.rcp[category + i].dw2_b_ofs);
      KmNthw_RcpSw4BDyn      (be->pKmNthw, km->v7.rcp[category + i].sw4_b_dyn);
      KmNthw_RcpSw4BOfs      (be->pKmNthw, km->v7.rcp[category + i].sw4_b_ofs);
      KmNthw_RcpSw5BDyn      (be->pKmNthw, km->v7.rcp[category + i].sw5_b_dyn);
      KmNthw_RcpSw5BOfs      (be->pKmNthw, km->v7.rcp[category + i].sw5_b_ofs);
      KmNthw_RcpFlush(be->pKmNthw);
    }
  }

  _CHECK_DEBUG_OFF(km, be->pKmNthw);
  return 0;
}

static int km_cam_flush(void *be_dev, const struct km_func_s *km, int bank,
                        int record, int cnt)
{
  struct backend_dev_s *be = (struct backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, km, be->pKmNthw);

  if (km->ver == 7) {
    KmNthw_CamCnt(be->pKmNthw, 1);
    for (int i = 0; i < cnt; i++) {
      KmNthw_CamSelect(be->pKmNthw, (bank << 11) + record + i);
      KmNthw_CamW0       (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].w0);
      KmNthw_CamW1       (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].w1);
      KmNthw_CamW2       (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].w2);
      KmNthw_CamW3       (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].w3);
      KmNthw_CamW4       (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].w4);
      KmNthw_CamW5       (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].w5);
      KmNthw_CamFt0      (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].ft0);
      KmNthw_CamFt1      (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].ft1);
      KmNthw_CamFt2      (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].ft2);
      KmNthw_CamFt3      (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].ft3);
      KmNthw_CamFt4      (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].ft4);
      KmNthw_CamFt5      (be->pKmNthw, km->v7.cam[(bank << 11) + record + i].ft5);
      KmNthw_CamFlush(be->pKmNthw);
    }
  }

  _CHECK_DEBUG_OFF(km, be->pKmNthw);
  return 0;
}

static int km_tcam_flush(void *be_dev, const struct km_func_s *km, int bank, int byte, int value, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, km, be->pKmNthw);

  if (km->ver == 7) {
    int start_idx = bank * 4 * 256 + byte * 256 + value;
#if 0
    /* TODO consider cont > 1 */
#endif
    KmNthw_TcamCnt(be->pKmNthw, 1);
    for (int i = 0; i < cnt; i++) {
      if (km->v7.tcam[start_idx + i].dirty) {
        KmNthw_TcamSelect(be->pKmNthw, start_idx + i);
        KmNthw_TcamT(be->pKmNthw, km->v7.tcam[start_idx + i].t);
        KmNthw_TcamFlush(be->pKmNthw);
        km->v7.tcam[start_idx + i].dirty = 0;
      }
    }
  }

  _CHECK_DEBUG_OFF(km, be->pKmNthw);
  return 0;
}

/*
 * bank is the TCAM bank, index is the index within the bank (0..71)
 */
static int km_tci_flush(void *be_dev, const struct km_func_s *km, int bank, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, km, be->pKmNthw);

  if (km->ver == 7) {
    /* TCAM bank width in version 3 = 72 */
    KmNthw_TciCnt(be->pKmNthw, 1);
    for (int i = 0; i < cnt; i++) {
      KmNthw_TciSelect(be->pKmNthw, bank * 72 + index + i);
      KmNthw_TciColor(be->pKmNthw, km->v7.tci[bank * 72 + index + i].color);
      KmNthw_TciFt(be->pKmNthw, km->v7.tci[bank * 72 + index + i].ft);
      KmNthw_TciFlush(be->pKmNthw);
    }
  }

  _CHECK_DEBUG_OFF(km, be->pKmNthw);
  return 0;
}

/*
 * bank is the TCAM bank, index is the index within the bank (0..71)
 */
static int km_tcq_flush(void *be_dev, const struct km_func_s *km, int bank, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, km, be->pKmNthw);

  if (km->ver == 7) {
    /* TCAM bank width in version 3 = 72 */
    KmNthw_TcqCnt(be->pKmNthw, 1);
    for (int i = 0; i < cnt; i++) {
      /* adr = lover 4 bits = bank, upper 7 bits = index */
      KmNthw_TcqSelect(be->pKmNthw, bank + (index << 4) + i);
      KmNthw_TcqBankMask(be->pKmNthw, km->v7.tcq[bank + (index << 4) + i].bank_mask);
      KmNthw_TcqQual(be->pKmNthw, km->v7.tcq[bank + (index << 4) + i].qual);
      KmNthw_TcqFlush(be->pKmNthw);
    }
  }

  _CHECK_DEBUG_OFF(km, be->pKmNthw);
  return 0;
}

/*
 *  *****************  FLM  *******************
 */

static bool flm_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pFlmNthw != NULL;
}

static uint32_t flm_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pFlmNthw->mFlm) << 16) |
                    (nthw_module_get_minor_version(be->pFlmNthw->mFlm) & 0xffff));
}

static int flm_control_flush(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_ControlEnable(be->pFlmNthw, flm->v17.control->enable);
    FlmNthw_ControlInit(be->pFlmNthw, flm->v17.control->init);
    FlmNthw_ControlLds(be->pFlmNthw, flm->v17.control->lds);
    FlmNthw_ControlLfs(be->pFlmNthw, flm->v17.control->lfs);
    FlmNthw_ControlLis(be->pFlmNthw, flm->v17.control->lis);
    FlmNthw_ControlUds(be->pFlmNthw, flm->v17.control->uds);
    FlmNthw_ControlUis(be->pFlmNthw, flm->v17.control->uis);
    FlmNthw_ControlRds(be->pFlmNthw, flm->v17.control->rds);
    FlmNthw_ControlRis(be->pFlmNthw, flm->v17.control->ris);
    FlmNthw_ControlPds(be->pFlmNthw, flm->v17.control->pds);
    FlmNthw_ControlPis(be->pFlmNthw, flm->v17.control->pis);
    FlmNthw_ControlCrcwr(be->pFlmNthw, flm->v17.control->crcwr);
    FlmNthw_ControlCrcrd(be->pFlmNthw, flm->v17.control->crcrd);
    FlmNthw_ControlRbl(be->pFlmNthw, flm->v17.control->rbl);
    FlmNthw_ControlEab(be->pFlmNthw, flm->v17.control->eab);
    FlmNthw_ControlSplitSdramUsage(be->pFlmNthw, flm->v17.control->split_sdram_usage);
    FlmNthw_ControlFlush(be->pFlmNthw);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_status_flush(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    /* CALIBDONE, INITDONE, IDLE, and EFT_BP is read only */
    FlmNthw_StatusCritical(be->pFlmNthw, &flm->v17.status->critical, 0);
    FlmNthw_StatusPanic(be->pFlmNthw, &flm->v17.status->panic, 0);
    FlmNthw_StatusCrcerr(be->pFlmNthw, &flm->v17.status->crcerr, 0);
    FlmNthw_StatusFlush(be->pFlmNthw);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_status_update(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_StatusUpdate(be->pFlmNthw);
    FlmNthw_StatusCalibdone(be->pFlmNthw, &flm->v17.status->calibdone, 1);
    FlmNthw_StatusInitdone(be->pFlmNthw, &flm->v17.status->initdone, 1);
    FlmNthw_StatusIdle(be->pFlmNthw, &flm->v17.status->idle, 1);
    FlmNthw_StatusCritical(be->pFlmNthw, &flm->v17.status->critical, 1);
    FlmNthw_StatusPanic(be->pFlmNthw, &flm->v17.status->panic, 1);
    FlmNthw_StatusCrcerr(be->pFlmNthw, &flm->v17.status->crcerr, 1);
    FlmNthw_StatusEftBp(be->pFlmNthw, &flm->v17.status->eft_bp, 1);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_timeout_flush(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_TimeoutT(be->pFlmNthw, flm->v17.timeout->t);
    FlmNthw_TimeoutFlush(be->pFlmNthw);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_scrub_flush(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_ScrubI(be->pFlmNthw, flm->v17.scrub->i);
    FlmNthw_ScrubFlush(be->pFlmNthw);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_load_bin_flush(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_LoadBin(be->pFlmNthw, flm->v17.load_bin->bin);
    FlmNthw_LoadBinFlush(be->pFlmNthw);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_load_pps_flush(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_LoadPps(be->pFlmNthw, flm->v17.load_pps->pps);
    FlmNthw_LoadPpsFlush(be->pFlmNthw);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_load_lps_flush(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_LoadLps(be->pFlmNthw, flm->v17.load_lps->lps);
    FlmNthw_LoadLpsFlush(be->pFlmNthw);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_load_aps_flush(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_LoadAps(be->pFlmNthw, flm->v17.load_aps->aps);
    FlmNthw_LoadApsFlush(be->pFlmNthw);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_prio_flush(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_PrioLimit0(be->pFlmNthw, flm->v17.prio->limit0);
    FlmNthw_PrioFt0(be->pFlmNthw, flm->v17.prio->ft0);
    FlmNthw_PrioLimit1(be->pFlmNthw, flm->v17.prio->limit1);
    FlmNthw_PrioFt1(be->pFlmNthw, flm->v17.prio->ft1);
    FlmNthw_PrioLimit2(be->pFlmNthw, flm->v17.prio->limit2);
    FlmNthw_PrioFt2(be->pFlmNthw, flm->v17.prio->ft2);
    FlmNthw_PrioLimit3(be->pFlmNthw, flm->v17.prio->limit3);
    FlmNthw_PrioFt3(be->pFlmNthw, flm->v17.prio->ft3);
    FlmNthw_PrioFlush(be->pFlmNthw);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_pst_flush(void *be_dev, const struct flm_func_s *flm, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_PstCnt(be->pFlmNthw, 1);
    for (int i = 0; i < cnt; i++) {
      FlmNthw_PstSelect(be->pFlmNthw, index + i);
      FlmNthw_PstBp(be->pFlmNthw, flm->v17.pst[index + i].bp);
      FlmNthw_PstPp(be->pFlmNthw, flm->v17.pst[index + i].pp);
      FlmNthw_PstTp(be->pFlmNthw, flm->v17.pst[index + i].tp);
      FlmNthw_PstFlush (be->pFlmNthw);
    }
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_rcp_flush(void *be_dev, const struct flm_func_s *flm, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_RcpCnt(be->pFlmNthw, 1);
    for (int i = 0; i < cnt; i++) {
      FlmNthw_RcpSelect(be->pFlmNthw, index + i);
      FlmNthw_RcpLookup(be->pFlmNthw, flm->v17.rcp[index + i].lookup);
      FlmNthw_RcpQw0Dyn(be->pFlmNthw, flm->v17.rcp[index + i].qw0_dyn);
      FlmNthw_RcpQw0Ofs(be->pFlmNthw, flm->v17.rcp[index + i].qw0_ofs);
      FlmNthw_RcpQw0Sel(be->pFlmNthw, flm->v17.rcp[index + i].qw0_sel);
      FlmNthw_RcpQw4Dyn(be->pFlmNthw, flm->v17.rcp[index + i].qw4_dyn);
      FlmNthw_RcpQw4Ofs(be->pFlmNthw, flm->v17.rcp[index + i].qw4_ofs);
      FlmNthw_RcpSw8Dyn(be->pFlmNthw, flm->v17.rcp[index + i].sw8_dyn);
      FlmNthw_RcpSw8Ofs(be->pFlmNthw, flm->v17.rcp[index + i].sw8_ofs);
      FlmNthw_RcpSw8Sel(be->pFlmNthw, flm->v17.rcp[index + i].sw8_sel);
      FlmNthw_RcpSw9Dyn(be->pFlmNthw, flm->v17.rcp[index + i].sw9_dyn);
      FlmNthw_RcpSw9Ofs(be->pFlmNthw, flm->v17.rcp[index + i].sw9_ofs);
      FlmNthw_RcpMask  (be->pFlmNthw, flm->v17.rcp[index + i].mask);
      FlmNthw_RcpKid   (be->pFlmNthw, flm->v17.rcp[index + i].kid);
      FlmNthw_RcpOpn   (be->pFlmNthw, flm->v17.rcp[index + i].opn);
      FlmNthw_RcpIpn   (be->pFlmNthw, flm->v17.rcp[index + i].ipn);
      FlmNthw_RcpBytDyn(be->pFlmNthw, flm->v17.rcp[index + i].byt_dyn);
      FlmNthw_RcpBytOfs(be->pFlmNthw, flm->v17.rcp[index + i].byt_ofs);
      FlmNthw_RcpTxplm (be->pFlmNthw, flm->v17.rcp[index + i].txplm);
      FlmNthw_RcpAutoIpv4Mask(be->pFlmNthw, flm->v17.rcp[index + i].auto_ipv4_mask);
      FlmNthw_RcpFlush (be->pFlmNthw);
    }
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_buf_ctrl_update(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_BufCtrlUpdate(be->pFlmNthw,
                          &flm->v17.buf_ctrl->lrn_free,
                          &flm->v17.buf_ctrl->inf_avail,
                          &flm->v17.buf_ctrl->sta_avail);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_stat_update(void *be_dev, const struct flm_func_s *flm)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  if (flm->ver >= 17) {
    FlmNthw_StatLrnDoneUpdate  (be->pFlmNthw);
    FlmNthw_StatLrnIgnoreUpdate(be->pFlmNthw);
    FlmNthw_StatLrnFailUpdate  (be->pFlmNthw);
    FlmNthw_StatUnlDoneUpdate  (be->pFlmNthw);
    FlmNthw_StatUnlIgnoreUpdate(be->pFlmNthw);
    FlmNthw_StatRelDoneUpdate  (be->pFlmNthw);
    FlmNthw_StatRelIgnoreUpdate(be->pFlmNthw);
    FlmNthw_StatAulDoneUpdate  (be->pFlmNthw);
    FlmNthw_StatAulIgnoreUpdate(be->pFlmNthw);
    FlmNthw_StatAulFailUpdate  (be->pFlmNthw);
    FlmNthw_StatTulDoneUpdate  (be->pFlmNthw);
    FlmNthw_StatFlowsUpdate    (be->pFlmNthw);

    FlmNthw_StatLrnDoneCnt  (be->pFlmNthw, &flm->v17.lrn_done->cnt, 1);
    FlmNthw_StatLrnIgnoreCnt(be->pFlmNthw, &flm->v17.lrn_ignore->cnt, 1);
    FlmNthw_StatLrnFailCnt  (be->pFlmNthw, &flm->v17.lrn_fail->cnt, 1);
    FlmNthw_StatUnlDoneCnt  (be->pFlmNthw, &flm->v17.unl_done->cnt, 1);
    FlmNthw_StatUnlIgnoreCnt(be->pFlmNthw, &flm->v17.unl_ignore->cnt, 1);
    FlmNthw_StatRelDoneCnt  (be->pFlmNthw, &flm->v17.rel_done->cnt, 1);
    FlmNthw_StatRelIgnoreCnt(be->pFlmNthw, &flm->v17.rel_ignore->cnt, 1);
    FlmNthw_StatAulDoneCnt  (be->pFlmNthw, &flm->v17.aul_done->cnt, 1);
    FlmNthw_StatAulIgnoreCnt(be->pFlmNthw, &flm->v17.aul_ignore->cnt, 1);
    FlmNthw_StatAulFailCnt  (be->pFlmNthw, &flm->v17.aul_fail->cnt, 1);
    FlmNthw_StatTulDoneCnt  (be->pFlmNthw, &flm->v17.tul_done->cnt, 1);
    FlmNthw_StatFlowsCnt    (be->pFlmNthw, &flm->v17.flows->cnt, 1);

    FlmNthw_StatPrbDoneUpdate  (be->pFlmNthw);
    FlmNthw_StatPrbIgnoreUpdate(be->pFlmNthw);
    FlmNthw_StatPrbDoneCnt  (be->pFlmNthw, &flm->v17.prb_done->cnt, 1);
    FlmNthw_StatPrbIgnoreCnt(be->pFlmNthw, &flm->v17.prb_ignore->cnt, 1);
  }
  if (flm->ver >= 20) {
    FlmNthw_StatStaDoneUpdate  (be->pFlmNthw);
    FlmNthw_StatInfDoneUpdate  (be->pFlmNthw);
    FlmNthw_StatInfSkipUpdate  (be->pFlmNthw);
    FlmNthw_StatPckHitUpdate   (be->pFlmNthw);
    FlmNthw_StatPckMissUpdate  (be->pFlmNthw);
    FlmNthw_StatPckUnhUpdate   (be->pFlmNthw);
    FlmNthw_StatPckDisUpdate   (be->pFlmNthw);
    FlmNthw_StatCshHitUpdate   (be->pFlmNthw);
    FlmNthw_StatCshMissUpdate  (be->pFlmNthw);
    FlmNthw_StatCshUnhUpdate   (be->pFlmNthw);
    FlmNthw_StatCucStartUpdate (be->pFlmNthw);
    FlmNthw_StatCucMoveUpdate  (be->pFlmNthw);

    FlmNthw_StatStaDoneCnt  (be->pFlmNthw, &flm->v20.sta_done->cnt, 1);
    FlmNthw_StatInfDoneCnt  (be->pFlmNthw, &flm->v20.inf_done->cnt, 1);
    FlmNthw_StatInfSkipCnt  (be->pFlmNthw, &flm->v20.inf_skip->cnt, 1);
    FlmNthw_StatPckHitCnt   (be->pFlmNthw, &flm->v20.pck_hit->cnt, 1);
    FlmNthw_StatPckMissCnt  (be->pFlmNthw, &flm->v20.pck_miss->cnt, 1);
    FlmNthw_StatPckUnhCnt   (be->pFlmNthw, &flm->v20.pck_unh->cnt, 1);
    FlmNthw_StatPckDisCnt   (be->pFlmNthw, &flm->v20.pck_dis->cnt, 1);
    FlmNthw_StatCshHitCnt   (be->pFlmNthw, &flm->v20.csh_hit->cnt, 1);
    FlmNthw_StatCshMissCnt  (be->pFlmNthw, &flm->v20.csh_miss->cnt, 1);
    FlmNthw_StatCshUnhCnt   (be->pFlmNthw, &flm->v20.csh_unh->cnt, 1);
    FlmNthw_StatCucStartCnt (be->pFlmNthw, &flm->v20.cuc_start->cnt, 1);
    FlmNthw_StatCucMoveCnt  (be->pFlmNthw, &flm->v20.cuc_move->cnt, 1);
  }

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return 0;
}

static int flm_lrn_data_flush(void *be_dev, const struct flm_func_s *flm, const uint32_t *lrn_data, uint32_t size)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  int ret = FlmNthw_LrnDataFlush(be->pFlmNthw, lrn_data, size,
                                 &flm->v17.buf_ctrl->lrn_free,
                                 &flm->v17.buf_ctrl->inf_avail,
                                 &flm->v17.buf_ctrl->sta_avail);

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return ret;
}

static int flm_inf_data_update(void *be_dev, const struct flm_func_s *flm, uint32_t *inf_data, uint32_t size)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  int ret = FlmNthw_InfDataUpdate(be->pFlmNthw, inf_data, size,
                                  &flm->v17.buf_ctrl->lrn_free,
                                  &flm->v17.buf_ctrl->inf_avail,
                                  &flm->v17.buf_ctrl->sta_avail);

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return ret;
}

static int flm_sta_data_update(void *be_dev, const struct flm_func_s *flm, uint32_t *sta_data, uint32_t size)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, flm, be->pFlmNthw);

  int ret = FlmNthw_StaDataUpdate(be->pFlmNthw, sta_data, size,
                                  &flm->v17.buf_ctrl->lrn_free,
                                  &flm->v17.buf_ctrl->inf_avail,
                                  &flm->v17.buf_ctrl->sta_avail);

  _CHECK_DEBUG_OFF(flm, be->pFlmNthw);
  return ret;
}


/*
 *  *****************  HSH  *******************
 */

static bool hsh_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pHshNthw != NULL;
}

static uint32_t hsh_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pHshNthw->mHsh) << 16) |
                    (nthw_module_get_minor_version(be->pHshNthw->mHsh) & 0xffff));
}


static int hsh_rcp_flush(void *be_dev, const struct hsh_func_s *hsh, int category, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, hsh, be->pHshNthw);

  if (hsh->ver == 5) {
    HshNthw_RcpCnt(be->pHshNthw, 1);
    for (int i = 0; i < cnt; i++) {
      HshNthw_RcpSelect(be->pHshNthw, category + i);
      HshNthw_RcpLoadDistType(be->pHshNthw, hsh->v5.rcp[category + i].load_dist_type);
      HshNthw_RcpMacPortMask (be->pHshNthw, hsh->v5.rcp[category + i].mac_port_mask);
      HshNthw_RcpSort        (be->pHshNthw, hsh->v5.rcp[category + i].sort);
      HshNthw_RcpQw0Pe       (be->pHshNthw, hsh->v5.rcp[category + i].qw0_pe);
      HshNthw_RcpQw0Ofs      (be->pHshNthw, hsh->v5.rcp[category + i].qw0_ofs);
      HshNthw_RcpQw4Pe       (be->pHshNthw, hsh->v5.rcp[category + i].qw4_pe);
      HshNthw_RcpQw4Ofs      (be->pHshNthw, hsh->v5.rcp[category + i].qw4_ofs);
      HshNthw_RcpW8Pe        (be->pHshNthw, hsh->v5.rcp[category + i].w8_pe);
      HshNthw_RcpW8Ofs       (be->pHshNthw, hsh->v5.rcp[category + i].w8_ofs);
      HshNthw_RcpW8Sort      (be->pHshNthw, hsh->v5.rcp[category + i].w8_sort);
      HshNthw_RcpW9Pe        (be->pHshNthw, hsh->v5.rcp[category + i].w9_pe);
      HshNthw_RcpW9Ofs       (be->pHshNthw, hsh->v5.rcp[category + i].w9_ofs);
      HshNthw_RcpW9Sort      (be->pHshNthw, hsh->v5.rcp[category + i].w9_sort);
      HshNthw_RcpW9P         (be->pHshNthw, hsh->v5.rcp[category + i].w9_p);
      HshNthw_RcpPMask       (be->pHshNthw, hsh->v5.rcp[category + i].p_mask);
      HshNthw_RcpWordMask    (be->pHshNthw, hsh->v5.rcp[category + i].word_mask);
      HshNthw_RcpSeed        (be->pHshNthw, hsh->v5.rcp[category + i].seed);
      HshNthw_RcpTnlP        (be->pHshNthw, hsh->v5.rcp[category + i].tnl_p);
      HshNthw_RcpHshValid    (be->pHshNthw, hsh->v5.rcp[category + i].hsh_valid);
      HshNthw_RcpHshType     (be->pHshNthw, hsh->v5.rcp[category + i].hsh_type);
      HshNthw_RcpAutoIpv4Mask(be->pHshNthw, hsh->v5.rcp[category + i].auto_ipv4_mask);
      HshNthw_RcpFlush(be->pHshNthw);
    }
  }
  _CHECK_DEBUG_OFF(hsh, be->pHshNthw);
  return 0;
}

/*
 *  *****************  QSL  *******************
 */

static bool qsl_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pQslNthw != NULL;
}

static uint32_t qsl_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pQslNthw->mQsl) << 16) |
                    (nthw_module_get_minor_version(be->pQslNthw->mQsl) & 0xffff));
}

static int qsl_rcp_flush(void *be_dev, const struct qsl_func_s *qsl, int category, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, qsl, be->pQslNthw);

  if (qsl->ver == 7) {
    QslNthw_RcpCnt(be->pQslNthw, 1);
    for (int i = 0; i < cnt; i++) {
      QslNthw_RcpSelect(be->pQslNthw, category + i);
      QslNthw_RcpDiscard(be->pQslNthw, qsl->v7.rcp[category + i].discard);
      QslNthw_RcpDrop   (be->pQslNthw, qsl->v7.rcp[category + i].drop);
      QslNthw_RcpTblLo  (be->pQslNthw, qsl->v7.rcp[category + i].tbl_lo);
      QslNthw_RcpTblHi  (be->pQslNthw, qsl->v7.rcp[category + i].tbl_hi);
      QslNthw_RcpTblIdx (be->pQslNthw, qsl->v7.rcp[category + i].tbl_idx);
      QslNthw_RcpTblMsk (be->pQslNthw, qsl->v7.rcp[category + i].tbl_msk);
      QslNthw_RcpLr     (be->pQslNthw, qsl->v7.rcp[category + i].lr);
      QslNthw_RcpTsa    (be->pQslNthw, qsl->v7.rcp[category + i].tsa);
      QslNthw_RcpVli    (be->pQslNthw, qsl->v7.rcp[category + i].vli);
      QslNthw_RcpFlush(be->pQslNthw);
    }
  }

  _CHECK_DEBUG_OFF(qsl, be->pQslNthw);
  return 0;
}


static int qsl_qst_flush(void *be_dev, const struct qsl_func_s *qsl, int entry, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, qsl, be->pQslNthw);

  if (qsl->ver == 7) {
    QslNthw_QstCnt(be->pQslNthw, 1);
    for (int i = 0; i < cnt; i++) {
      QslNthw_QstSelect(be->pQslNthw, entry + i);
      QslNthw_QstQueue(be->pQslNthw, qsl->v7.qst[entry + i].queue);
      QslNthw_QstEn(be->pQslNthw, qsl->v7.qst[entry + i].en);

      QslNthw_QstTxPort(be->pQslNthw, qsl->v7.qst[entry + i].tx_port);
      QslNthw_QstLre(be->pQslNthw, qsl->v7.qst[entry + i].lre);
      QslNthw_QstTci(be->pQslNthw, qsl->v7.qst[entry + i].tci);
      QslNthw_QstVen(be->pQslNthw, qsl->v7.qst[entry + i].ven);
      QslNthw_QstFlush(be->pQslNthw);
    }
  }

  _CHECK_DEBUG_OFF(qsl, be->pQslNthw);
  return 0;
}

static int qsl_qen_flush(void *be_dev, const struct qsl_func_s *qsl, int entry, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, qsl, be->pQslNthw);

  if (qsl->ver == 7) {
    QslNthw_QenCnt(be->pQslNthw, 1);
    for (int i = 0; i < cnt; i++) {
      QslNthw_QenSelect(be->pQslNthw, entry + i);
      QslNthw_QenEn(be->pQslNthw, qsl->v7.qen[entry + i].en);
      QslNthw_QenFlush(be->pQslNthw);
    }
  }

  _CHECK_DEBUG_OFF(qsl, be->pQslNthw);
  return 0;
}

static int qsl_unmq_flush(void *be_dev, const struct qsl_func_s *qsl, int entry, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, qsl, be->pQslNthw);

  if (qsl->ver == 7) {
    QslNthw_UnmqCnt(be->pQslNthw, 1);
    for (int i = 0; i < cnt; i++) {
      QslNthw_UnmqSelect(be->pQslNthw, entry + i);
      QslNthw_UnmqDestQueue(be->pQslNthw, qsl->v7.unmq[entry + i].dest_queue);
      QslNthw_UnmqEn(be->pQslNthw, qsl->v7.unmq[entry + i].en);
      QslNthw_UnmqFlush(be->pQslNthw);
    }
  }

  _CHECK_DEBUG_OFF(qsl, be->pQslNthw);
  return 0;
}


/*
 *  *****************  SLC  *******************
 */

static bool slc_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pSlcNthw != NULL;
}

static uint32_t slc_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pSlcNthw->mSlc) << 16) |
                    (nthw_module_get_minor_version(be->pSlcNthw->mSlc) & 0xffff));
}

static int slc_rcp_flush(void *be_dev, const struct slc_func_s *slc, int category, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, slc, be->pSlcNthw);

  if (slc->ver == 1) {
    SlcNthw_RcpCnt(be->pSlcNthw, 1);
    for (int i = 0; i < cnt; i++) {
      SlcNthw_RcpSelect(be->pSlcNthw, category + i);
      SlcNthw_RcpTailSlcEn (be->pSlcNthw, slc->v1.rcp[category + i].tail_slc_en);
      SlcNthw_RcpTailDyn   (be->pSlcNthw, slc->v1.rcp[category + i].tail_dyn);
      SlcNthw_RcpTailOfs   (be->pSlcNthw, slc->v1.rcp[category + i].tail_ofs);
      SlcNthw_RcpPcap      (be->pSlcNthw, slc->v1.rcp[category + i].pcap);
      SlcNthw_RcpFlush(be->pSlcNthw);
    }
  }

  _CHECK_DEBUG_OFF(slc, be->pSlcNthw);
  return 0;
}


/*
 *  *****************  SLC LR *******************
 */

static bool slc_lr_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pSlcLrNthw != NULL;
}

static uint32_t slc_lr_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pSlcLrNthw->mSlcLr) << 16) |
                    (nthw_module_get_minor_version(be->pSlcLrNthw->mSlcLr) & 0xffff));
}

static int slc_lr_rcp_flush(void *be_dev, const struct slc_lr_func_s *slcLr, int category, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, slcLr, be->pSlcLrNthw);

  if (slcLr->ver == 2) {
    SlcLrNthw_RcpCnt(be->pSlcLrNthw, 1);
    for (int i = 0; i < cnt; i++) {
      SlcLrNthw_RcpSelect(be->pSlcLrNthw, category + i);
      SlcLrNthw_RcpHeadSlcEn (be->pSlcLrNthw, slcLr->v2.rcp[category + i].head_slc_en);
      SlcLrNthw_RcpHeadDyn   (be->pSlcLrNthw, slcLr->v2.rcp[category + i].head_dyn);
      SlcLrNthw_RcpHeadOfs   (be->pSlcLrNthw, slcLr->v2.rcp[category + i].head_ofs);
      SlcLrNthw_RcpTailSlcEn (be->pSlcLrNthw, slcLr->v2.rcp[category + i].tail_slc_en);
      SlcLrNthw_RcpTailDyn   (be->pSlcLrNthw, slcLr->v2.rcp[category + i].tail_dyn);
      SlcLrNthw_RcpTailOfs   (be->pSlcLrNthw, slcLr->v2.rcp[category + i].tail_ofs);
      SlcLrNthw_RcpPcap      (be->pSlcLrNthw, slcLr->v2.rcp[category + i].pcap);
      SlcLrNthw_RcpFlush(be->pSlcLrNthw);
    }
  }

  _CHECK_DEBUG_OFF(slcLr, be->pSlcLrNthw);
  return 0;
}


/*
 *  *****************  PDB  *******************
 */

static bool pdb_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pPdbNthw != NULL;
}

static uint32_t pdb_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pPdbNthw->mPdb) << 16) |
                    (nthw_module_get_minor_version(be->pPdbNthw->mPdb) & 0xffff));
}


static int pdb_rcp_flush(void *be_dev, const struct pdb_func_s *pdb, int category, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, pdb, be->pPdbNthw);

  if (pdb->ver == 9) {
    PdbNthw_RcpCnt(be->pPdbNthw, 1);
    for (int i = 0; i < cnt; i++) {
      PdbNthw_RcpSelect(be->pPdbNthw, category + i);
      PdbNthw_RcpDescriptor (be->pPdbNthw, pdb->v9.rcp[category + i].descriptor);
      PdbNthw_RcpDescLen (be->pPdbNthw, pdb->v9.rcp[category + i].desc_len);
      PdbNthw_RcpTxPort (be->pPdbNthw, pdb->v9.rcp[category + i].tx_port);
      PdbNthw_RcpTxIgnore (be->pPdbNthw, pdb->v9.rcp[category + i].tx_ignore);
      PdbNthw_RcpTxNow (be->pPdbNthw, pdb->v9.rcp[category + i].tx_now);
      PdbNthw_RcpCrcOverwrite (be->pPdbNthw, pdb->v9.rcp[category + i].crc_overwrite);
      PdbNthw_RcpAlign (be->pPdbNthw, pdb->v9.rcp[category + i].align);
      PdbNthw_RcpOfs0Dyn (be->pPdbNthw, pdb->v9.rcp[category + i].ofs0_dyn);
      PdbNthw_RcpOfs0Rel (be->pPdbNthw, pdb->v9.rcp[category + i].ofs0_rel);
      PdbNthw_RcpOfs1Dyn (be->pPdbNthw, pdb->v9.rcp[category + i].ofs1_dyn);
      PdbNthw_RcpOfs1Rel (be->pPdbNthw, pdb->v9.rcp[category + i].ofs1_rel);
      PdbNthw_RcpOfs2Dyn (be->pPdbNthw, pdb->v9.rcp[category + i].ofs2_dyn);
      PdbNthw_RcpOfs2Rel (be->pPdbNthw, pdb->v9.rcp[category + i].ofs2_rel);
      PdbNthw_RcpIpProtTnl (be->pPdbNthw, pdb->v9.rcp[category + i].ip_prot_tnl);
      PdbNthw_RcpPpcHsh (be->pPdbNthw, pdb->v9.rcp[category + i].ppc_hsh);
      PdbNthw_RcpDuplicateEn (be->pPdbNthw, pdb->v9.rcp[category + i].duplicate_en);
      PdbNthw_RcpDuplicateBit (be->pPdbNthw, pdb->v9.rcp[category + i].duplicate_bit);
      PdbNthw_RcpDuplicateBit (be->pPdbNthw, pdb->v9.rcp[category + i].pcap_keep_fcs);
      PdbNthw_RcpFlush(be->pPdbNthw);
    }
  }
  _CHECK_DEBUG_OFF(pdb, be->pPdbNthw);
  return 0;
}


static int pdb_config_flush(void *be_dev, const struct pdb_func_s *pdb)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, pdb, be->pPdbNthw);

  if (pdb->ver == 9) {
    PdbNthw_ConfigTsFormat(be->pPdbNthw, pdb->v9.config->ts_format);
    PdbNthw_ConfigPortOfs(be->pPdbNthw, pdb->v9.config->port_ofs);
    PdbNthw_ConfigFlush(be->pPdbNthw);
  }

  _CHECK_DEBUG_OFF(pdb, be->pPdbNthw);
  return 0;
}

/*
 *  *****************  IOA  *******************
 */

static bool ioa_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pIoaNthw != NULL;
}

static uint32_t ioa_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pIoaNthw->mIoa) << 16) |
                    (nthw_module_get_minor_version(be->pIoaNthw->mIoa) & 0xffff));
}

static int ioa_rcp_flush(void *be_dev, const struct ioa_func_s *ioa,
                         int category, int cnt)
{
  struct backend_dev_s *be = (struct backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, ioa, be->pIoaNthw);

  if (ioa->ver == 4) {
    IoaNthw_RcpCnt(be->pIoaNthw, 1);
    for (int i = 0; i < cnt; i++) {
      IoaNthw_RcpSelect(be->pIoaNthw, category + i);
      IoaNthw_RcpTunnelPop(be->pIoaNthw, ioa->v4.rcp[category + i].tunnel_pop);
      IoaNthw_RcpVlanPop(be->pIoaNthw, ioa->v4.rcp[category + i].vlan_pop);
      IoaNthw_RcpVlanPush(be->pIoaNthw, ioa->v4.rcp[category + i].vlan_push);
      IoaNthw_RcpVlanVid(be->pIoaNthw, ioa->v4.rcp[category + i].vlan_vid);
      IoaNthw_RcpVlanDei(be->pIoaNthw, ioa->v4.rcp[category + i].vlan_dei);
      IoaNthw_RcpVlanPcp(be->pIoaNthw, ioa->v4.rcp[category + i].vlan_pcp);
      IoaNthw_RcpVlanTpidSel(be->pIoaNthw, ioa->v4.rcp[category + i].vlan_tpid_sel);
      IoaNthw_RcpQueueOverrideEn(be->pIoaNthw, ioa->v4.rcp[category + i].queue_override_en);
      IoaNthw_RcpQueueId(be->pIoaNthw, ioa->v4.rcp[category + i].queue_id);
      IoaNthw_RcpFlush(be->pIoaNthw);
    }
  }

  _CHECK_DEBUG_OFF(ioa, be->pIoaNthw);
  return 0;
}

static int ioa_special_tpid_flush(void *be_dev, const struct ioa_func_s *ioa)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, ioa, be->pIoaNthw);

  if (ioa->ver == 4) {
    IoaNthw_SpecialVlanTpidCustTpid0(be->pIoaNthw, ioa->v4.tpid->cust_tpid_0);
    IoaNthw_SpecialVlanTpidCustTpid1(be->pIoaNthw, ioa->v4.tpid->cust_tpid_1);
    IoaNthw_SpecialVlanTpidFlush(be->pIoaNthw);
  }

  _CHECK_DEBUG_OFF(ioa, be->pIoaNthw);
  return 0;
}

static int ioa_roa_epp_flush(void *be_dev, const struct ioa_func_s *ioa, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, ioa, be->pIoaNthw);

  if (ioa->ver == 4) {
    IoaNthw_RoaEppCnt(be->pIoaNthw, 1);
    for (int i = 0; i < cnt; i++) {
      IoaNthw_RoaEppSelect(be->pIoaNthw, index + i);
      IoaNthw_RoaEppPushTunnel(be->pIoaNthw, ioa->v4.roa_epp[index + i].push_tunnel);
      IoaNthw_RoaEppTxPort(be->pIoaNthw, ioa->v4.roa_epp[index + i].tx_port);
      IoaNthw_RoaEppFlush(be->pIoaNthw);
    }
  }

  _CHECK_DEBUG_OFF(ioa, be->pIoaNthw);
  return 0;
}


/*
 *  *****************  ROA  *******************
 */

static bool roa_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pRoaNthw != NULL;
}

static uint32_t roa_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pRoaNthw->mRoa) << 16) |
                    (nthw_module_get_minor_version(be->pRoaNthw->mRoa) & 0xffff));
}


static int roa_tunhdr_flush(void *be_dev, const struct roa_func_s *roa, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, roa, be->pRoaNthw);

  if (roa->ver == 6) {
    RoaNthw_TunHdrCnt(be->pRoaNthw, 4);
    for (int i = 0; i < cnt; i++) {
      for (int ii = 0; ii < 4; ii++) {
        RoaNthw_TunHdrSelect(be->pRoaNthw, index + (i * 4) + ii);
        RoaNthw_TunHdrTunnelHdr(be->pRoaNthw, &roa->v6.tunhdr[index/4 + i].tunnel_hdr[ii*4]);
        RoaNthw_TunHdrFlush(be->pRoaNthw);
      }
    }
  }

  _CHECK_DEBUG_OFF(roa, be->pRoaNthw);
  return 0;
}

static int roa_tuncfg_flush(void *be_dev, const struct roa_func_s *roa,
                            int category, int cnt)
{
  struct backend_dev_s *be = (struct backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, roa, be->pRoaNthw);

  if (roa->ver == 6) {
    RoaNthw_TunCfgCnt(be->pRoaNthw, 1);
    for (int i = 0; i < cnt; i++) {
      RoaNthw_TunCfgSelect(be->pRoaNthw, category + i);
      RoaNthw_TunCfgTunLen(be->pRoaNthw, roa->v6.tuncfg[category + i].tun_len);
      RoaNthw_TunCfgTunType(be->pRoaNthw,  roa->v6.tuncfg[category + i].tun_type);
      RoaNthw_TunCfgTunVlan(be->pRoaNthw, roa->v6.tuncfg[category + i].tun_vlan);
      RoaNthw_TunCfgIpType(be->pRoaNthw, roa->v6.tuncfg[category + i].ip_type);
      RoaNthw_TunCfgIpcsUpd(be->pRoaNthw, roa->v6.tuncfg[category + i].ipcs_upd);
      RoaNthw_TunCfgIpcsPrecalc(be->pRoaNthw, roa->v6.tuncfg[category + i].ipcs_precalc);
      RoaNthw_TunCfgIptlUpd(be->pRoaNthw, roa->v6.tuncfg[category + i].iptl_upd);
      RoaNthw_TunCfgIptlPrecalc(be->pRoaNthw, roa->v6.tuncfg[category + i].iptl_precalc);
      RoaNthw_TunCfgVxlanUdpLenUpd(be->pRoaNthw, roa->v6.tuncfg[category + i].vxlan_udp_len_upd);
      RoaNthw_TunCfgTxLagIx(be->pRoaNthw, roa->v6.tuncfg[category + i].tx_lag_ix);
      RoaNthw_TunCfgRecirculate(be->pRoaNthw, roa->v6.tuncfg[category + i].recirculate);
      RoaNthw_TunCfgPushTunnel(be->pRoaNthw, roa->v6.tuncfg[category + i].push_tunnel);
      RoaNthw_TunCfgRecircPort(be->pRoaNthw, roa->v6.tuncfg[category + i].recirc_port);
      RoaNthw_TunCfgRecircBypass(be->pRoaNthw, roa->v6.tuncfg[category + i].recirc_bypass);
      RoaNthw_TunCfgFlush(be->pRoaNthw);
    }
  }

  _CHECK_DEBUG_OFF(roa, be->pRoaNthw);
  return 0;
}

static int roa_config_flush(void *be_dev, const struct roa_func_s *roa)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, roa, be->pRoaNthw);

  if (roa->ver == 6) {
    RoaNthw_ConfigFwdRecirculate (be->pRoaNthw, roa->v6.config->fwd_recirculate);
    RoaNthw_ConfigFwdNormalPcks (be->pRoaNthw, roa->v6.config->fwd_normal_pcks);
    RoaNthw_ConfigFwdTxPort0 (be->pRoaNthw, roa->v6.config->fwd_txport0);
    RoaNthw_ConfigFwdTxPort1 (be->pRoaNthw, roa->v6.config->fwd_txport1);
    RoaNthw_ConfigFwdCellBuilderPcks (be->pRoaNthw, roa->v6.config->fwd_cellbuilder_pcks);
    RoaNthw_ConfigFwdNonNormalPcks (be->pRoaNthw, roa->v6.config->fwd_non_normal_pcks);
    RoaNthw_ConfigFlush(be->pRoaNthw);
  }

  _CHECK_DEBUG_OFF(roa, be->pRoaNthw);
  return 0;
}

static int roa_lagcfg_flush(void *be_dev, const struct roa_func_s *roa, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, roa, be->pRoaNthw);

  if (roa->ver == 6) {
    RoaNthw_LagCfgCnt(be->pRoaNthw, 1);
    for (int i = 0; i < cnt; i++) {
      RoaNthw_LagCfgSelect(be->pRoaNthw, index + i);
      RoaNthw_LagCfgTxPhyPort  (be->pRoaNthw, roa->v6.lagcfg[index + i].txphy_port);
      RoaNthw_LagCfgFlush(be->pRoaNthw);
    }
  }

  _CHECK_DEBUG_OFF(roa, be->pRoaNthw);
  return 0;
}



/*
 *  *****************  RMC  *******************
 */

static bool rmc_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pRmcNthw != NULL;
}

static uint32_t rmc_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  return (uint32_t)((nthw_module_get_major_version(be->pRmcNthw->mRmc) << 16) |
                    (nthw_module_get_minor_version(be->pRmcNthw->mRmc) & 0xffff));
}

static int rmc_ctrl_flush(void *be_dev, const struct rmc_func_s *rmc)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, rmc, be->pRmcNthw);

  if (rmc->ver == 0x10003) {
    RmcNthw_CtrlBlockStatt(be->pRmcNthw, rmc->v1_3.ctrl->block_statt);
    RmcNthw_CtrlBlockKeepA(be->pRmcNthw, rmc->v1_3.ctrl->block_keepa);
    RmcNthw_CtrlBlockRppSlice(be->pRmcNthw, rmc->v1_3.ctrl->block_rpp_slice);
    RmcNthw_CtrlBlockMacPort(be->pRmcNthw, rmc->v1_3.ctrl->block_mac_port);
    RmcNthw_CtrlLagPhyOddEven(be->pRmcNthw, rmc->v1_3.ctrl->lag_phy_odd_even);
    RmcNthw_CtrlFlush(be->pRmcNthw);
  }

  _CHECK_DEBUG_OFF(rmc, be->pRmcNthw);
  return 0;
}

/*
 *  *****************  TPE  *******************
 */

static bool tpe_get_present(void *be_dev)
{
  struct backend_dev_s *be = (struct backend_dev_s*)be_dev;
  return be->pCsuNthw != NULL &&
         be->pHfuNthw != NULL && be->pRppLrNthw != NULL &&
         be->pTxCpyNthw != NULL && be->pTxInsNthw != NULL &&
         be->pTxRplNthw != NULL;
}

static uint32_t tpe_get_version(void *be_dev)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;

  const uint32_t csu_version = (uint32_t)(
    (nthw_module_get_major_version(be->pCsuNthw->mCsu) << 16) |
    (nthw_module_get_minor_version(be->pCsuNthw->mCsu) & 0xffff));

  const uint32_t hfu_version = (uint32_t)(
    (nthw_module_get_major_version(be->pHfuNthw->mHfu) << 16) |
    (nthw_module_get_minor_version(be->pHfuNthw->mHfu) & 0xffff));

  const uint32_t rpp_lr_version = (uint32_t)(
    (nthw_module_get_major_version(be->pRppLrNthw->mRppLr) << 16) |
    (nthw_module_get_minor_version(be->pRppLrNthw->mRppLr) & 0xffff));

  const uint32_t tx_cpy_version = (uint32_t)(
    (nthw_module_get_major_version(be->pTxCpyNthw->mTxCpy) << 16) |
    (nthw_module_get_minor_version(be->pTxCpyNthw->mTxCpy) & 0xffff));

  const uint32_t tx_ins_version = (uint32_t)(
    (nthw_module_get_major_version(be->pTxInsNthw->mTxIns) << 16) |
    (nthw_module_get_minor_version(be->pTxInsNthw->mTxIns) & 0xffff));

  const uint32_t tx_rpl_version = (uint32_t)(
    (nthw_module_get_major_version(be->pTxRplNthw->mTxRpl) << 16) |
    (nthw_module_get_minor_version(be->pTxRplNthw->mTxRpl) & 0xffff));

  if (csu_version == 0 &&
    hfu_version == 2 && rpp_lr_version == 1 && tx_cpy_version == 2 &&
    tx_ins_version == 1 && tx_rpl_version == 2) {
    return 3;
  }

  assert(false);
  return 0;
}

static int tpe_rpp_rcp_flush(void *be_dev, const struct tpe_func_s *rppLr, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, rppLr, be->pRppLrNthw);

  if (rppLr->ver >= 1) {
    RppLrNthw_RcpCnt(be->pRppLrNthw, 1);
    for (int i = 0; i < cnt; i++) {
      RppLrNthw_RcpSelect(be->pRppLrNthw, index + i);
      RppLrNthw_RcpExp(be->pRppLrNthw, rppLr->v3.rpp_rcp[index + i].exp);
      RppLrNthw_RcpFlush(be->pRppLrNthw);
    }
  }

  _CHECK_DEBUG_OFF(rppLr, be->pRppLrNthw);
  return 0;
}

static int tpe_rpp_ifr_rcp_flush(void *be_dev, const struct tpe_func_s *rppLr, int index, int cnt)
{
  int res = 0;
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, rppLr, be->pRppLrNthw);

  if (rppLr->ver >= 2) {
    RppLrNthw_IfrRcpCnt(be->pRppLrNthw, 1);
    for (int i = 0; i < cnt; i++) {
      RppLrNthw_IfrRcpSelect(be->pRppLrNthw, index + i);
      RppLrNthw_IfrRcpEn(be->pRppLrNthw, rppLr->v3.rpp_ifr_rcp[index + i].en);
      RppLrNthw_IfrRcpMtu(be->pRppLrNthw, rppLr->v3.rpp_ifr_rcp[index + i].mtu);
      RppLrNthw_IfrRcpFlush(be->pRppLrNthw);
    }
  }
  else {
    res = -1;
  }

  _CHECK_DEBUG_OFF(rppLr, be->pRppLrNthw);
  return res;
}

static int tpe_ifr_rcp_flush(void *be_dev, const struct tpe_func_s *ifr, int index, int cnt)
{
  int res = 0;
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, ifr, be->pIfrNthw);

  if (ifr->ver >= 2) {
    IfrNthw_RcpCnt(be->pIfrNthw, 1);
    for (int i = 0; i < cnt; i++) {
      IfrNthw_RcpSelect(be->pIfrNthw, index + i);
      IfrNthw_RcpEn(be->pIfrNthw, ifr->v3.ifr_rcp[index + i].en);
      IfrNthw_RcpMtu(be->pIfrNthw, ifr->v3.ifr_rcp[index + i].mtu);
      IfrNthw_RcpFlush(be->pIfrNthw);
    }
  }
  else {
    res = -1;
  }

  _CHECK_DEBUG_OFF(ifr, be->pIfrNthw);
  return res;
}

static int tpe_ins_rcp_flush(void *be_dev, const struct tpe_func_s *txIns, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, txIns, be->pTxInsNthw);

  if (txIns->ver >= 1) {
    TxInsNthw_RcpCnt(be->pTxInsNthw, 1);
    for (int i = 0; i < cnt; i++) {
      TxInsNthw_RcpSelect(be->pTxInsNthw, index + i);
      TxInsNthw_RcpDyn(be->pTxInsNthw, txIns->v3.ins_rcp[index + i].dyn);
      TxInsNthw_RcpOfs(be->pTxInsNthw, txIns->v3.ins_rcp[index + i].ofs);
      TxInsNthw_RcpLen(be->pTxInsNthw, txIns->v3.ins_rcp[index + i].len);
      TxInsNthw_RcpFlush(be->pTxInsNthw);
    }
  }

  _CHECK_DEBUG_OFF(txIns, be->pTxInsNthw);
  return 0;
}

static int tpe_rpl_rcp_flush(void *be_dev, const struct tpe_func_s *txRpl, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, txRpl, be->pTxRplNthw);

  if (txRpl->ver >= 1) {
    TxRplNthw_RcpCnt(be->pTxRplNthw, 1);
    for (int i = 0; i < cnt; i++) {
      TxRplNthw_RcpSelect(be->pTxRplNthw, index + i);
      TxRplNthw_RcpDyn(be->pTxRplNthw, txRpl->v3.rpl_rcp[index + i].dyn);
      TxRplNthw_RcpOfs(be->pTxRplNthw, txRpl->v3.rpl_rcp[index + i].ofs);
      TxRplNthw_RcpLen(be->pTxRplNthw, txRpl->v3.rpl_rcp[index + i].len);
      TxRplNthw_RcpRplPtr(be->pTxRplNthw, txRpl->v3.rpl_rcp[index + i].rpl_ptr);
      TxRplNthw_RcpExtPrio(be->pTxRplNthw, txRpl->v3.rpl_rcp[index + i].ext_prio);
      TxRplNthw_RcpFlush(be->pTxRplNthw);
    }
  }

  _CHECK_DEBUG_OFF(txRpl, be->pTxRplNthw);
  return 0;
}

static int tpe_rpl_ext_flush(void *be_dev, const struct tpe_func_s *txRpl, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, txRpl, be->pTxRplNthw);

  if (txRpl->ver >= 1) {
    TxRplNthw_ExtCnt(be->pTxRplNthw, 1);
    for (int i = 0; i < cnt; i++) {
      TxRplNthw_ExtSelect(be->pTxRplNthw, index + i);
      TxRplNthw_ExtRplPtr(be->pTxRplNthw, txRpl->v3.rpl_ext[index + i].rpl_ptr);
      TxRplNthw_ExtFlush(be->pTxRplNthw);
    }
  }

  _CHECK_DEBUG_OFF(txRpl, be->pTxRplNthw);
  return 0;
}

static int tpe_rpl_rpl_flush(void *be_dev, const struct tpe_func_s *txRpl, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, txRpl, be->pTxRplNthw);

  if (txRpl->ver >= 1) {
    TxRplNthw_RplCnt(be->pTxRplNthw, 1);
    for (int i = 0; i < cnt; i++) {
      TxRplNthw_RplSelect(be->pTxRplNthw, index + i);
      TxRplNthw_RplValue(be->pTxRplNthw, txRpl->v3.rpl_rpl[index + i].value);
      TxRplNthw_RplFlush(be->pTxRplNthw);
    }
  }

  _CHECK_DEBUG_OFF(txRpl, be->pTxRplNthw);
  return 0;
}

static int tpe_cpy_rcp_flush(void *be_dev, const struct tpe_func_s *txCpy, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  unsigned int wr_index = -1;

  _CHECK_DEBUG_ON(be, txCpy, be->pTxCpyNthw);

  if (txCpy->ver >= 1) {
    for (int i = 0; i < cnt; i++) {
      if (wr_index != (index + i) / txCpy->nb_rcp_categories) {
        wr_index = (index + i) / txCpy->nb_rcp_categories;
        TxCpyNthw_WriterCnt(be->pTxCpyNthw, wr_index, 1);
      }

      TxCpyNthw_WriterSelect(be->pTxCpyNthw, wr_index, (index + i) % txCpy->nb_rcp_categories);
      TxCpyNthw_WriterReaderSelect(be->pTxCpyNthw, wr_index, txCpy->v3.cpy_rcp[index + i].reader_select);
      TxCpyNthw_WriterDyn(be->pTxCpyNthw, wr_index, txCpy->v3.cpy_rcp[index + i].dyn);
      TxCpyNthw_WriterOfs(be->pTxCpyNthw, wr_index, txCpy->v3.cpy_rcp[index + i].ofs);
      TxCpyNthw_WriterLen(be->pTxCpyNthw, wr_index, txCpy->v3.cpy_rcp[index + i].len);
      TxCpyNthw_WriterFlush(be->pTxCpyNthw, wr_index);
    }
  }

  _CHECK_DEBUG_OFF(txCpy, be->pTxCpyNthw);
  return 0;
}

static int tpe_hfu_rcp_flush(void *be_dev, const struct tpe_func_s *hfu, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, hfu, be->pHfuNthw);

  if (hfu->ver >= 1) {
    HfuNthw_RcpCnt(be->pHfuNthw, 1);
    for (int i = 0; i < cnt; i++) {
      HfuNthw_RcpSelect(be->pHfuNthw, index + i);
      HfuNthw_RcpLenAWr(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_a_wr);
      HfuNthw_RcpLenAOl4len(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_a_outer_l4_len);
      HfuNthw_RcpLenAPosDyn(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_a_pos_dyn);
      HfuNthw_RcpLenAPosOfs(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_a_pos_ofs);
      HfuNthw_RcpLenAAddDyn(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_a_add_dyn);
      HfuNthw_RcpLenAAddOfs(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_a_add_ofs);
      HfuNthw_RcpLenASubDyn(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_a_sub_dyn);
      HfuNthw_RcpLenBWr(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_b_wr);
      HfuNthw_RcpLenBPosDyn(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_b_pos_dyn);
      HfuNthw_RcpLenBPosOfs(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_b_pos_ofs);
      HfuNthw_RcpLenBAddDyn(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_b_add_dyn);
      HfuNthw_RcpLenBAddOfs(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_b_add_ofs);
      HfuNthw_RcpLenBSubDyn(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_b_sub_dyn);
      HfuNthw_RcpLenCWr(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_c_wr);
      HfuNthw_RcpLenCPosDyn(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_c_pos_dyn);
      HfuNthw_RcpLenCPosOfs(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_c_pos_ofs);
      HfuNthw_RcpLenCAddDyn(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_c_add_dyn);
      HfuNthw_RcpLenCAddOfs(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_c_add_ofs);
      HfuNthw_RcpLenCSubDyn(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].len_c_sub_dyn);
      HfuNthw_RcpTtlWr(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].ttl_wr);
      HfuNthw_RcpTtlPosDyn(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].ttl_pos_dyn);
      HfuNthw_RcpTtlPosOfs(be->pHfuNthw, hfu->v3.hfu_rcp[index + i].ttl_pos_ofs);
      HfuNthw_RcpFlush(be->pHfuNthw);
    }
  }

  _CHECK_DEBUG_OFF(hfu, be->pHfuNthw);
  return 0;
}

static int tpe_csu_rcp_flush(void *be_dev, const struct tpe_func_s *csu, int index, int cnt)
{
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  _CHECK_DEBUG_ON(be, csu, be->pCsuNthw);

  if (csu->ver >= 1) {
    CsuNthw_RcpCnt(be->pCsuNthw, 1);
    for (int i = 0; i < cnt; i++) {
      CsuNthw_RcpSelect(be->pCsuNthw, index + i);
      CsuNthw_RcpOuterL3Cmd(be->pCsuNthw, csu->v3.csu_rcp[index + i].ol3_cmd);
      CsuNthw_RcpOuterL4Cmd(be->pCsuNthw, csu->v3.csu_rcp[index + i].ol4_cmd);
      CsuNthw_RcpInnerL3Cmd(be->pCsuNthw, csu->v3.csu_rcp[index + i].il3_cmd);
      CsuNthw_RcpInnerL4Cmd(be->pCsuNthw, csu->v3.csu_rcp[index + i].il4_cmd);
      CsuNthw_RcpFlush(be->pCsuNthw);
    }
  }

  _CHECK_DEBUG_OFF(csu, be->pCsuNthw);
  return 0;
}


/*
 *  *****************  DBS  *******************
 */

static int alloc_rx_queue(void *be_dev, int queue_id)
{
#if 0
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  HostBufferInfo *pHostBufferInfo = NULL;
  if (queue_id > 255) return -1;
  pHostBufferInfo = hbh4gaRxReserveHostBuffer(be->adapter_no, (uint8_t)queue_id);

  if (pHostBufferInfo) {
    printf("Got HB size %lu, Index %i, queue id %i\n",
        pHostBufferInfo->hostBufferSize, pHostBufferInfo->hostBufferIndex, (uint8_t)queue_id);

    return pHostBufferInfo->hostBufferIndex;
  }
#else
  (void)be_dev;
  (void)queue_id;
#endif
  printf("ERROR alloc Rx queue\n");
  return -1;
}


static int free_rx_queue(void *be_dev, int hw_queue)
{
#if 0
  struct  backend_dev_s *be = (struct  backend_dev_s *)be_dev;
  printf("Free HB queue id %i\n", (uint8_t)hw_queue);
  hbh4gaRxReleaseHostBuffer(be->adapter_no, hw_queue);
#else
  (void)be_dev;
  (void)hw_queue;
#endif
  printf("ERROR free Rx queue\n");
  return 0;
}


const struct flow_api_backend_ops flow_be_iface = {
  1,

  set_debug_mode,
  get_nb_phy_ports,
  get_nb_rx_ports,
  get_ltx_avail,
  get_nb_cat_funcs,
  get_nb_categories,
  get_nb_cat_km_if_cnt,
  get_nb_cat_km_if_m0,
  get_nb_cat_km_if_m1,
  get_nb_queues,
  get_nb_km_flow_types,
  get_nb_pm_ext,
  get_nb_len,
  get_kcc_size,
  get_kcc_banks,
  get_nb_km_categories,
  get_nb_km_cam_banks,
  get_nb_km_cam_record_words,
  get_nb_km_cam_records,
  get_nb_km_tcam_banks,
  get_nb_km_tcam_bank_width,
  get_nb_flm_categories,
  get_nb_flm_size_mb,
  get_nb_flm_entry_size,
  get_nb_flm_variant,
  get_nb_flm_prios,
  get_nb_flm_pst_profiles,
  get_nb_qsl_categories,
  get_nb_qsl_qst_entries,
  get_nb_pdb_categories,
  get_nb_ioa_categories,
  get_nb_roa_categories,
  get_nb_tpe_categories,
  get_nb_tx_cpy_writers,
  get_nb_tx_cpy_mask_mem,
  get_nb_tx_rpl_depth,
  get_nb_tx_rpl_ext_categories,
  get_nb_tpe_ifr_categories,

  alloc_rx_queue,
  free_rx_queue,

  cat_get_present,
  cat_get_version,
  cat_cfn_flush,

  cat_kce_flush,
  cat_kcs_flush,
  cat_fte_flush,

  cat_cte_flush,
  cat_cts_flush,
  cat_cot_flush,
  cat_cct_flush,
  cat_exo_flush,
  cat_rck_flush,
  cat_len_flush,
  cat_kcc_flush,
  cat_cce_flush,
  cat_ccs_flush,

  km_get_present,
  km_get_version,
  km_rcp_flush,
  km_cam_flush,
  km_tcam_flush,
  km_tci_flush,
  km_tcq_flush,

  flm_get_present,
  flm_get_version,
  flm_control_flush,
  flm_status_flush,
  flm_status_update,
  flm_timeout_flush,
  flm_scrub_flush,
  flm_load_bin_flush,
  flm_load_pps_flush,
  flm_load_lps_flush,
  flm_load_aps_flush,
  flm_prio_flush,
  flm_pst_flush,
  flm_rcp_flush,
  flm_buf_ctrl_update,
  flm_stat_update,
  flm_lrn_data_flush,
  flm_inf_data_update,
  flm_sta_data_update,

  hsh_get_present,
  hsh_get_version,
  hsh_rcp_flush,

  qsl_get_present,
  qsl_get_version,
  qsl_rcp_flush,
  qsl_qst_flush,
  qsl_qen_flush,
  qsl_unmq_flush,

  slc_get_present,
  slc_get_version,
  slc_rcp_flush,

  slc_lr_get_present,
  slc_lr_get_version,
  slc_lr_rcp_flush,

  pdb_get_present,
  pdb_get_version,
  pdb_rcp_flush,
  pdb_config_flush,

  ioa_get_present,
  ioa_get_version,
  ioa_rcp_flush,
  ioa_special_tpid_flush,
  ioa_roa_epp_flush,

  roa_get_present,
  roa_get_version,
  roa_tunhdr_flush,
  roa_tuncfg_flush,
  roa_config_flush,
  roa_lagcfg_flush,

  rmc_get_present,
  rmc_get_version,
  rmc_ctrl_flush,

  tpe_get_present,
  tpe_get_version,
  tpe_rpp_rcp_flush,
  tpe_rpp_ifr_rcp_flush,
  tpe_ifr_rcp_flush,
  tpe_ins_rcp_flush,
  tpe_rpl_rcp_flush,
  tpe_rpl_ext_flush,
  tpe_rpl_rpl_flush,
  tpe_cpy_rcp_flush,
  tpe_hfu_rcp_flush,
  tpe_csu_rcp_flush,
};

const struct flow_api_backend_ops *BinFlowBackendInit(nthw_fpga_t *pFpga,
                                                      void **dev)
{
  uint8_t physicalAdapterNo = (uint8_t)pFpga->p_fpga_info->adapterNo;

  struct InfoNthw *pinfonthw = InfoNthw_new();
  InfoNthw_init(pinfonthw, pFpga, physicalAdapterNo);
  be_devs[physicalAdapterNo].pInfoNthw = pinfonthw;

  /* Init nthw CAT */
  if (CatNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct CatNthw *pcatnthw = CatNthw_new();
    CatNthw_init(pcatnthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pCatNthw = pcatnthw;
  }
  else {
    be_devs[physicalAdapterNo].pCatNthw = NULL;
  }

  /* Init nthw KM */
  if (KmNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct KmNthw *pkmnthw = KmNthw_new();
    KmNthw_init(pkmnthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pKmNthw = pkmnthw;
  }
  else {
    be_devs[physicalAdapterNo].pKmNthw = NULL;
  }

  /* Init nthw FLM */
  if (FlmNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct FlmNthw *pflmnthw = FlmNthw_new();
    FlmNthw_init(pflmnthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pFlmNthw = pflmnthw;
  }
  else {
    be_devs[physicalAdapterNo].pFlmNthw = NULL;
  }

  /* Init nthw IFR */
  if (IfrNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct IfrNthw *ifrnthw = IfrNthw_new();
    IfrNthw_init(ifrnthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pIfrNthw = ifrnthw;
  }
  else {
    be_devs[physicalAdapterNo].pIfrNthw = NULL;
  }

  /* Init nthw HSH */
  if (HshNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct HshNthw *phshnthw = HshNthw_new();
    HshNthw_init(phshnthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pHshNthw = phshnthw;
  }
  else {
    be_devs[physicalAdapterNo].pHshNthw = NULL;
  }

  /* Init nthw QSL */
  if (QslNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct QslNthw *pqslnthw = QslNthw_new();
    QslNthw_init(pqslnthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pQslNthw = pqslnthw;
  }
  else {
    be_devs[physicalAdapterNo].pQslNthw = NULL;
  }

  /* Init nthw SLC */
  if (SlcNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct SlcNthw *pslcnthw = SlcNthw_new();
    SlcNthw_init(pslcnthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pSlcNthw = pslcnthw;
  }
  else {
    be_devs[physicalAdapterNo].pSlcNthw = NULL;
  }

  /* Init nthw SLC LR */
  if (SlcLrNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct SlcLrNthw *pslclrnthw = SlcLrNthw_new();
    SlcLrNthw_init(pslclrnthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pSlcLrNthw = pslclrnthw;
  }
  else {
    be_devs[physicalAdapterNo].pSlcLrNthw = NULL;
  }

  /* Init nthw PDB */
  if (PdbNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct PdbNthw *ppdbnthw = PdbNthw_new();
    PdbNthw_init(ppdbnthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pPdbNthw = ppdbnthw;
  }
  else {
    be_devs[physicalAdapterNo].pPdbNthw = NULL;
  }

  /* Init nthw IOA */
  if (IoaNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct IoaNthw *pioanthw = IoaNthw_new();
    IoaNthw_init(pioanthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pIoaNthw = pioanthw;
  }
  else {
    be_devs[physicalAdapterNo].pIoaNthw = NULL;
  }

  /* Init nthw ROA */
  if (RoaNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct RoaNthw *proanthw = RoaNthw_new();
    RoaNthw_init(proanthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pRoaNthw = proanthw;
  }
  else {
    be_devs[physicalAdapterNo].pRoaNthw = NULL;
  }

  /* Init nthw RMC */
  if (RmcNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct RmcNthw *prmcnthw = RmcNthw_new();
    RmcNthw_init(prmcnthw, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pRmcNthw = prmcnthw;
  }
  else {
    be_devs[physicalAdapterNo].pRmcNthw = NULL;
  }

  /* Init nthw HFU */
  if (HfuNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct HfuNthw *ptr = HfuNthw_new();
    HfuNthw_init(ptr, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pHfuNthw = ptr;
  }
  else {
    be_devs[physicalAdapterNo].pHfuNthw = NULL;
  }

  /* Init nthw RPP_LR */
  if (RppLrNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct RppLrNthw *ptr = RppLrNthw_new();
    RppLrNthw_init(ptr, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pRppLrNthw = ptr;
  }
  else {
    be_devs[physicalAdapterNo].pRppLrNthw = NULL;
  }

  /* Init nthw TX_CPY */
  if (TxCpyNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct TxCpyNthw *ptr = TxCpyNthw_new();
    TxCpyNthw_init(ptr, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pTxCpyNthw = ptr;
  }
  else {
    be_devs[physicalAdapterNo].pTxCpyNthw = NULL;
  }

  /* Init nthw CSU */
  if (CsuNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct CsuNthw *ptr = CsuNthw_new();
    CsuNthw_init(ptr, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pCsuNthw = ptr;
  }
  else {
    be_devs[physicalAdapterNo].pCsuNthw = NULL;
  }

  /* Init nthw TX_INS */
  if (TxInsNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct TxInsNthw *ptr = TxInsNthw_new();
    TxInsNthw_init(ptr, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pTxInsNthw = ptr;
  }
  else {
    be_devs[physicalAdapterNo].pTxInsNthw = NULL;
  }

  /* Init nthw TX_RPL */
  if (TxRplNthw_init(NULL, pFpga, physicalAdapterNo) == 0) {
    struct TxRplNthw *ptr = TxRplNthw_new();
    TxRplNthw_init(ptr, pFpga, physicalAdapterNo);
    be_devs[physicalAdapterNo].pTxRplNthw = ptr;
  }
  else {
    be_devs[physicalAdapterNo].pTxRplNthw = NULL;
  }

  be_devs[physicalAdapterNo].adapter_no = physicalAdapterNo;
  *dev = (void *)&be_devs[physicalAdapterNo];

  return &flow_be_iface;
}

void BinFlowBackendDone(void *dev)
{
  struct backend_dev_s *be_dev = (struct backend_dev_s *)dev;
  InfoNthw_delete(be_dev->pInfoNthw);
  CatNthw_delete(be_dev->pCatNthw);
  KmNthw_delete(be_dev->pKmNthw);
  FlmNthw_delete(be_dev->pFlmNthw);
  HshNthw_delete(be_dev->pHshNthw);
  QslNthw_delete(be_dev->pQslNthw);
  SlcNthw_delete(be_dev->pSlcNthw);
  SlcLrNthw_delete(be_dev->pSlcLrNthw);
  PdbNthw_delete(be_dev->pPdbNthw);
  IoaNthw_delete(be_dev->pIoaNthw);
  RoaNthw_delete(be_dev->pRoaNthw);
  RmcNthw_delete(be_dev->pRmcNthw);
  CsuNthw_delete(be_dev->pCsuNthw);
  HfuNthw_delete(be_dev->pHfuNthw);
  RppLrNthw_delete(be_dev->pRppLrNthw);
  TxCpyNthw_delete(be_dev->pTxCpyNthw);
  TxInsNthw_delete(be_dev->pTxInsNthw);
  TxRplNthw_delete(be_dev->pTxRplNthw);
}
