/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"
#include "nthw_drv.h"
#include "nthw_register.h"

#include "flow_nthw_hfu.h"

#include <stdlib.h>
#include <string.h>

void hfuNthw_setDebugMode(struct HfuNthw *p, unsigned int nDebugMode)
{
  nthw_module_set_debug_mode(p->mHfu, nDebugMode);
}

struct HfuNthw *HfuNthw_new(void)
{
  struct HfuNthw *p = malloc(sizeof(struct HfuNthw));
  if (p) {
    (void)memset(p, 0, sizeof(*p));
  }
  return p;
}

void HfuNthw_delete(struct HfuNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int HfuNthw_init(struct HfuNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->p_fpga_info->mpAdapterIdStr;
  nthw_module_t* pMod = nthw_fpga_query_module(pFpga, MOD_HFU, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: Hfu %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mHfu = nthw_fpga_query_module(pFpga, MOD_HFU, nInstance);

  p->mpRcpCtrl = nthw_module_get_register(p->mHfu, HFU_RCP_CTRL);
  p->mpRcpAddr = nthw_register_get_field(p->mpRcpCtrl, HFU_RCP_CTRL_ADR);
  p->mpRcpCnt = nthw_register_get_field(p->mpRcpCtrl, HFU_RCP_CTRL_CNT);

  p->mpRcpData = nthw_module_get_register(p->mHfu, HFU_RCP_DATA);
  p->mpRcpDataLenAWr = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_A_WR);
  p->mpRcpDataLenAOl4len = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_A_OL4LEN);
  p->mpRcpDataLenAPosDyn = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_A_POS_DYN);
  p->mpRcpDataLenAPosOfs = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_A_POS_OFS);
  p->mpRcpDataLenAAddDyn = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_A_ADD_DYN);
  p->mpRcpDataLenAAddOfs = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_A_ADD_OFS);
  p->mpRcpDataLenASubDyn = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_A_SUB_DYN);
  p->mpRcpDataLenBWr = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_B_WR);
  p->mpRcpDataLenBPosDyn = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_B_POS_DYN);
  p->mpRcpDataLenBPosOfs = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_B_POS_OFS);
  p->mpRcpDataLenBAddDyn = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_B_ADD_DYN);
  p->mpRcpDataLenBAddOfs = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_B_ADD_OFS);
  p->mpRcpDataLenBSubDyn = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_B_SUB_DYN);
  p->mpRcpDataLenCWr = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_C_WR);
  p->mpRcpDataLenCPosDyn = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_C_POS_DYN);
  p->mpRcpDataLenCPosOfs = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_C_POS_OFS);
  p->mpRcpDataLenCAddDyn = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_C_ADD_DYN);
  p->mpRcpDataLenCAddOfs = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_C_ADD_OFS);
  p->mpRcpDataLenCSubDyn = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_LEN_C_SUB_DYN);
  p->mpRcpDataTtlWr = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_TTL_WR);
  p->mpRcpDataTtlPosDyn = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_TTL_POS_DYN);
  p->mpRcpDataTtlPosOfs = nthw_register_get_field(p->mpRcpData, HFU_RCP_DATA_TTL_POS_OFS);

  return 0;
}

void HfuNthw_RcpSelect(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpAddr, val);
}

void HfuNthw_RcpCnt(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpCnt, val);
}

void HfuNthw_RcpLenAWr(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenAWr, val);
}

void HfuNthw_RcpLenAOl4len(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenAOl4len, val);
}

void HfuNthw_RcpLenAPosDyn(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenAPosDyn, val);
}

void HfuNthw_RcpLenAPosOfs(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenAPosOfs, val);
}

void HfuNthw_RcpLenAAddDyn(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenAAddDyn, val);
}

void HfuNthw_RcpLenAAddOfs(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenAAddOfs, val);
}

void HfuNthw_RcpLenASubDyn(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenASubDyn, val);
}

void HfuNthw_RcpLenBWr(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenBWr, val);
}

void HfuNthw_RcpLenBPosDyn(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenBPosDyn, val);
}

void HfuNthw_RcpLenBPosOfs(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenBPosOfs, val);
}

void HfuNthw_RcpLenBAddDyn(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenBAddDyn, val);
}

void HfuNthw_RcpLenBAddOfs(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenBAddOfs, val);
}

void HfuNthw_RcpLenBSubDyn(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenBSubDyn, val);
}

void HfuNthw_RcpLenCWr(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenCWr, val);
}

void HfuNthw_RcpLenCPosDyn(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenCPosDyn, val);
}

void HfuNthw_RcpLenCPosOfs(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenCPosOfs, val);
}

void HfuNthw_RcpLenCAddDyn(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenCAddDyn, val);
}

void HfuNthw_RcpLenCAddOfs(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenCAddOfs, val);
}

void HfuNthw_RcpLenCSubDyn(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataLenCSubDyn, val);
}

void HfuNthw_RcpTtlWr(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTtlWr, val);
}

void HfuNthw_RcpTtlPosDyn(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTtlPosDyn, val);
}

void HfuNthw_RcpTtlPosOfs(const struct HfuNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTtlPosOfs, val);
}

void HfuNthw_RcpFlush(const struct HfuNthw *p)
{
  nthw_register_flush(p->mpRcpCtrl, 1);
  nthw_register_flush(p->mpRcpData, 1);
}
