/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"

#include "flow_nthw_qsl.h"

#include <stdlib.h> /* malloc */
#include <string.h> /* memset */

void qslNthw_setDebugMode(struct QslNthw *p, unsigned int nDebugMode)
{
  nthw_module_set_debug_mode(p->mQsl, nDebugMode);
}

struct QslNthw *QslNthw_new(void)
{
  struct QslNthw *p = malloc(sizeof(struct QslNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void QslNthw_delete(struct QslNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int QslNthw_init(struct QslNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->p_fpga_info->mpAdapterIdStr;
  nthw_module_t* pMod = nthw_fpga_query_module(pFpga, MOD_QSL, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: QSL %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mQsl = pMod;

#if defined(DEBUG) && (0)
  nthw_module_set_debug_mode(p->mQsl, 0xFF); /* DEBUG: */
#endif

  /* RCP */
  p->mpRcpCtrl = nthw_module_get_register(p->mQsl, QSL_RCP_CTRL);
  p->mpRcpAddr = nthw_register_get_field(p->mpRcpCtrl, QSL_RCP_CTRL_ADR);
  p->mpRcpCnt = nthw_register_get_field(p->mpRcpCtrl, QSL_RCP_CTRL_CNT);
  p->mpRcpData = nthw_module_get_register(p->mQsl, QSL_RCP_DATA);
  p->mpRcpDataDiscard = nthw_register_get_field(p->mpRcpData, QSL_RCP_DATA_DISCARD);
  p->mpRcpDataDrop = nthw_register_get_field(p->mpRcpData, QSL_RCP_DATA_DROP);
  p->mpRcpDataTblLo = nthw_register_get_field(p->mpRcpData, QSL_RCP_DATA_TBL_LO);
  p->mpRcpDataTblHi = nthw_register_get_field(p->mpRcpData, QSL_RCP_DATA_TBL_HI);
  p->mpRcpDataTblIdx = nthw_register_get_field(p->mpRcpData, QSL_RCP_DATA_TBL_IDX);
  p->mpRcpDataTblMsk = nthw_register_get_field(p->mpRcpData, QSL_RCP_DATA_TBL_MSK);
  p->mpRcpDataCao = nthw_register_query_field(p->mpRcpData, QSL_RCP_DATA_CAO);
  p->mpRcpDataLr = nthw_register_query_field(p->mpRcpData, QSL_RCP_DATA_LR);
  p->mpRcpDataTsa = nthw_register_query_field(p->mpRcpData, QSL_RCP_DATA_TSA);
  p->mpRcpDataVli = nthw_register_query_field(p->mpRcpData, QSL_RCP_DATA_VLI);

  /* QST */
  p->mpQstCtrl = nthw_module_get_register(p->mQsl, QSL_QST_CTRL);
  p->mpQstAddr = nthw_register_get_field(p->mpQstCtrl, QSL_QST_CTRL_ADR);
  p->mpQstCnt = nthw_register_get_field(p->mpQstCtrl, QSL_QST_CTRL_CNT);
  p->mpQstData = nthw_module_get_register(p->mQsl, QSL_QST_DATA);
  p->mpQstDataQueue = nthw_register_get_field(p->mpQstData, QSL_QST_DATA_QUEUE);
  p->mpQstDataEn = nthw_register_query_field(p->mpQstData, QSL_QST_DATA_EN);
  p->mpQstDataTxPort = nthw_register_query_field(p->mpQstData, QSL_QST_DATA_TX_PORT);
  p->mpQstDataLre = nthw_register_query_field(p->mpQstData, QSL_QST_DATA_LRE);
  p->mpQstDataTci = nthw_register_query_field(p->mpQstData, QSL_QST_DATA_TCI);
  p->mpQstDataVen = nthw_register_query_field(p->mpQstData, QSL_QST_DATA_VEN);
  /* QEN */
  p->mpQenCtrl = nthw_module_get_register(p->mQsl, QSL_QEN_CTRL);
  p->mpQenAddr = nthw_register_get_field(p->mpQenCtrl, QSL_QEN_CTRL_ADR);
  p->mpQenCnt = nthw_register_get_field(p->mpQenCtrl, QSL_QEN_CTRL_CNT);
  p->mpQenData = nthw_module_get_register(p->mQsl, QSL_QEN_DATA);
  p->mpQenDataEn = nthw_register_get_field(p->mpQenData, QSL_QEN_DATA_EN);
  /* UNMQ */
  p->mpUnmqCtrl = nthw_module_get_register(p->mQsl, QSL_UNMQ_CTRL);
  p->mpUnmqAddr = nthw_register_get_field(p->mpUnmqCtrl, QSL_UNMQ_CTRL_ADR);
  p->mpUnmqCnt = nthw_register_get_field(p->mpUnmqCtrl, QSL_UNMQ_CTRL_CNT);
  p->mpUnmqData = nthw_module_get_register(p->mQsl, QSL_UNMQ_DATA);
  p->mpUnmqDataDestQueue = nthw_register_get_field(p->mpUnmqData, QSL_UNMQ_DATA_DEST_QUEUE);
  p->mpUnmqDataEn = nthw_register_get_field(p->mpUnmqData, QSL_UNMQ_DATA_EN);

  if (!p->mpQstDataEn) {
    /* changed name from EN to QEN in v0.7 */
    p->mpQstDataEn = nthw_register_get_field(p->mpQstData, QSL_QST_DATA_QEN);
  }

  /* LTX - not there anymore from v0.7+ */
  p->mpLtxCtrl = nthw_module_query_register(p->mQsl, QSL_LTX_CTRL);
  if (p->mpLtxCtrl) {
    p->mpLtxAddr = nthw_register_get_field(p->mpLtxCtrl, QSL_LTX_CTRL_ADR);
    p->mpLtxCnt = nthw_register_get_field(p->mpLtxCtrl, QSL_LTX_CTRL_CNT);
  }
  else {
    p->mpLtxAddr = NULL;
    p->mpLtxCnt = NULL;
  }
  p->mpLtxData = nthw_module_query_register(p->mQsl, QSL_LTX_DATA);
  if (p->mpLtxData) {
    p->mpLtxDataLr = nthw_register_get_field(p->mpLtxData, QSL_LTX_DATA_LR);
    p->mpLtxDataTxPort = nthw_register_get_field(p->mpLtxData, QSL_LTX_DATA_TX_PORT);
    p->mpLtxDataTsa = nthw_register_get_field(p->mpLtxData, QSL_LTX_DATA_TSA);
  }
  else {
    p->mpLtxDataLr = NULL;
    p->mpLtxDataTxPort = NULL;
    p->mpLtxDataTsa = NULL;
  }
  return 0;
}

int QslNthw_setup(struct QslNthw *p, int nIdx, int nIdxCnt)
{
  (void)p;
  (void)nIdx;
  (void)nIdxCnt;

  return 0;
}

/* RCP */
void QslNthw_RcpSelect(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpAddr, val);
};

void QslNthw_RcpCnt(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpCnt, val);
}

void QslNthw_RcpDiscard(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataDiscard, val);
}

void QslNthw_RcpDrop(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataDrop, val);
}

void QslNthw_RcpTblLo(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTblLo, val);
}
void QslNthw_RcpTblHi(const struct QslNthw *p, uint32_t val)

{
  nthw_field_set_val32(p->mpRcpDataTblHi, val);
}
void QslNthw_RcpTblIdx(const struct QslNthw *p, uint32_t val)

{
  nthw_field_set_val32(p->mpRcpDataTblIdx, val);
}

void QslNthw_RcpTblMsk(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTblMsk, val);
}

void QslNthw_RcpCao(const struct QslNthw *p, uint32_t val)
{
  if (p->mpRcpDataCao) {
    nthw_field_set_val32(p->mpRcpDataCao, val);
  }
}

void QslNthw_RcpLr(const struct QslNthw *p, uint32_t val)
{
  if (p->mpRcpDataLr) {
    nthw_field_set_val32(p->mpRcpDataLr, val);
  }
}

void QslNthw_RcpTsa(const struct QslNthw *p, uint32_t val)
{
  if (p->mpRcpDataTsa) {
    nthw_field_set_val32(p->mpRcpDataTsa, val);
  }
}

void QslNthw_RcpVli(const struct QslNthw *p, uint32_t val)
{
  if (p->mpRcpDataVli) {
    nthw_field_set_val32(p->mpRcpDataVli, val);
  }
}

void QslNthw_RcpFlush(const struct QslNthw *p)
{
  nthw_register_flush(p->mpRcpCtrl, 1);
  nthw_register_flush(p->mpRcpData, 1);
}

  /* LTX */
void QslNthw_LtxSelect(const struct QslNthw *p, uint32_t val)
{
  if (p->mpLtxAddr) {
    nthw_field_set_val32(p->mpLtxAddr, val);
  }
}

void QslNthw_LtxCnt(const struct QslNthw *p, uint32_t val)
{
  if (p->mpLtxAddr) {
    nthw_field_set_val32(p->mpLtxCnt, val);
  }
}

void QslNthw_LtxLr(const struct QslNthw *p, uint32_t val)
{
  if (p->mpLtxDataLr) {
    nthw_field_set_val32(p->mpLtxDataLr, val);
  }
}

void QslNthw_LtxTxPort(const struct QslNthw *p, uint32_t val)
{
  if (p->mpLtxDataTxPort) {
    nthw_field_set_val32(p->mpLtxDataTxPort, val);
  }
}

void QslNthw_LtxTsa(const struct QslNthw *p, uint32_t val)
{
  if (p->mpLtxDataTsa) {
    nthw_field_set_val32(p->mpLtxDataTsa, val);
  }
};

void QslNthw_LtxFlush(const struct QslNthw *p)
{
  nthw_register_flush(p->mpLtxCtrl, 1);
  nthw_register_flush(p->mpLtxData, 1);
}

/* QST */
void QslNthw_QstSelect(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpQstAddr, val);
}

void QslNthw_QstCnt(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpQstCnt, val);
}

void QslNthw_QstQueue(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpQstDataQueue, val);
}

void QslNthw_QstEn(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpQstDataEn, val);
}

void QslNthw_QstTxPort(const struct QslNthw *p, uint32_t val)
{
  if (p->mpQstDataTxPort) {
    nthw_field_set_val32(p->mpQstDataTxPort, val);
  }
}

void QslNthw_QstLre(const struct QslNthw *p, uint32_t val)
{
  if (p->mpQstDataLre) {
    nthw_field_set_val32(p->mpQstDataLre, val);
  }
}

void QslNthw_QstTci(const struct QslNthw *p, uint32_t val)
{
  if (p->mpQstDataTci) {
    nthw_field_set_val32(p->mpQstDataTci, val);
  }
}

void QslNthw_QstVen(const struct QslNthw *p, uint32_t val)
{
  if (p->mpQstDataVen) {
    nthw_field_set_val32(p->mpQstDataVen, val);
  }
}

void QslNthw_QstFlush(const struct QslNthw *p)
{
  nthw_register_flush(p->mpQstCtrl, 1);
  nthw_register_flush(p->mpQstData, 1);
}

/* QEN */
void QslNthw_QenSelect(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpQenAddr, val);
}

void QslNthw_QenCnt(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpQenCnt, val);
}

void QslNthw_QenEn(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpQenDataEn, val);
}

void QslNthw_QenFlush(const struct QslNthw *p)
{
  nthw_register_flush(p->mpQenCtrl, 1);
  nthw_register_flush(p->mpQenData, 1);
}

/* UNMQ */
void QslNthw_UnmqSelect(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpUnmqAddr, val);
}

void QslNthw_UnmqCnt(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpUnmqCnt, val);
}

void QslNthw_UnmqDestQueue(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpUnmqDataDestQueue, val);
}

void QslNthw_UnmqEn(const struct QslNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpUnmqDataEn, val);
}

void QslNthw_UnmqFlush(const struct QslNthw *p)
{
  nthw_register_flush(p->mpUnmqCtrl, 1);
  nthw_register_flush(p->mpUnmqData, 1);
}
