/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"

#include "flow_nthw_slc.h"

#include <stdlib.h> /* malloc */
#include <string.h> /* memset */

void slcNthw_setDebugMode(struct SlcNthw *p, unsigned int nDebugMode)
{
  nthw_module_set_debug_mode(p->mSlc, nDebugMode);
}

struct SlcNthw *SlcNthw_new(void)
{
  struct SlcNthw *p = malloc(sizeof(struct SlcNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void SlcNthw_delete(struct SlcNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int SlcNthw_init(struct SlcNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->p_fpga_info->mpAdapterIdStr;
  nthw_module_t* pMod = nthw_fpga_query_module(pFpga, MOD_SLC, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: Slc %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mSlc = nthw_fpga_query_module(pFpga, MOD_SLC, nInstance);

  /* RCP */
  p->mpRcpCtrl = nthw_module_get_register(p->mSlc, SLC_RCP_CTRL);
  p->mpRcpAddr = nthw_register_get_field(p->mpRcpCtrl, SLC_RCP_CTRL_ADR);
  p->mpRcpCnt = nthw_register_get_field(p->mpRcpCtrl, SLC_RCP_CTRL_CNT);
  p->mpRcpData = nthw_module_get_register(p->mSlc, SLC_RCP_DATA);
  p->mpRcpDataTailSlcEn = nthw_register_get_field(p->mpRcpData, SLC_RCP_DATA_TAIL_SLC_EN);
  p->mpRcpDataTailDyn = nthw_register_get_field(p->mpRcpData, SLC_RCP_DATA_TAIL_DYN);
  p->mpRcpDataTailOfs = nthw_register_get_field(p->mpRcpData, SLC_RCP_DATA_TAIL_OFS);
  p->mpRcpDataPcap = nthw_register_get_field(p->mpRcpData, SLC_RCP_DATA_PCAP);

  return 0;
}

/* RCP */
void SlcNthw_RcpSelect(const struct SlcNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpAddr, val);
}

void SlcNthw_RcpCnt(const struct SlcNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpCnt, val);
}

void SlcNthw_RcpTailSlcEn(const struct SlcNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTailSlcEn, val);
}

void SlcNthw_RcpTailDyn(const struct SlcNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTailDyn, val);
}

void SlcNthw_RcpTailOfs(const struct SlcNthw *p, int32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTailOfs, val);
}

void SlcNthw_RcpPcap(const struct SlcNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataPcap, val);
}

void SlcNthw_RcpFlush(const struct SlcNthw *p)
{
  nthw_register_flush(p->mpRcpCtrl, 1);
  nthw_register_flush(p->mpRcpData, 1);
}
