/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include <unistd.h>
#include <sched.h>
#include <rte_vhost.h>
#include <linux/virtio_net.h>
#include <rte_mbuf.h>
#include <rte_cycles.h>
#include <rte_meter.h>
#include <rte_ethdev.h>
#include <rte_bus_pci.h>

#include "ntlog.h"

#include "ntnic_ethdev.h"
#include "ntnic_relay.h"

#if 0
#define MEASURE_CPU_UTIL
#define DUMP_PKTS_TO_VM
#define DUMP_PKTS_FROM_VM
#endif

/*
 * Tx relay copy process. From VM virtio queues to mbuf
 *
 * It is important to keep this mbuf memory buffer large for performance (8192)
 * It is used to request packets from virtio queues using the dpdk vhost library
 *
 * 2 copies are made at the moment. virtio ->cpy-> mbuf(this one) ->cpy-> nt vq
 */
#define MBUF_MEM_POOL_SIZE   (2048*4)

enum {VIRTIO_RXQ, VIRTIO_TXQ, VIRTIO_QNUM};

#define IF_NAME_SZ 256

#define DVIO_MAX_QUEUES (MAX_QUEUES)

#define MAX_DVIO_CONNECTIONS  64

/* structure to handle Rx and Tx relay functions */
struct queue {
  struct hwq_s hwq;
  struct hwq_s hwq_tx;
  int core;
  pthread_t tids;
  struct rte_mempool *mb_pool;
  /* Handle to call from PMD drive first queue (queue 0) virtio data to and from NIC */
  void *execute_handle;
};

struct dvio_conn {
  char socket[IF_NAME_SZ];

  int hw_port_id;
  int first_queue_hw_id;

  int num_descr;
  int num_descr_tx;

  int rxqs;
  int txqs;

  struct queue queue[DVIO_MAX_QUEUES];
};
static struct dvio_conn dvio_conns[MAX_DVIO_CONNECTIONS];

#define MAX_DVIO_INSTANCE_NUM  64
#define NUM_MBUFS   32


struct rx_dvio {
  int vid;
  uint16_t queue_id;
  uint32_t port;
  int running;

  int vf_num;  /* info only */
  struct nthw_virt_queue  *vq;

  struct rte_mbuf base_bufs[NUM_MBUFS*2];
  struct rte_mbuf *bufs[NUM_MBUFS];
  struct rte_mbuf *seg_bufs[NUM_MBUFS];
};

struct tx_dvio {
  int vid;
  uint16_t queue_id;
  uint32_t port;
  int running;
  int core;  /* if pinned then on what core */

  int vf_num;  /* info only */
  struct nthw_virt_queue  *vq;
  int num_descr_tx;

  struct rte_mempool *mb_pool;

  struct rx_dvio *hdl_rx; /* to run Rx relay function from Tx relay core */

};
static struct rx_dvio rx_dvios[MAX_DVIO_INSTANCE_NUM][DVIO_MAX_QUEUES];
static struct tx_dvio tx_dvios[MAX_DVIO_INSTANCE_NUM][DVIO_MAX_QUEUES];

static int get_dvio_conn_vhid(const char *socket_path)
{
  int i;
  for (i = 0; i < MAX_DVIO_CONNECTIONS; i++) {
    if (strncmp(dvio_conns[i].socket, socket_path, IF_NAME_SZ) == 0) {
      return i;
    }
  }
  return -1;
}



#if defined(DUMP_PKTS_TO_VM) || defined(DUMP_PKTS_FROM_VM)
static inline void dump_packet_seg(const char *header, uint8_t *bb, uint32_t data_len)
{
  uint32_t x;
  char string[65536], val[32];
  strlcpy(string, header, 65535);
  for(x=0; x < data_len; x++) {
    if(!(x%16)) {
      sprintf(val,"\n>>%02d :", x);
      strcat(string, val);
    }
    sprintf(val, " %02X", *(bb+x));
    strcat(string, val);
  }

  if (string[0]) {
    NT_LOG(INF, ETHDEV, "\n%s\n", string);
  }
}
#endif



static inline int stdvio_send_to_vm(struct rx_dvio *rx_dvio)
{
  struct nthw_received_packets hw_recv[NUM_MBUFS];

  if (rx_dvio == NULL) return 0;

  uint16_t whole_pkts;
  uint16_t hw_recv_pkts = nthw_get_rx_packets(rx_dvio->vq, NUM_MBUFS, hw_recv, &whole_pkts);
  if (!whole_pkts) return 0;

  /* We have at least one packet in NT segment */
  int i, nb_pkts = 0, nb = 0;

  for (i = 0; i < whole_pkts; i++) {

    struct rte_mbuf *mbuf = rx_dvio->bufs[i];
    rte_mbuf_refcnt_set(mbuf, 1);
    rte_pktmbuf_reset(mbuf);

    struct _pkt_hdr_rx *phdr = (struct _pkt_hdr_rx *)hw_recv[nb++].addr;
    uint16_t data_len = (uint16_t)phdr->cap_len;

    uint16_t len = data_len > SG_HW_RX_PKT_BUFFER_SIZE ? SG_HW_RX_PKT_BUFFER_SIZE : data_len;
      mbuf->data_off = SG_HDR_SIZE;
      mbuf->buf_addr = (uint8_t *)phdr;
      mbuf->buf_len  = data_len - SG_HDR_SIZE;
      mbuf->pkt_len  = data_len - SG_HDR_SIZE;
      mbuf->data_len = len - SG_HDR_SIZE;

      data_len -= len;
      struct rte_mbuf *mbuf_sg = mbuf;
      while (data_len > 0) {
        len = data_len > SG_HW_RX_PKT_BUFFER_SIZE ? SG_HW_RX_PKT_BUFFER_SIZE : data_len;
        mbuf_sg->next = rx_dvio->seg_bufs[nb];
        mbuf_sg = mbuf_sg->next;

        rte_mbuf_refcnt_set(mbuf_sg, 1);
        rte_pktmbuf_reset(mbuf_sg);

        mbuf_sg->data_off = 0;
        mbuf_sg->buf_addr = (uint8_t *)hw_recv[nb].addr;
        mbuf_sg->buf_len  = len;
        mbuf_sg->pkt_len  = len;
        mbuf_sg->data_len = len;

        mbuf->nb_segs++;
        data_len -= len;
        if (++nb > NUM_MBUFS)
          break;
      }

    nb_pkts++;

#ifdef DUMP_PKTS_TO_VM
    {
      struct rte_mbuf *m = mbuf;

      printf("\nRelay Core To VM: vid %i, Rx vq-queue %i, pkt #%i, rx on VF # %i, hdr rx port %i, tx to port %i, queue id %i, len %i, segs %i\n", rx_dvio->vid,
          rx_dvio->queue_id * VIRTIO_QNUM, nb_pkts,
          rx_dvio->vf_num,
          phdr->port,
          rx_dvio->port, rx_dvio->queue_id, m->pkt_len,
          m->nb_segs);

      for (int ii = 0; m && ii < mbuf->nb_segs; ii++) {
        printf("  seg %i len %i\n", ii, m->data_len);
        printf("  seg dump:\n");
        dump_packet_seg("mbuf seg:", rte_pktmbuf_mtod(m, uint8_t *), m->data_len);
        m = m->next;
      }
    }
#endif
  }

  /* next mbuf burst ready - send it up the virtio channel */
  struct rte_mbuf **cur_pkts = &rx_dvio->bufs[0];
  int cnt = nb_pkts;
  unsigned int tx_pkts;
  unsigned int retry = 0;

  do {
    tx_pkts = rte_vhost_enqueue_burst(rx_dvio->vid, rx_dvio->queue_id * VIRTIO_QNUM, cur_pkts, cnt);
    if (likely(tx_pkts)) {
      /* Packets have been sent.*/
      cnt -= tx_pkts;
      /* Prepare for possible retry.*/
      cur_pkts = &cur_pkts[tx_pkts];
    } else {
      /* No packets sent - retry */
      if (++retry > 1000000)
        break;
    }
  } while (cnt);

  nthw_release_rx_packets(rx_dvio->vq, hw_recv_pkts);
  return nb_pkts;
}



#ifdef MEASURE_CPU_UTIL
static volatile uint64_t cycles_tx[MAX_DVIO_INSTANCE_NUM] = {0};
static volatile uint64_t rx_util[MAX_DVIO_INSTANCE_NUM] = {0};
#endif


#ifdef DUMP_PKTS_FROM_VM
#define vq_log_arg(dvio, prefix, format, ...) printf("[%s:%s]: vPort %i: "format, prefix, __func__, dvio->port, __VA_ARGS__)
#else
#define vq_log_arg(dvio, prefix, format, ...)
#endif


#define TX_PACKETS_BULK_SIZE  32

static int stdvio_recv_from_vm(struct tx_dvio *tx_dvio)
{
  struct rte_mbuf *packets[32];
  int nb_rx, i;

  struct nthw_cvirtq_desc cvq_desc;

  struct nthw_memory_descriptor* vq_bufs;
  int pkts_sent = 0;

  if (tx_dvio == NULL) return 0;

#ifdef MEASURE_CPU_UTIL
  uint64_t tsc = rte_rdtsc();
#endif

  nb_rx = rte_vhost_dequeue_burst(tx_dvio->vid, tx_dvio->queue_id * VIRTIO_QNUM + VIRTIO_TXQ, tx_dvio->mb_pool, packets, TX_PACKETS_BULK_SIZE);

  if (likely(nb_rx))
  {
    int nb_segs = 0;
    uint16_t first_vq_descr_idx = 0;
    uint16_t nb_segs_arr[TX_PACKETS_BULK_SIZE];

    vq_log_arg(tx_dvio, "Relay-Core", "vid %i, Tx vq-queue %i, pkts:%i, mbuf size %i, vq size %i\n", tx_dvio->vid, tx_dvio->queue_id * VIRTIO_QNUM + VIRTIO_TXQ,
        nb_rx, packets[0]->buf_len - RTE_PKTMBUF_HEADROOM, SG_HW_TX_PKT_BUFFER_SIZE);

    /*
     * count all segments needed to contain all packets in vq buffers
     */
    for (i = 0; i < nb_rx; i++) {
      /*
       * Build the num segments array for segmentation control and release
       * function
       */
      int vq_segs = NUM_VQ_SEGS(packets[i]->pkt_len);
      nb_segs_arr[i] = vq_segs;
      nb_segs += vq_segs;
    }

    if (!nb_segs) {
      goto _exit;
    }

    int got_nb_segs = nthw_get_tx_buffers(tx_dvio->vq, nb_segs, &first_vq_descr_idx, &cvq_desc, &vq_bufs);
    if (!got_nb_segs) {
      vq_log_arg(tx_dvio, "Relay-Core", "vid %i Zero segments got - back pressure from HW\n", tx_dvio->vid);
      goto _exit;
    }

    int seg_idx = 0;
    uint16_t next_idx;

    for (i = 0; i < nb_rx; i++) {
      uint16_t vq_descr_idx = VIRTQ_DESCR_IDX(seg_idx);

      if (i + 1 < nb_rx) {
        next_idx = VIRTQ_DESCR_IDX(seg_idx + 1);
        rte_prefetch0((void *)(uintptr_t)vq_bufs[next_idx].virtAddr);
      }

      struct _pkt_hdr_tx *hdr_tx = (struct _pkt_hdr_tx *)vq_bufs[vq_descr_idx].virtAddr;
      /* Set the header to all zeros */
      memset(hdr_tx, 0, SG_HDR_SIZE);
      hdr_tx->bypass_port = 0x1FFF;

      /* set packet length */
      hdr_tx->cap_len = packets[i]->pkt_len + SG_HDR_SIZE;

#ifdef DUMP_PKTS_FROM_VM
      printf("pkt size %i recv need %i vq segs, mbuf segs %i\n", packets[i]->pkt_len, nb_segs_arr[i], packets[i]->nb_segs);
#endif

      /* Do we have all segs needed? */
      if ((seg_idx + nb_segs_arr[i]) > got_nb_segs) {
        /*
         * Drop remaining mbuf packets received and send the ones we got vq
         * buffers for
         */
        break;
      }

      if (packets[i]->nb_segs == 1 && nb_segs_arr[i] == 1) {
#ifdef DUMP_PKTS_FROM_VM
        printf("Simple copy %i bytes - mbuf -> vq\n", packets[i]->pkt_len);
#endif
        rte_memcpy((void *)((char *)vq_bufs[vq_descr_idx].virtAddr + SG_HDR_SIZE), rte_pktmbuf_mtod(packets[i], void *), packets[i]->pkt_len);

        cvq_desc.b[vq_descr_idx].len = packets[i]->pkt_len + SG_HDR_SIZE;
        vq_set_next (&cvq_desc, vq_descr_idx, 0);
        vq_set_flags(&cvq_desc, vq_descr_idx, 0);

        seg_idx++;

        if (i + 1 < nb_rx) {
          rte_prefetch0((void *)(uintptr_t)&cvq_desc.b[next_idx]);
        }

      } else {

        int cpy_segs = copy_mbuf_to_virtqueue(&cvq_desc, vq_descr_idx, vq_bufs, got_nb_segs - seg_idx, packets[i]);
        if (cpy_segs < 0) {
          break;
        }
        seg_idx += cpy_segs;


#ifdef DUMP_PKTS_FROM_VM
        int d, tot_size = 0;
        int last_seg_idx = seg_idx - cpy_segs;

        for (d = last_seg_idx; d < seg_idx; d++) {
          tot_size += cvq_desc.b[VIRTQ_DESCR_IDX(d)].len;
        }

        printf("\nDump final Tx vq pkt %i, size %i, tx port %i, bypass id %i, using hw queue index %i\n", i, tot_size, tx_dvio->port,
            hdr_tx->bypass_port, tx_dvio->queue_id);
#if 0
        printf("Header:\n");
        printf("cap_len: %i\n", hdr_tx->cap_len);
        printf("lso_cso0: %i\n", hdr_tx->lso_cso0);
        printf("lso_cso1: %i\n", hdr_tx->lso_cso1);
        printf("lso_cso2: %i\n", hdr_tx->lso_cso2);
        printf("bypass port: %i\n", hdr_tx->bypass_port);
        printf("desc_typ: %i\n", hdr_tx->descr);
        printf("desc_12b: %i\n", hdr_tx->descr_12b);
        printf("color_typ: %i\n", hdr_tx->color_type);
        printf("color: %08x\n", hdr_tx->color);
#endif
        for (d = last_seg_idx; d < seg_idx; d++) {
          char str[32];
          sprintf(str, "Vq seg %i:", d - last_seg_idx);
          dump_packet_seg(str, vq_bufs[VIRTQ_DESCR_IDX(d)].virtAddr, cvq_desc.b[VIRTQ_DESCR_IDX(d)].len);
        }
#endif
      }

      pkts_sent++;
    }

    nthw_release_tx_buffers(tx_dvio->vq, pkts_sent, nb_segs_arr);
  }

_exit:
  for (i = 0; i < nb_rx; i++) {
    rte_pktmbuf_free(packets[i]);
  }

#ifdef MEASURE_CPU_UTIL
    cycles_tx[tx_dvio->vid] += rte_rdtsc() - tsc;
#endif

  return pkts_sent;
}




/*
 * Takes care of 1st queue pair rx/tx called from PMD recv function
 */
void dvio_relay_execute(void *handle)
{
  struct tx_dvio *tx_dvio = (struct tx_dvio *)handle;

  /* First check for any fully offloaded packets to send to VM */
  if (tx_dvio->hdl_rx) {
    stdvio_send_to_vm(tx_dvio->hdl_rx);
  } else {
    printf("ERROR Rx handler\n");
  }

  stdvio_recv_from_vm(tx_dvio);
}






static void *stdvio_queue_relay(void *handle)
{
#define LOOP_TO_SLEEP 1000
  struct tx_dvio *tx_dvio = (struct tx_dvio *)handle;
  int pkts_handled;
  int loop_until_sleep = LOOP_TO_SLEEP;

  NT_LOG(INF, ETHDEV, "(port %i, queue %i) NTNIC: Running Rx/Tx queue relay thread from VP index %i\n", tx_dvio->port, tx_dvio->queue_id, tx_dvio->vid);


  tx_dvio->running = 1;
  while (tx_dvio->running) {

    /* First check for any fully offloaded packets to send to VM */
    if (tx_dvio->hdl_rx) {
      pkts_handled = stdvio_send_to_vm(tx_dvio->hdl_rx);
    } else {
      printf("ERROR Rx handler\n");
      pkts_handled = 0;
    }

    pkts_handled += stdvio_recv_from_vm(tx_dvio);

    if (tx_dvio->core >= 0) {
      /*
       * Just always yield. if we are alone on core we will return immediately
       * again
       */
      if (!pkts_handled) sched_yield();

    } else {
      /*
       * if core is not specifically pinned, then try to dynamically
       * increase polling on high throughput
       */
      if (!pkts_handled) {
        if (loop_until_sleep) {
          --loop_until_sleep;
          sched_yield();
        } else {
          usleep(10);
        }
      } else {
        if (loop_until_sleep != LOOP_TO_SLEEP) loop_until_sleep = LOOP_TO_SLEEP;
      }
    }

  }
  tx_dvio->running = 0;
  NT_LOG(INF, ETHDEV, "(port %i, queue %i) NTNIC: Ending tx thread\n", tx_dvio->port, tx_dvio->queue_id);
  return NULL;
}

static void set_affinity(pthread_t tid, unsigned int core)
{
  rte_cpuset_t cpuset;
  CPU_ZERO(&cpuset);
  CPU_SET(core, &cpuset);
  pthread_setaffinity_np(tid, sizeof(rte_cpuset_t), &cpuset);
}




/*
 * A new virtio-net device is added to a vhost port.
 */
int dvio_relay_new_device(int vid, struct rte_pci_addr addr)
{
  char ifname[IF_NAME_SZ];
  int i, nb_buf;
  NthwDbs_t *pNthwDbs = get_pdbs_from_pci(addr);

  NT_LOG(INF, ETHDEV, "DVIO: Adding new device  vid = %i\n", vid);

  rte_vhost_get_ifname(vid, ifname, sizeof(ifname));
  int vhid = get_dvio_conn_vhid(ifname);
  if (vhid < 0 || vid >= MAX_DVIO_INSTANCE_NUM) {
    NT_LOG(ERR, ETHDEV, "Did not find vhid index\n");
    return -1;
  }

  if (dvio_conns[vhid].rxqs != dvio_conns[vhid].txqs) {
    NT_LOG(ERR, ETHDEV, "Rx queues and Tx queues must be equal\n");
    return -1;
  }

  NT_LOG(INF, ETHDEV, "**** VID %i rxq %i txq %i ****\n", vid, dvio_conns[vhid].rxqs, dvio_conns[vhid].txqs);

  /* Create a structure for each queue. we only support 1 queue by now */
  for (i = 0; i < dvio_conns[vhid].rxqs; i++) {

    memset(&rx_dvios[vid][i], 0, sizeof(rx_dvios[0][0]));
    memset(&tx_dvios[vid][i], 0, sizeof(tx_dvios[0][0]));

    rx_dvios[vid][i].vq = nthw_setup_managed_rx_virt_queue(pNthwDbs,
                          dvio_conns[vhid].first_queue_hw_id + i,          /* index */
                          dvio_conns[vhid].num_descr,                      /* queue size */
                          dvio_conns[vhid].queue[i].hwq.vf_num,            /* hostid */
                          1,                                               /* header NT DVIO header */
                          &dvio_conns[vhid].queue[i].hwq.virt_queues_ctrl, /* HW virtqueue ctrl */
                          dvio_conns[vhid].queue[i].hwq.pkt_buffers,       /* HW virtqueue IOMMU mapped buffers */
                          SPLIT_RING,                                     /* vq type */
                          -1);                                             /* interrupt vector */

    rx_dvios[vid][i].vf_num   = dvio_conns[vhid].queue[i].hwq.vf_num;
    rx_dvios[vid][i].vid      = vid;
    rx_dvios[vid][i].queue_id = i;
    rx_dvios[vid][i].port     = dvio_conns[vhid].hw_port_id;

    register_release_virtqueue_info(rx_dvios[vid][i].vq, 1, 1);

    /* Build mbuf pointer array */
    for (nb_buf = 0; nb_buf < NUM_MBUFS; nb_buf++) {
      rx_dvios[vid][i].bufs[nb_buf] = &rx_dvios[vid][i].base_bufs[nb_buf];
      rx_dvios[vid][i].seg_bufs[nb_buf] = &rx_dvios[vid][i].base_bufs[nb_buf + NUM_MBUFS];
    }

    /* rx is called from tx execute */
    tx_dvios[vid][i].hdl_rx = &rx_dvios[vid][i];

    tx_dvios[vid][i].vq = nthw_setup_managed_tx_virt_queue(pNthwDbs,
                          dvio_conns[vhid].first_queue_hw_id + i,             /* same index as Rx */
                          dvio_conns[vhid].num_descr_tx,                      /* queue size */
                          dvio_conns[vhid].queue[i].hwq_tx.vf_num,            /* hostId */
                          0,                                                  /* transmit port - not used in vswitch enabled mode */
                          dvio_conns[vhid].hw_port_id,                        /* in_port equals vf_num for now */
                          1,                                                  /* header NT DVIO header */
                          &dvio_conns[vhid].queue[i].hwq_tx.virt_queues_ctrl, /* HW virtqueue ctrl */
                          dvio_conns[vhid].queue[i].hwq_tx.pkt_buffers,       /* HW virtqueue IOMMU mapped buffers */
                          SPLIT_RING,
                          -1,
                          IN_ORDER);

    tx_dvios[vid][i].vf_num       = dvio_conns[vhid].queue[i].hwq_tx.vf_num;
    tx_dvios[vid][i].num_descr_tx = dvio_conns[vhid].num_descr_tx;
    tx_dvios[vid][i].vid          = vid;
    tx_dvios[vid][i].queue_id     = i;
    tx_dvios[vid][i].port         = dvio_conns[vhid].hw_port_id;
    tx_dvios[vid][i].mb_pool      = dvio_conns[vhid].queue[i].mb_pool;
    tx_dvios[vid][i].running      = 0;
    tx_dvios[vid][i].core         = dvio_conns[vhid].queue[i].core;

    register_release_virtqueue_info(tx_dvios[vid][i].vq, 0, 1);

    if (i == 0) {
      /* 1st queue pair is handled from OVS polling rep thread */
      dvio_conns[vhid].queue[0].execute_handle = (void *)&tx_dvios[vid][0];
    } else {
      /*
       * Remaining queues in a RSS multiqueue setup needs own threads - may be
       * pinned on same core or a pool of cores
       */
      pthread_create(&dvio_conns[vhid].queue[i].tids, NULL, stdvio_queue_relay, &tx_dvios[vid][i]);
      if (tx_dvios[vid][i].core >= 0) {
        set_affinity(dvio_conns[vhid].queue[i].tids, tx_dvios[vid][i].core);
      }
    }

  }

  return 0;
}

/*
 * return vhid
 */
int dvio_relay_destroy_device(int vid)
{
  int i;
  char ifname[IF_NAME_SZ];
  NT_LOG(INF, ETHDEV, "DVIO: Destroy device\n");

  rte_vhost_get_ifname(vid, ifname, sizeof(ifname));
  int vhid = get_dvio_conn_vhid(ifname);
  if (vhid < 0 || vid >= MAX_DVIO_INSTANCE_NUM) {
    NT_LOG(ERR, ETHDEV, "Did not find vhid index - error destroying device!\n");
    return -1;
  }

  /* Close down Tx queues */
  for (i = 0; i < dvio_conns[vhid].txqs; i++) {
    dvio_conns[vhid].queue[i].execute_handle = NULL;

    if (tx_dvios[vid][i].running) {
      tx_dvios[vid][i].running = 0;
      pthread_join(dvio_conns[vhid].queue[i].tids, NULL);
    }

  }

  /* Close down Rx queues */
  for (i = 0; i < dvio_conns[vhid].rxqs; i++) {
    if (rx_dvios[vid][i].running) {
      rx_dvios[vid][i].running = 0;
    }
  }

  rte_mb();
  usleep(100000);

  for (i = 0; i < dvio_conns[vhid].txqs; i++) {
    if (tx_dvios[vid][i].vq) {
      de_register_release_virtqueue_info(tx_dvios[vid][i].vq);
      nthw_release_managed_tx_virt_queue(tx_dvios[vid][i].vq);
    }
  }

  for (i = 0; i < dvio_conns[vhid].rxqs; i++) {
    if (rx_dvios[vid][i].vq) {
      de_register_release_virtqueue_info(rx_dvios[vid][i].vq);
      nthw_release_managed_rx_virt_queue(rx_dvios[vid][i].vq);
    }
  }

  return vhid;
}


/* NOTE: txq from DPDK is rx-handling for virtio guest */
int dvio_relay_configure_txq(int vhid, int queue_id,
    struct hwq_s *hwq, int num_descr)
{
  /* Check if valid and created */
  if (vhid >= 0 && vhid < MAX_DVIO_CONNECTIONS &&
      queue_id >= 0 && queue_id < DVIO_MAX_QUEUES) {
    NT_LOG(INF, ETHDEV, "stdvio: configure Tx(egress from HW) for virtio id:%i queue %i\n", vhid, queue_id);

    dvio_conns[vhid].queue[queue_id].hwq_tx  = *hwq;
    dvio_conns[vhid].num_descr_tx = num_descr;
    return 0;
  }
  return -1;
}

/* NOTE: rxq from DPDK is tx-handling for virtio guest */
int dvio_relay_configure_rxq(int vhid, int queue_id,
    struct hwq_s *hwq, int num_descr, int core_id, handle_t **handle)
{
  char name[64];
  /* Check if valid and created */
  if (vhid >= 0 && vhid < MAX_DVIO_CONNECTIONS &&
      queue_id >= 0 && queue_id < DVIO_MAX_QUEUES &&
      handle) {

    dvio_conns[vhid].queue[queue_id].core = core_id;

    /* Need a pool for each queue */
    snprintf(name, 64, "MBUF_POOL_%i_%i", vhid, queue_id);

    dvio_conns[vhid].queue[queue_id].mb_pool = rte_pktmbuf_pool_create(name, 1024,
            RTE_CACHE_LINE_SIZE, 0, MBUF_MEM_POOL_SIZE + RTE_PKTMBUF_HEADROOM, rte_socket_id());

    if (!dvio_conns[vhid].queue[queue_id].mb_pool) {
      NT_LOG(ERR, ETHDEV, "Could not create memory pool for virtqueue relay core:  virtio id:%i queue %i\n", vhid, queue_id);
      return -1;
    }

    NT_LOG(INF, ETHDEV, "stdvio: configure Rx(ingress to HW) for virtio id:%i queue %i, socket id %i, memory pool %p\n", vhid, queue_id, rte_socket_id(),
          dvio_conns[vhid].queue[queue_id].mb_pool);

    dvio_conns[vhid].queue[queue_id].hwq  = *hwq;
    dvio_conns[vhid].num_descr = num_descr;

    dvio_conns[vhid].queue[queue_id].execute_handle = NULL;
    *handle = &dvio_conns[vhid].queue[queue_id].execute_handle;
    return 0;
  }
  return -1;
}


int dvio_relay_create(int vhid, int first_queue_hw_id, int hw_port_id,
    int rxqs, int txqs,
    const char *socket_path)
{
  int i;
  NT_LOG(INF, ETHDEV, "VDPA Relay: Create device for %s\n", socket_path);

  /* Create entry in dvio_conn with index vhid */
  strlcpy(dvio_conns[vhid].socket, socket_path, IF_NAME_SZ);

  dvio_conns[vhid].hw_port_id        = hw_port_id;
  dvio_conns[vhid].first_queue_hw_id = first_queue_hw_id; /* then contiguous */

  dvio_conns[vhid].rxqs = txqs;  /* from HW to VM */
  dvio_conns[vhid].txqs = rxqs;  /* from VM to HW */

  for (i = 0; i < DVIO_MAX_QUEUES; i++) {
    dvio_conns[vhid].queue[i].mb_pool = NULL;
  }

  return 0;
}
