/*-
 *   BSD LICENSE
 *
 *   Copyright(c) 2010-2015 Intel Corporation. All rights reserved.
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <unistd.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <linux/limits.h>
#include <getopt.h>
#include <termios.h>
#include <time.h>

#include <rte_eal.h>
#include <rte_ethdev.h>
#include <rte_cycles.h>
#include <rte_lcore.h>
#include <rte_mbuf.h>
#include <rte_flow.h>
#include <rte_flow_driver.h>
#include <rte_net.h>
#include <rte_mtr.h>
#include <rte_mtr_driver.h>

#include <rte_malloc.h>

#include "filter.h"


#define RX_RING_SIZE 1024
#define TX_RING_SIZE 1024

#define NUM_MBUFS 8192
#define MBUF_CACHE_SIZE 250
#define BURST_SIZE 64

#define MAX_PKT_SIZE 2000
#define MAX_NO_PORTS 8

pthread_mutex_t mutex;

static void unlearn(void);

static const struct rte_eth_conf port_conf_default = {
    .rxmode = { .max_lro_pkt_size = RTE_ETHER_MAX_LEN }
};

struct stat_s statistic[NO_PORTS][MAX_QUEUES];
struct flows_stat_s *flows_stat;

static uint32_t number_of_queues = 1;
static bool set_filter = true;
static bool use_x_stat = false;
static int disable_hw_offload = false;

bool use_meter = false;
bool use_decap = false;
bool debug_print_out = false;
uint32_t ingress_port = 0;
uint32_t eegress_port = 1;
uint32_t ingress_vlan = 44;
uint32_t eegress_vlan = 45;

static uint64_t mtr_cir = 100ULL*1000ULL*1000ULL*1000ULL;

enum {
  USR_SIG_NONE,
  USR_SIG_UNLEARN=SIGUSR1,
  USR_SIG_LEARN=SIGUSR2,
  USR_SIG_QUIT=SIGTERM,
  USR_SIG_DESTROY_DEFAULT=40,
  USR_SIG_SETUP_DEFAULT=41,
};

uint8_t rte_eal_init_done = 0;
uint8_t quit_signal = 0;
uint32_t usr_signal = USR_SIG_NONE;

struct rte_flow *def_flows[NO_PORTS][10] = {NULL};
uint64_t dev_flow_count[NO_PORTS] = {0};

int meter_profile_id[NO_PORTS];
int meter_policy_id[NO_PORTS];
int meter_mtr_id[NO_PORTS];

struct port_t {
  uint8_t rxPort;
  uint8_t txPort;
  uint32_t queue;
  struct stat_s *stat;
  struct flows_s *flows;
};

static void int_handler(int sig_num)
{
  switch (sig_num)
  {
    case USR_SIG_UNLEARN:
    case USR_SIG_LEARN:
    case USR_SIG_DESTROY_DEFAULT:
    case USR_SIG_SETUP_DEFAULT:
      printf("\nSignal %d received\n", sig_num);
      usr_signal=sig_num;
      break;
    case USR_SIG_QUIT:
    default:
      printf("\nExiting on signal %d\n", sig_num);
      /* Until rte_eal_init() is complete and waiting for clients to connect,
       * the signal SIGTERM will not terminate the application.
       * That's why we have to terminate it this way. */
      if (rte_eal_init_done == 0) {
        exit(1);
      }
      /* set quit flag for rx thread to exit */
      quit_signal = 1;
      break;
  }
}


static int SetupFilter(uint8_t port, uint8_t numQueues)
{
  struct rte_flow_error error;
  struct rte_flow_attr attr;
  struct rte_flow_item pattern[2];
  struct rte_flow_action actions[4];
  int i;

  struct rte_ether_addr addr;
  rte_eth_macaddr_get(port, &addr);

  // Action struct
  struct rte_flow_action_queue queue;
  struct rte_flow_action_rss rss;
  uint16_t queues[MAX_QUEUES];

  if (numQueues > MAX_QUEUES) {
    printf("ERROR: To many queues needed. Reduce the number of queues\n");
    return -1;
  }

  /* Poisoning to make sure PMDs update it in case of error. */
  memset(&error, 0x22, sizeof(struct rte_flow_error));

  memset(&attr, 0, sizeof(attr));
  attr.ingress = 1;
  attr.priority = 10;
  attr.group = 0;

  memset(&actions, 0, sizeof(actions));
  memset(&pattern, 0, sizeof(pattern));

  uint32_t actionCount = 0;
  uint32_t patternCount = 0;

  // Set IPV4 default filter
  memset(&actions, 0, sizeof(actions));
  memset(&pattern, 0, sizeof(pattern));

  actionCount = 0;
  patternCount = 0;

  pattern[patternCount].type = RTE_FLOW_ITEM_TYPE_IPV4;
  patternCount++;

  pattern[patternCount].type = RTE_FLOW_ITEM_TYPE_END;
  patternCount++;

  struct rte_flow_action_jump jump;
  jump.group = 1;
  actions[actionCount].type = RTE_FLOW_ACTION_TYPE_JUMP;
  actions[actionCount].conf = &jump;
  actionCount++;

  actions[actionCount].type = RTE_FLOW_ACTION_TYPE_END;
  actionCount++;

  def_flows[port][dev_flow_count[port]] = rte_flow_create(port, &attr, pattern, actions, &error);
  if (def_flows[port][dev_flow_count[port]] == NULL) {
    printf("ERROR (%s): Port: %u, Type: RTE_FLOW_ITEM_TYPE_IPV4, MSG: %s\n", __func__, port, error.message);
    return -1;
  }
  dev_flow_count[port]++;

  // Set IPV6 default filter
  memset(&actions, 0, sizeof(actions));
  memset(&pattern, 0, sizeof(pattern));

  actionCount = 0;
  patternCount = 0;

  pattern[patternCount].type = RTE_FLOW_ITEM_TYPE_IPV6;
  patternCount++;

  pattern[patternCount].type = RTE_FLOW_ITEM_TYPE_END;
  patternCount++;

  jump.group = 2;
  actions[actionCount].type = RTE_FLOW_ACTION_TYPE_JUMP;
  actions[actionCount].conf = &jump;
  actionCount++;

  actions[actionCount].type = RTE_FLOW_ACTION_TYPE_END;
  actionCount++;

  def_flows[port][dev_flow_count[port]] = rte_flow_create(port, &attr, pattern, actions, &error);
  if (def_flows[port][dev_flow_count[port]] == NULL) {
    printf("ERROR (%s): Port: %u, Type: RTE_FLOW_ITEM_TYPE_IPV6, MSG: %s\n", __func__, port, error.message);
    return -1;
  }
  dev_flow_count[port]++;

  // Set default filter queues
  for (int g = 0; g < 2; g++) {
    memset(&actions, 0, sizeof(actions));
    memset(&pattern, 0, sizeof(pattern));

    attr.group = g + 1;

    actionCount = 0;
    patternCount = 0;

    pattern[patternCount].type = RTE_FLOW_ITEM_TYPE_END;
    patternCount++;

    if (numQueues > 1) {
      rss.func = RTE_ETH_HASH_FUNCTION_DEFAULT;
      rss.level = 0;
      rss.types  = RTE_ETH_RSS_UDP | RTE_ETH_RSS_TCP | RTE_ETH_RSS_SCTP;
      rss.queue_num = numQueues;
      for (i = 0; i < numQueues; i++) {
        queues[i] = i;
      }
      rss.queue = queues;
      actions[actionCount].type = RTE_FLOW_ACTION_TYPE_RSS;
      actions[actionCount].conf = &rss;
      actionCount++;
    }
    else {
      queue.index = 0;
      actions[actionCount].type = RTE_FLOW_ACTION_TYPE_QUEUE;
      actions[actionCount].conf = &queue;
      actionCount++;
    }

    actions[actionCount].type = RTE_FLOW_ACTION_TYPE_END;
    actionCount++;

    def_flows[port][dev_flow_count[port]] = rte_flow_create(port, &attr, pattern, actions, &error);
    if (def_flows[port][dev_flow_count[port]] == NULL) {
      printf("ERROR (%s): Port: %u, Group: %u, Type: RTE_FLOW_ACTION_TYPE_..., MSG: %s\n", __func__, port, g, error.message);
      return -1;
    }
    dev_flow_count[port]++;
  }

  return 0;
}

/*
 * Initializes a given port using global settings and with the RX buffers
 * coming from the mbuf_pool passed as a parameter.
 */
static inline int port_init(uint8_t port, struct rte_mempool *mbuf_pool)
{
    struct rte_eth_conf port_conf = port_conf_default;
    int retval;
    uint32_t q;

    if (port >= rte_eth_dev_count_avail())
        return -1;

    /* Configure the Ethernet device. */
    retval = rte_eth_dev_configure(port, number_of_queues, number_of_queues, &port_conf);
    if (retval != 0)
        return retval;

    /* Allocate and set up 1 RX queue per Ethernet port. */
    for (q = 0; q < number_of_queues; q++) {
        retval = rte_eth_rx_queue_setup(port, q, RX_RING_SIZE, rte_eth_dev_socket_id(port), NULL, mbuf_pool);
        if (retval < 0)
            return retval;
    }

    /* Allocate and set up 1 TX queue per Ethernet port. */
    for (q = 0; q < number_of_queues; q++) {
        retval = rte_eth_tx_queue_setup(port, q, TX_RING_SIZE, rte_eth_dev_socket_id(port), NULL);
        if (retval < 0)
            return retval;
    }

  struct rte_ether_addr addr;
  rte_eth_macaddr_get(port, &addr);
  printf("Port %u MAC: %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 " %02" PRIx8 "\n", (unsigned int)port,
                                                                                                          addr.addr_bytes[0], addr.addr_bytes[1],
                                                                                                          addr.addr_bytes[2], addr.addr_bytes[3],
                                                                                                          addr.addr_bytes[4], addr.addr_bytes[5]);

  set_eth_mac_address(port, &addr);

    /* Start the Ethernet port. */
    retval = rte_eth_dev_start(port);
    if (retval < 0)
        return retval;

    /* Disable RX in promiscuous mode for the Ethernet device. */
    rte_eth_promiscuous_disable(port);

    return 0;
}

static inline void TxPackets(struct rte_mbuf **bufs, uint32_t nb_rx, uint8_t port, uint16_t queue)
{
  uint32_t nbidx = 0;
  if (nb_rx) {
    do {
      uint16_t nb_tx = rte_eth_tx_burst(port, queue, &bufs[nbidx], nb_rx-nbidx);
      nbidx += nb_tx;
    } while (nbidx < nb_rx);
  }
}

static int lcore_rx(void *p)
{
  struct port_t *port_io = (struct port_t *)p;

  uint8_t rxPort = port_io->rxPort;
  uint8_t txPort = port_io->txPort;
  uint16_t queue = port_io->queue;
  struct stat_s *stat = port_io->stat;
  struct flows_s *flows = port_io->flows;

  printf("Core %u RX port %u -> TX port %u using queue %u\n", rte_lcore_id(), rxPort, txPort, queue);
  fflush(stdout);

  /* Run until the application is quit or killed. */
  for (;;) {
    if (quit_signal != 0) {
      printf("Breaking lcore %u\n", rte_lcore_id());
      break;
    }

    /* Get burst of RX packets */
    struct rte_mbuf *bufs[BURST_SIZE];
    const uint16_t nb_rx = rte_eth_rx_burst(rxPort, queue, bufs, BURST_SIZE);
    if (likely(nb_rx)) {
      uint16_t buf;
      for (buf = 0; buf < nb_rx; buf++) {
        struct rte_mbuf *mbuf = bufs[buf];
        if (!disable_hw_offload) {
          offload_filter_setup(mbuf, rxPort, txPort, stat, flows);
          if (use_decap) {
            rte_pktmbuf_free(mbuf);
          }
        }
        stat->numRxPackets++;
      }
      if (!use_decap) {
        TxPackets(bufs, nb_rx, txPort, queue);
      }
    }
  }
  fflush(stdout);
  return 0;
}

static void nonblock(int state)
{
    struct termios ttystate;

    //get the terminal state
    tcgetattr(STDIN_FILENO, &ttystate);

    if (state==1)
    {
        //turn off canonical mode
        ttystate.c_lflag &= ~ICANON;
        //minimum of number input read.
        ttystate.c_cc[VMIN] = 1;
    }
    else if (state==0)
    {
        //turn on canonical mode
        ttystate.c_lflag |= ICANON;
    }
    //set the terminal attributes.
    tcsetattr(STDIN_FILENO, TCSANOW, &ttystate);
}

static int delete_default_flows(void)
{
  struct rte_flow_error error;
  uint64_t i;

  printf("Port %u: Destroying %llu default filters: ", eegress_port, (long long unsigned int)dev_flow_count[eegress_port]);
  for (i = 0; i < dev_flow_count[eegress_port]; i++) {
    if (def_flows[eegress_port][i]) {
      printf("%p ", def_flows[eegress_port][i]);
      if (rte_flow_destroy(eegress_port, def_flows[eegress_port][i], &error) != 0) {
        printf("ERROR deleting default flow %lu, port %u, MSG: %s\n", i, eegress_port, error.message);
      }
      def_flows[eegress_port][i] = NULL;
    }
  }
  printf("\n");
  dev_flow_count[eegress_port] = 0;
  printf("Port %u: Destroying %llu default filters: ", ingress_port, (long long unsigned int)dev_flow_count[ingress_port]);
  for (i = 0; i < dev_flow_count[ingress_port]; i++) {
    printf("%p ", def_flows[ingress_port][i]);
    if (def_flows[ingress_port][i]) {
      if (rte_flow_destroy(ingress_port, def_flows[ingress_port][i], &error) != 0) {
        printf("ERROR deleting default flow %lu, port %u, MSG: %s\n", i, ingress_port, error.message);
      }
      def_flows[ingress_port][i] = NULL;
    }
  }
  printf("\n");
  dev_flow_count[ingress_port] = 0;
  return 0;
}

static int setup_default_flows(void)
{
  if (SetupFilter(ingress_port, number_of_queues) != 0) {
    printf("Cannot default setup flows for port %u\n", ingress_port);
    return 1;
  }
  if (SetupFilter(eegress_port, number_of_queues) != 0) {
    printf("Cannot default setup flows for port %u\n", eegress_port);
    return 1;
  }
  return 0;
}

static int learn_unlearn(void *p)
{
  (void)p;
  printf("Core %u acting as worker core for learn/unlearn\n", rte_lcore_id()); fflush(stdout);
  nonblock(1);
  for (;;) {
    fd_set fd_list;
    int32_t result;
    struct timeval tv;

    if (quit_signal != 0) return 0;

    FD_ZERO(&fd_list);
    FD_SET(fileno(stdin), &fd_list);

    tv.tv_sec = 1;
    tv.tv_usec = 0;
    result = select(fileno(stdin) + 1, &fd_list, NULL, NULL, &tv);

    switch (usr_signal)
    {
      case USR_SIG_UNLEARN:
        printf("\n\nUnlearn flows\n\n");
        disable_hw_offload = true;
        unlearn();
        usr_signal = USR_SIG_NONE;
        break;
      case USR_SIG_LEARN:
        printf("\n\nLearn flows\n\n");
        disable_hw_offload = false;
        usr_signal = USR_SIG_NONE;
        break;
      case USR_SIG_DESTROY_DEFAULT:
        printf("\n\nDelete default flows\n\n");
        delete_default_flows();
        usr_signal = USR_SIG_NONE;
        break;
      case USR_SIG_SETUP_DEFAULT:
        printf("\n\nSetup default flows\n\n");
        setup_default_flows();
        usr_signal = USR_SIG_NONE;
        break;
    }

    if (result < 0) {
      fprintf(stderr, ">>> Error: Error on select: %s\n", strerror(errno));
    }

    if (FD_ISSET(0, &fd_list))
    {
      const uint32_t key = getchar();
      switch (key) {
      case 113:
      case 81:
        printf("\n\nQuit is pressed\n\n");
        quit_signal = 1;
        break;
      case 117:
      case 85:
        printf("\n\nUnlearn flows\n\n");
        disable_hw_offload = true;
        unlearn();
        break;
      case 108:
      case 76:
        printf("\n\nLearn flows\n\n");
        disable_hw_offload = false;
        break;
      case 0x31:
        printf("\n\nDelete default flows\n\n");
        delete_default_flows();
        usr_signal = 0;
        break;
      case 0x32:
        printf("\n\nSetup default flows\n\n");
        setup_default_flows();
        usr_signal = 0;
        break;
      }
    }
   }
  nonblock(0);
  return 0;
}

/*
 * The lcore main. This is the main thread that does the work, reading from
 * an input port and writing to an output port.
 */
static void lcore_main1(void)
{
  printf("Core %u acting as worker core for statistics output\n\n", rte_lcore_id()); fflush(stdout);

  printf("Press U or send SIGNAL %d to unlearn flows\n", USR_SIG_UNLEARN);
  printf("Press L or send SIGNAL %d to learn flows\n", USR_SIG_LEARN);
  printf("Press 1 or send SIGNAL %d to delete default flows\n", USR_SIG_DESTROY_DEFAULT);
  printf("Press 2 or send SIGNAL %d to setup default flows\n", USR_SIG_SETUP_DEFAULT);
  printf("Press Q or send SIGNAL %d to quit\n", USR_SIG_QUIT);
  printf("Process ID: %d\n\n", getpid());
  printf("Usage example:\nkill -%d %d\n\n", USR_SIG_UNLEARN, getpid());
  fflush(stdout);

  for (;;) {
    if (quit_signal != 0) break;
    sleep(1);

    if (!debug_print_out) {
      uint i;
      printf("Q: %u ", number_of_queues);
      for (i = 0; i < number_of_queues; i++) {
        printf("R%u.%u: %llu ", ingress_port, i, (long long unsigned int)statistic[ingress_port][i].numRxPackets);
        printf("F: %llu ", (long long unsigned int)statistic[eegress_port][i].numFilter);
      }
      for (i = 0; i < number_of_queues; i++) {
        printf("R%u.%u: %llu ", eegress_port, i, (long long unsigned int)statistic[eegress_port][i].numRxPackets);
        printf("F: %llu ", (long long unsigned int)statistic[eegress_port][i].numFilter);
      }


      if (use_meter) {
        struct rte_mtr_stats stats;
        struct rte_mtr_error error;
        uint64_t stats_mask;
        uint64_t meter = 0;
        int port;
        for (port = 0; port < NO_PORTS; port++) {
          if (!rte_mtr_stats_read(port, meter_mtr_id[port], &stats, &stats_mask, 0, &error)) {
            if (stats_mask & RTE_MTR_STATS_N_PKTS_GREEN) {
              meter += (long long unsigned int)stats.n_pkts[RTE_COLOR_GREEN];
            }
          }
        }
        printf("M: %llu ", (long long unsigned int)meter);
      }
      printf("                \r"); fflush(stdout);
    }
  }
}

/*
 * The lcore worker This is the worker thread that collects and prints statistics
 */
static void lcore_main2(void)
{
  uint8_t port;
  uint8_t counter = 0;

  struct rte_eth_xstat *xstats[RTE_MAX_ETHPORTS];
  int cnt_xstats[RTE_MAX_ETHPORTS];
  struct rte_eth_xstat_name *xstats_names[RTE_MAX_ETHPORTS];

  char name[100];
  struct rte_eth_link eth_link;
  uint16_t mtu;
  struct rte_ether_addr addr;
  int lf;
  uint32_t i;

  for (port = 0; port < RTE_MAX_ETHPORTS; port++) {
    xstats[port] = NULL;
    cnt_xstats[port] = 0;
    xstats_names[port] = NULL;
  }

  for (port = 0; port < NO_PORTS; port++) {
    /* Get count */
    cnt_xstats[port] = rte_eth_xstats_get_names(port, NULL, 0);
    if (cnt_xstats[port]  < 0) {
      printf("Error: Cannot get count of xstats on port %u\n", port);
    }
    else {
      xstats_names[port] = malloc(sizeof(struct rte_eth_xstat_name) * cnt_xstats[port]);
      if (xstats_names[port] == NULL) {
        printf("Cannot allocate memory for xstats lookup\n");
      }
      else {
        if (cnt_xstats[port] != rte_eth_xstats_get_names(port, xstats_names[port], cnt_xstats[port])) {
          printf("Error: Cannot get xstats lookup\n");
          free(xstats_names[port]);
          xstats_names[port] = NULL;
        }
      }
    }

    if (xstats_names[port]) {
      xstats[port] = malloc(sizeof(struct rte_eth_xstat) * cnt_xstats[port]);
      if (xstats[port] == NULL) {
        printf("Cannot allocate memory for xstats\n");
        free(xstats_names[port]);
        xstats_names[port] = NULL;
      }
    }
  }

  printf("\033[2J");
  while (1) {
    if (quit_signal != 0) return;
    if (counter++ >= 10) {
      counter = 0;
      printf("\033[2J");
    }
    sleep(1);
    printf("\033[0;0H\n");
    printf("Port status\n");
    printf("----------------------------------------------------------------------------------------\n");
    for (port = 0; port < NO_PORTS; port++) {
      rte_eth_link_get(port, &eth_link);
      printf("P%u: link_status: %-4s - ", port, eth_link.link_status == RTE_ETH_LINK_UP ? "UP" : "DOWN");
      rte_eth_dev_get_name_by_port(port, name);
      printf("Device name: \"%s\" - MTU: ", name);
      if (rte_eth_dev_get_mtu(port, &mtu) == 0) {
        printf("%u - ", mtu);
      }
      else {
        printf("None - ");
      }

      rte_eth_macaddr_get(port, &addr);
      printf("MAC: %02X:%02X:%02X:%02X:%02X:%02X", addr.addr_bytes[0], addr.addr_bytes[1],
                                                   addr.addr_bytes[2], addr.addr_bytes[3],
                                                   addr.addr_bytes[4], addr.addr_bytes[5]);
      printf("\n");
    }
    printf("\n");

    printf("SW statistics\n");
    printf("-------------------------------------\n");
    for (port = 0; port < NO_PORTS; port++) {
      printf("P%u: ", port);
      for (i = 0; i < number_of_queues; i++) {
        printf("Q%u: %10llu ", i, (long long unsigned int)statistic[port][i].numRxPackets);
        printf("F: %10llu ", (long long unsigned int)statistic[port][i].numFilter);
      }
      printf("\n");
    }

    for (port = 0; port < NO_PORTS; port++) {
      int cnt_xstats_port = cnt_xstats[port];
      lf = 0;
      printf("\nVQ statistics port %u\n", port);
      printf("------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
      if (xstats[port]) {
        struct rte_eth_xstat *xstats_port = xstats[port];
        struct rte_eth_xstat_name *xstats_names_port = xstats_names[port];

        if (rte_eth_xstats_get(port, xstats_port, cnt_xstats_port)) {
          for (int i = 0; i < cnt_xstats_port; i++) {
            printf("%-30s: %10llu  ", xstats_names_port[xstats_port[i].id].name, (long long unsigned int)xstats_port[i].value);
            if (!(++lf % 4) && (lf > 0)) {
              printf("\n");
            }
          }
        }
      }
      printf("\n");
    }

    if (use_meter) {
      struct rte_mtr_stats stats;
      struct rte_mtr_error error;
      uint64_t stats_mask;
      for (port = 0; port < NO_PORTS; port++) {
        printf("\nMeter statistics port %u\n", port);
        printf("------------------------------------------------------------------------------------------------\n");
        if (!rte_mtr_stats_read(port, meter_mtr_id[port], &stats, &stats_mask, 0, &error)) {
          printf("PR: %5u, PO: %5u, ME: %5u, Mask: %016lX ", meter_profile_id[port], meter_mtr_id[port], meter_policy_id[port], stats_mask);
          if (stats_mask & RTE_MTR_STATS_N_PKTS_GREEN) {
            printf("G: %10llu packets", (long long unsigned int)stats.n_pkts[RTE_COLOR_GREEN]);
            if (stats_mask & RTE_MTR_STATS_N_BYTES_GREEN) {
              printf(", %10llu octets", (long long unsigned int)stats.n_bytes[RTE_COLOR_GREEN]);
            }
          }
          if (stats_mask & RTE_MTR_STATS_N_PKTS_YELLOW) {
            printf(", Y: %10llu packets", (long long unsigned int)stats.n_pkts[RTE_COLOR_YELLOW]);
            if (stats_mask & RTE_MTR_STATS_N_BYTES_YELLOW) {
              printf(", %10llu octets", (long long unsigned int)stats.n_bytes[RTE_COLOR_YELLOW]);
            }
          }
          if (stats_mask & RTE_MTR_STATS_N_PKTS_RED) {
            printf(", R: %10llu packets", (long long unsigned int)stats.n_pkts[RTE_COLOR_RED]);
            if (stats_mask & RTE_MTR_STATS_N_BYTES_RED) {
              printf(", %10llu octets", (long long unsigned int)stats.n_bytes[RTE_COLOR_RED]);
            }
          }
          if (stats_mask & RTE_MTR_STATS_N_PKTS_DROPPED) {
            printf(", D: %10llu packets", (long long unsigned int)stats.n_pkts_dropped);
            if (stats_mask & RTE_MTR_STATS_N_BYTES_DROPPED) {
              printf(", %10llu octets", (long long unsigned int)stats.n_bytes_dropped);
            }
          }
          printf("\n");
        }
        else {
          printf("Error: %s\n", error.message);
        }
      }
    }
    printf("\n");
    fflush(stdout);
  }
  printf("\n");
}

static int setup_meter(uint8_t port)
{
  struct rte_mtr_error error;
  struct rte_mtr_capabilities cap;

  srand(time(NULL) + port * 10);
  meter_profile_id[port] = rand() % 10000;
  meter_policy_id[port] = rand() % 10000;
  meter_mtr_id[port] = rand() % 10000;

  if (rte_mtr_capabilities_get(port, &cap, &error) != 0) {
    printf("Failed to setup meter for port %u: %s\n", port, error.message);
    return -1;
  }
  // printf("ERROR: Type %u, MSG %s\n", error.type, error.message);
  // printf("%u: identical                              %d\n", port, cap.identical);
  // printf("%u: shared_identical                       %d\n", port, cap.shared_identical);
  // printf("%u: chaining_use_prev_mtr_color_supported  %d\n", port, cap.chaining_use_prev_mtr_color_supported);
  // printf("%u: chaining_use_prev_mtr_color_enforced   %d\n", port, cap.chaining_use_prev_mtr_color_enforced);
  // printf("%u: color_aware_srtcm_rfc2697_supported    %d\n", port, cap.color_aware_srtcm_rfc2697_supported);
  // printf("%u: color_aware_trtcm_rfc2698_supported    %d\n", port, cap.color_aware_trtcm_rfc2698_supported);
  // printf("%u: color_aware_trtcm_rfc4115_supported    %d\n", port, cap.color_aware_trtcm_rfc4115_supported);
  // printf("%u: srtcm_rfc2697_byte_mode_supported      %d\n", port, cap.srtcm_rfc2697_byte_mode_supported);
  // printf("%u: srtcm_rfc2697_packet_mode_supported    %d\n", port, cap.srtcm_rfc2697_packet_mode_supported);
  // printf("%u: trtcm_rfc2698_byte_mode_supported      %d\n", port, cap.trtcm_rfc2698_byte_mode_supported);
  // printf("%u: trtcm_rfc2698_packet_mode_supported    %d\n", port, cap.trtcm_rfc2698_packet_mode_supported);
  // printf("%u: trtcm_rfc4115_byte_mode_supported      %d\n", port, cap.trtcm_rfc4115_byte_mode_supported);
  // printf("%u: trtcm_rfc4115_packet_mode_supported    %d\n", port, cap.trtcm_rfc4115_packet_mode_supported);
  // printf("%u: separate_input_color_table_per_port    %d\n", port, cap.separate_input_color_table_per_port);
  // printf("%u: n_max                                  %u\n", port, cap.n_max);
  // printf("%u: n_shared_max                           %u\n", port, cap.n_shared_max);
  // printf("%u: meter_srtcm_rfc2697_n_max              %u\n", port, cap.meter_srtcm_rfc2697_n_max);
  // printf("%u: meter_trtcm_rfc2698_n_max              %u\n", port, cap.meter_trtcm_rfc2698_n_max);
  // printf("%u: meter_trtcm_rfc4115_n_max              %u\n", port, cap.meter_trtcm_rfc4115_n_max);
  // printf("%u: shared_n_flows_per_mtr_max             %u\n", port, cap.shared_n_flows_per_mtr_max);
  // printf("%u: chaining_n_mtrs_per_flow_max           %u\n", port, cap.chaining_n_mtrs_per_flow_max);
  // printf("%u: meter_rate_max                         %llu\n", port, (long long unsigned int)cap.meter_rate_max);
  // printf("%u: meter_policy_n_max                     %llu\n", port, (long long unsigned int)cap.meter_policy_n_max);
  // printf("%u: stats_mask                             %llu\n", port, (long long unsigned int)cap.stats_mask);
  // printf("%u: input_color_proto_mask                 %llu\n", port, (long long unsigned int)cap.input_color_proto_mask);
  // printf("\n");

  struct rte_mtr_meter_profile profile;
  profile.alg = RTE_MTR_TRTCM_RFC2698;
  profile.trtcm_rfc2698.cir = mtr_cir;
  profile.trtcm_rfc2698.pir = mtr_cir;
  profile.trtcm_rfc2698.cbs = mtr_cir/10;
  profile.trtcm_rfc2698.pbs = mtr_cir/10;
  profile.packet_mode = 0;
  if (rte_mtr_meter_profile_add(port, meter_profile_id[port], &profile, &error)) {
    printf("Error: Cannot add meter profile: %s\n", error.message);
    return -1;
  }

  struct rte_mtr_meter_policy_params policy;
  struct rte_flow_action actions_green[2];
  struct rte_flow_action actions_yellow[2];
  struct rte_flow_action actions_red[2];
  actions_green[0].type = RTE_FLOW_ACTION_TYPE_VOID;
  actions_green[1].type = RTE_FLOW_ACTION_TYPE_END;
  actions_yellow[0].type = RTE_FLOW_ACTION_TYPE_DROP;
  actions_yellow[1].type = RTE_FLOW_ACTION_TYPE_END;
  actions_red[0].type = RTE_FLOW_ACTION_TYPE_DROP;
  actions_red[1].type = RTE_FLOW_ACTION_TYPE_END;
  policy.actions[RTE_COLOR_GREEN] = actions_green;
  policy.actions[RTE_COLOR_YELLOW] = actions_yellow;
  policy.actions[RTE_COLOR_RED] = actions_red;
  if (rte_mtr_meter_policy_add(port, meter_policy_id[port], &policy, &error)) {
    printf("Error: Cannot add meter policy : %s\n", error.message);
    return -1;
  }

  struct rte_mtr_params params;
  memset(&params, 0, sizeof(struct rte_mtr_params));
  params.meter_profile_id = meter_profile_id[port];
  params.meter_enable = 1;
  params.stats_mask = RTE_MTR_STATS_N_PKTS_GREEN |
                      RTE_MTR_STATS_N_BYTES_GREEN;
  params.meter_policy_id = meter_policy_id[port];
  if (rte_mtr_create(port, meter_mtr_id[port], &params, 1, &error)) {
    printf("Error: Cannot create meter : %s\n", error.message);
    return -1;
  }
  return 0;
}

static int delete_rte_meter(uint8_t port)
{
  struct rte_mtr_error error;

  if (rte_mtr_destroy(port, meter_mtr_id[port], &error) != 0) {
    printf("Error: Cannot delete meter : %s\n", error.message);
    return -1;
  }

  if (rte_mtr_meter_policy_delete(port, meter_policy_id[port], &error) != 0) {
    printf("Error: Cannot delete meter policy : %s\n", error.message);
    return -1;
  }

  if (rte_mtr_meter_profile_delete(port, meter_profile_id[port], &error) != 0) {
    printf("Error: Cannot delete meter profile : %s\n", error.message);
    return -1;
  }
  return 0;
}

static const char short_options[] =
  "h"   /* Help                  */
  "i:"  /* ingress port          */
  "e:"  /* egress port           */
  "I:"  /* ingress port VLAN     */
  "E:"  /* egress port VLAN      */
  "c"   /* Decap/encap packers   */
  "f"   /* Set filter            */
  "q:"  /* Number of queues      */
  "n"   /* No HW offload         */
  "d"   /* debug print           */
  "s"   /* extended stats        */
  "m"   /* meter stats           */
  "l:"  /* limit rate            */
  ;

/* display usage */
static void forward_usage(const char *prgname)
{
  printf("\n%s [-h][-i port no][-e port no][-s][-n][-d][-f]\n"
         "  -h         : Display help\n"
         "  -i port no : Ingress port number. Default port no. is port 0\n"
         "  -e port no : Eegress port number. Default port no. is port 1\n"
         "  -I VLAN ID : Ingress port VLAN. Default VLAN is port 44\n"
         "  -E VLAN ID : Eegress port VLAN. Default VLAN is port 45\n"
         "  -c         : Decap and encap packets\n"
         "  -f         : Do not set filters (VLAN or MAC)\n"
         "  -n         : Disable HW offload\n"
         "  -s         : Extended statistics\n"
         "  -m         : Meter flow statistics\n"
         "  -l         : Data rate limits i 1G steps (requires -m)\n"
         "  -d         : Debug print out\n",
         prgname);
}

static unsigned int forward_parse_value(const char *q_arg)
{
  char *end = NULL;
  unsigned long n;

  /* parse hexadecimal string */
  n = strtoul(q_arg, &end, 10);
  if ((q_arg[0] == '\0') || (end == NULL) || (*end != '\0'))
    return 0;

  return n;
}

/* Parse the argument given in the command line of the application */
static int forward_parse_args(int argc, char **argv)
{
  int opt, ret;
  char **argvopt;
  char *prgname = argv[0];

  argvopt = argv;

  while ((opt = getopt(argc, argvopt, short_options)) != EOF) {
    switch (opt) {
    case 'i':
      ingress_port = forward_parse_value(optarg);
      break;

    case 'e':
      eegress_port = forward_parse_value(optarg);
      break;

    case 'I':
      ingress_vlan = forward_parse_value(optarg);
      break;

    case 'E':
      eegress_vlan = forward_parse_value(optarg);
      break;

    case 'l':
      {
        uint32_t val = forward_parse_value(optarg);
        mtr_cir = ((uint64_t)val * 1000ULL*1000ULL*1000ULL) / 8ULL;
      }
      break;

    case 'q':
      number_of_queues = forward_parse_value(optarg);
      if (number_of_queues > MAX_QUEUES) {
        rte_exit(EXIT_FAILURE, "To many queues requested\n");
      }
      break;

    case 's':
      use_x_stat = true;
      break;

    case 'm':
      use_meter = true;
      break;

    case 'c':
      use_decap = true;
      break;

    case 'f':
      set_filter = false;
      break;

    case 'n':
      disable_hw_offload = true;
      break;

    case 'd':
      debug_print_out = true;
      break;

    case 'h':
      forward_usage(prgname);
      exit(0);

    default:
      forward_usage(prgname);
      return -1;
    }
  }

  if (optind >= 0)
    argv[optind-1] = prgname;

  ret = optind-1;
  optind = 1; /* reset getopt lib */
  return ret;
}

static void unlearn(void)
{
  uint64_t i;
  uint32_t q;
  struct rte_flow_error error;

  for (q = 0; q < number_of_queues; q++) {
    printf("Port %u, queue %u: Destroying %llu flm filters\n", ingress_port, q, (long long unsigned int)flows_stat->f[ingress_port][q].off_flow_count);
    for (i = 0; i < flows_stat->f[ingress_port][q].off_flow_count; i++) {
      if (flows_stat->f[ingress_port][q].off_flows[i]) {
        rte_flow_destroy(ingress_port, flows_stat->f[ingress_port][q].off_flows[i], &error);
        flows_stat->f[ingress_port][q].off_flows[i] = NULL;
      }
    }
    flows_stat->f[ingress_port][q].off_flow_count = 0;
  }

  for (q = 0; q < number_of_queues; q++) {
    printf("Port %u, queue %u: Destroying %llu flm filters\n", eegress_port, q, (long long unsigned int)flows_stat->f[eegress_port][q].off_flow_count);
    for (i = 0; i < flows_stat->f[eegress_port][q].off_flow_count; i++) {
      if (flows_stat->f[eegress_port][q].off_flows[i]) {
        rte_flow_destroy(eegress_port, flows_stat->f[eegress_port][q].off_flows[i], &error);
        flows_stat->f[eegress_port][q].off_flows[i] = NULL;
      }
    }
    flows_stat->f[eegress_port][q].off_flow_count = 0;
  }
}

int main(int argc, char *argv[])
{
  struct rte_mempool *mbuf_pool = NULL;
  unsigned int nb_ports;
  unsigned int lcore_id;
  int ret;
  uint queue;

  rte_eal_init_done = 0;
  quit_signal = 0;

  /* catch ctrl-c so we can print on exit */
  signal(SIGINT, int_handler);
  signal(SIGTERM, int_handler);
  signal(SIGUSR1, int_handler);
  signal(SIGUSR2, int_handler);
  signal(40, int_handler);
  signal(41, int_handler);

  printf("Forward DPDK app - Forwarding packets between two ports.\n");
  printf("-------------------------------------------------------.\n");

  pthread_mutex_init(&mutex, NULL);

  /* Initialize the Environment Abstraction Layer (EAL). */
  ret = rte_eal_init(argc, argv);
  if (ret < 0) {
      rte_exit(EXIT_FAILURE, "Error with EAL initialization\n");
  }

  rte_eal_init_done = 1;

  argc -= ret;
  argv += ret;

  /* parse application arguments (after the EAL ones) */
  ret = forward_parse_args(argc, argv);
  if (ret < 0) {
    rte_exit(EXIT_FAILURE, "Invalid forward arguments\n");
  }

  if (rte_lcore_count() < ((NO_PORTS * number_of_queues) + 2)) {
    rte_exit(EXIT_FAILURE, "Need at least %u cores specified", ((NO_PORTS * number_of_queues) + 2));
    return -1;
  }

  /* Check that there is an even number of ports to send/receive on. */
  nb_ports = rte_eth_dev_count_avail();
  if (nb_ports < NO_PORTS) {
      rte_exit(EXIT_FAILURE, "Needs at least 2 ports\n");
  }

  if (nb_ports > MAX_NO_PORTS) {
    rte_exit(EXIT_FAILURE, "To many ports. Supported %u - Found %u\n", MAX_NO_PORTS, nb_ports);
  }

  if (ingress_port > nb_ports) {
    rte_exit(EXIT_FAILURE, "Invalid ingress port\n");
  }

  if (eegress_port > nb_ports) {
    rte_exit(EXIT_FAILURE, "Invalid eegress port\n");
  }

  flows_stat = rte_zmalloc("Flow", sizeof(struct flows_stat_s), 0);
  if (!flows_stat) {
    rte_exit(EXIT_FAILURE, "Failed to allocate flow stat memory\n");
  }
  memset(&statistic, 0, sizeof(statistic));

  if (disable_hw_offload) {
    use_decap = false;
  }
  printf("Ingress port:     %u\n", ingress_port);
  printf("Eegress port:     %u\n", eegress_port);
  printf("Set filter:       %s\n", set_filter ? "yes" : "no");
  printf("Number of queues: %u\n", number_of_queues);
  printf("HW offload:       %s\n", disable_hw_offload ? "no" : "yes");
  printf("Decap packets:    %s\n", use_decap ? "yes" : "no");
  printf("Debug print:      %s\n\n", debug_print_out ? "enabled" : "disabled");
  printf("Number of ports found:      %u\n", nb_ports);

  /* Creates a new mempool in memory to hold the mbufs. */
  mbuf_pool = rte_pktmbuf_pool_create("MBUF_POOL", NUM_MBUFS, MBUF_CACHE_SIZE, 0, MAX_PKT_SIZE + RTE_PKTMBUF_HEADROOM, rte_socket_id());

  if (mbuf_pool == NULL)
      rte_exit(EXIT_FAILURE, "Cannot create mbuf pool\n");

  /* Initialize selected port. */
  if (port_init(ingress_port, mbuf_pool) != 0) {
    if (flows_stat) rte_free(flows_stat);
    rte_exit(EXIT_FAILURE, "Cannot init port 0\n");
  }

  if (port_init(eegress_port, mbuf_pool) != 0) {
    if (flows_stat) rte_free(flows_stat);
    rte_exit(EXIT_FAILURE, "Cannot init port 1\n");
  }

  set_vlan(ingress_port, ingress_vlan);
  set_vlan(eegress_port, eegress_vlan);

  lcore_id = rte_get_next_lcore(-1, 0, 0);

  if (set_filter) {
    if (setup_default_flows() != 0) {
      if (flows_stat) rte_free(flows_stat);
      rte_exit(EXIT_FAILURE, "Cannot setup filter\n");
    }
  }

  if (use_meter) {
    setup_meter(ingress_port);
    setup_meter(eegress_port);
  }

  struct port_t rx0[MAX_QUEUES];
  struct port_t rx1[MAX_QUEUES];

  for (queue = 0; queue < number_of_queues; queue++) {
    lcore_id = rte_get_next_lcore(lcore_id, 0, 0);
    rx0[queue].rxPort = ingress_port;
    rx0[queue].txPort = eegress_port;
    rx0[queue].queue = queue;
    rx0[queue].stat = &statistic[ingress_port][queue];
    rx0[queue].flows = &flows_stat->f[ingress_port][queue];
    rte_eal_remote_launch((lcore_function_t *)lcore_rx, &rx0[queue], lcore_id);
  }

  for (queue = 0; queue < number_of_queues; queue++) {
    lcore_id = rte_get_next_lcore(lcore_id, 0, 0);
    rx1[queue].rxPort = eegress_port;
    rx1[queue].txPort = ingress_port;
    rx1[queue].queue = queue;
    rx1[queue].stat = &statistic[eegress_port][queue];
    rx1[queue].flows = &flows_stat->f[eegress_port][queue];
    rte_eal_remote_launch((lcore_function_t *)lcore_rx, &rx1[queue], lcore_id);
  }

  lcore_id = rte_get_next_lcore(lcore_id, 0, 0);
  rte_eal_remote_launch((lcore_function_t *)learn_unlearn, NULL, lcore_id);

  if (use_x_stat)
    lcore_main2();
  else
    lcore_main1();

  printf("\nClosing and cleaning\n");

  RTE_LCORE_FOREACH_WORKER(lcore_id) {
    printf("Trying to close lcore %u\n", lcore_id);
    if (rte_eal_wait_lcore(lcore_id) < 0) {
      if (flows_stat) rte_free(flows_stat);
      return -1;
    }
  }

  unlearn();

  if (use_meter) {
    delete_rte_meter(ingress_port);
    delete_rte_meter(eegress_port);
  }

  if (set_filter) {
    delete_default_flows();
  }

  rte_eth_dev_stop(ingress_port);
  rte_eth_dev_stop(eegress_port);

  rte_eth_dev_close(ingress_port);
  rte_eth_dev_close(eegress_port);

  if (flows_stat) rte_free(flows_stat);

  return 0;
}
