from trex_stl_lib.api import *
from scapy.layers.inet import Ether, IP, UDP
from scapy.contrib.gtp import GTP_U_Header, Raw

"""
Test Name: trex_inline_upf_demo_zerodrop
Test Overview: Display TREX TG operation.
Test Procedure:
1. Connect to the server.
2. Set-up the following instructions:
   - UL_flows_LEARN for learning;
   - DL_flows for transmitting;
   - UL_flows for transmitting.
3. Set the appropriate values for LEARN, UL and DL streams.
4. Transmit traffic by TG.
5. Verify that 2 000 000 flows are learned on 1st phase.
6. Verify that the drop events are not on the 1st and 2nd phase.
"""

trex_client = STLClient(server='127.0.0.1')
trex_client.connect()

GTP_ext_h = [
              0x01,  # Extension Header Length:
              0x10,  # PDU Type: UL PDU SESSION INFORMATION (1)
              0x06,  # QoS Flow Identifier (QFI): 6
              0x00,  # Next extension header type
            ]

'''
gtp_type it is GTP message type = 0xFF(255) T-PDU
next_ex it is next extension header type = 0x85(133) - PDU session container
PT it is GTP Protocol type = GTP 1
E it is flag (is next extension header present) = 1 - ext header presents
'''
ENCAP_H = Ether(src='16:16:16:16:16:16', dst='1a:1a:1a:1a:1a:1a') \
        / IP(src='20.0.0.0', dst='21.0.0.0') \
        / UDP(dport=2152) \
        / GTP_U_Header(PT=1, E=1, gtp_type=0xFF, teid=0xcafe101, next_ex=0x85) \
        / Raw(GTP_ext_h)

CLIENT_FLOW = IP(src='20.0.0.0', dst='21.0.0.0') \
            / UDP(sport=4096, dport=4097) \
            / Raw('NapaTech inline UPF demo test')

uplink_pkt = ENCAP_H / CLIENT_FLOW

# create a DL base packet
downlink_pkt = Ether(src='16:16:16:16:16:16', dst='1a:1a:1a:1a:1a:1a') \
             / CLIENT_FLOW

# packet size in bytes
PACKET_SIZE = 512
N3_PORT = 0
N6_PORT = 1
# ethernet header size
IPV4_H_LEN = 20
ETH_H_LEN = 14


def pad(base_pkt):
    padding = (PACKET_SIZE - len(base_pkt)) * '.'
    return padding


def create_stream(flows, packet, mode):
    vm = STLScVmRaw(flows)
    pkt_builder = STLPktBuilder(packet/pad(packet), vm=vm)
    stream = STLStream(packet=pkt_builder, mode=mode)
    return stream


def run_traffic_stream_learning(trex_client, ul_stream_LEARN, N3_PORT):
    trex_client.reset(ports=[N3_PORT, N6_PORT])
    trex_client.add_streams(ul_stream_LEARN, ports=[N3_PORT])
    trex_client.start(ports=[N3_PORT])
    trex_client.wait_on_traffic()


def run_traffic_stream_transmitting(trex_client, ul_stream, dl_stream, N3_PORT, N6_PORT):
    trex_client.reset(ports=[N3_PORT, N6_PORT])
    trex_client.add_streams(ul_stream, ports=[N3_PORT])
    trex_client.add_streams(dl_stream, ports=[N6_PORT])
    trex_client.start(ports=[N3_PORT, N6_PORT])
    trex_client.wait_on_traffic()


try:
    # Inctruction to generate range of L3 flows for N3 UPF interafce
    UL_flows = [
        STLVmFlowVar('ip_dst',
                     min_value='21.0.0.0',
                     max_value='21.14.255.255',
                     size=4,
                     step=1,
                     op='inc'
                     ),
        # write ip to packet IP.dst
        STLVmWrFlowVar(fv_name='ip_dst',
                       pkt_offset='IP.dst'),
        STLVmFlowVar('inner_ip_dst',
                     min_value='21.0.0.0',
                     max_value='21.14.255.255',
                     size=4,
                     step=1,
                     op='inc'
                     ),
        # write ip to packet inner IP.dst
        # [16] - offset to inner IPv4 address in the IPv4 header
        STLVmWrFlowVar(fv_name='inner_ip_dst',
                       pkt_offset=len(ENCAP_H) + 16),

        STLVmFixIpv4(offset='IP'),
        STLVmFixChecksumHw(l3_offset=len(ENCAP_H),
                           l4_offset=len(ENCAP_H) + IPV4_H_LEN,
                           l4_type=CTRexVmInsFixHwCs.L4_TYPE_UDP)
    ]

    DL_flows = [
        STLVmFlowVar('ip_dst',
                     min_value='21.0.0.0',
                     max_value='21.14.255.255',
                     size=4,
                     step=1,
                     op='inc'
                     ),
        # write ip to packet IP.dst
        STLVmWrFlowVar(fv_name='ip_dst',
                       pkt_offset='IP.dst'),

        STLVmFixIpv4(offset='IP'),
        STLVmFixChecksumHw(l3_offset=ETH_H_LEN,
                           l4_offset=ETH_H_LEN + IPV4_H_LEN,
                           l4_type=CTRexVmInsFixHwCs.L4_TYPE_UDP)
    ]

    UL_flows_LEARN = [
        STLVmFlowVar('ip_dst',
                     min_value='21.0.0.0',
                     max_value='21.255.255.255',
                     size=4,
                     step=1,
                     op='inc'
                     ),
        # write ip to packet IP.dst
        STLVmWrFlowVar(fv_name='ip_dst',
                       pkt_offset='IP.dst'),
        STLVmFlowVar('inner_ip_dst',
                     min_value='21.0.0.0',
                     max_value='21.255.255.255',
                     size=4,
                     step=1,
                     op='inc'
                     ),
        # write ip to packet inner IP.dst
        # [16] - offset to inner IPv4 address in the IPv4 header
        STLVmWrFlowVar(fv_name='inner_ip_dst',
                       pkt_offset=len(ENCAP_H) + 16),

        STLVmFixIpv4(offset='IP'),
        STLVmFixChecksumHw(l3_offset=len(ENCAP_H),
                           l4_offset=len(ENCAP_H) + IPV4_H_LEN,
                           l4_type=CTRexVmInsFixHwCs.L4_TYPE_UDP)
    ]

    ul_stream_LEARN = create_stream(UL_flows_LEARN, uplink_pkt, STLTXSingleBurst(
        pps=70000, total_pkts=1000000))

    ul_stream = create_stream(UL_flows, uplink_pkt, STLTXMultiBurst(
        pps=20000000, pkts_per_burst=100000000, count=8, ibg=0))

    dl_stream = create_stream(DL_flows, downlink_pkt, STLTXMultiBurst(
        pps=20000000, pkts_per_burst=100000000, count=8, ibg=0))

    print('\nTest execution 1st phase(learning) in progress. Press CTRL-C to stop...\n')
    run_traffic_stream_learning(trex_client, ul_stream_LEARN, N3_PORT)
    print('Learn Finished\n')

    print('Test execution 2nd phase(transmitting) in progress. Press CTRL-C to stop...\n')
    run_traffic_stream_transmitting(trex_client, ul_stream, dl_stream, N3_PORT, N6_PORT)
    print('Transmit Finished\n')

except STLError as e:
    print(e)

finally:
    trex_client.disconnect()
