from trex_stl_lib.api import *
from scapy.layers.inet import Ether, IP, UDP
from scapy.contrib.gtp import GTP_U_Header, Raw

"""
Test Name: trex_inline_upf_demo_zerodrop_NAT
Test Overview: Display TREX TG operation using NAT.
Test Procedure:
1. Connect to the server.
2. Set-up the following instructions:
   - UL_flows_LEARN for learning;
   - UL_flows for transmitting.
3. Set the appropriate values for LEARN and UL streams.
4. Transmit traffic by TG.
5. Verify that 2 000 000 flows are learned on 1st phase.
6. Verify that the drop events are not on the 1st and 2nd phase.
"""

trex_client = STLClient(server='127.0.0.1')
trex_client.connect()

GTP_ext_h = [
    0x01,  # Extension Header Length:
    0x10,  # PDU Type: UL PDU SESSION INFORMATION (1)
    0x06,  # QoS Flow Identifier (QFI): 6
    0x00,  # Next extension header type
]

N3_IF_PEER_S = '10.10.10.2'
N3_IF_ADDR = '10.10.10.3'
UE_ADDR_S = '192.15.0.0'
SERVER_UE = '1.1.1.2'

N6_IF_ADDR = '20.20.0.0'  # RSS queue id 0
# N6_IF_ADDR = '20.21.0.0'  # RSS queue id 1
# N6_IF_ADDR = '20.22.0.0'  # RSS queue id 2
# N6_IF_ADDR = '20.23.0.0'  # RSS queue id 3

# packet size in bytes
PACKET_SIZE = 512
N3_PORT = 0

'''
gtp_type it is GTP message type = 0xFF(255) T-PDU
next_ex it is next extension header type = 0x85(133) - PDU session container
PT it is GTP Protocol type = GTP 1
E it is flag (is next extension header present) = 1 - ext header presents
'''
GTP_U = GTP_U_Header(PT=1, E=1, gtp_type=0xFF, teid=0xcafe101, next_ex=0x85) \
      / Raw(GTP_ext_h)

OUTER_H = Ether(src='16:16:16:16:16:16', dst='1a:1a:1a:1a:1a:1a') \
        / IP(src=N3_IF_PEER_S, dst=N3_IF_ADDR) \
        / UDP(sport=2152, dport=2152, chksum=0) \
        / GTP_U

CLIENT_FLOW_UL = IP(src=UE_ADDR_S, dst=SERVER_UE) \
               / UDP(sport=4096, dport=4097, chksum=0) \
               / Raw('NapaTech inline UPF demo test')

uplink_pkt = OUTER_H / CLIENT_FLOW_UL


def pad(base_pkt):
    padding = (PACKET_SIZE - len(base_pkt)) * '.'
    return padding


def create_stream(flows, packet, mode):
    vm = STLScVmRaw(flows)
    pkt_builder = STLPktBuilder(packet/pad(packet), vm=vm)
    stream = STLStream(packet=pkt_builder, mode=mode)
    return stream


def run_traffic_stream(trex_client, stream, ports):
    trex_client.reset(ports=[N3_PORT])
    trex_client.add_streams(stream, ports=ports)
    trex_client.start(ports=ports)
    trex_client.wait_on_traffic()


try:
    # Inctruction to generate range of L3 flows for N3 UPF interafce
    def generate_flow_config(ip_src_min, ip_src_max, inner_ip_src_min, inner_ip_src_max):
        flow_config = [
            STLVmFlowVar('ip_src',
                         min_value=ip_src_min,
                         max_value=ip_src_max,
                         size=4,
                         step=1,
                         op='inc'
                         ),
            # write ip to packet IP.dst
            STLVmWrFlowVar(fv_name='ip_src',
                           pkt_offset='IP.src'),

            # # Due to the unpredictable packet sequencing in the TRex API,
            # # the utilization of TEID becomes unfeasible (TRex limitation)
            # STLVmFlowVar('teid',
            #              min_value=1,
            #              max_value=1000000,
            #              size=4,
            #              step=1,
            #              op='inc'
            #              ),
            # # write ip to packet inner gtp teid
            # # [-12] - offset to teid from end of outer header
            # STLVmWrFlowVar(fv_name='teid',
            #                pkt_offset=len(OUTER_H) - 12),

            STLVmFlowVar('inner_ip_src',
                         min_value=inner_ip_src_min,
                         max_value=inner_ip_src_max,
                         size=4,
                         step=1,
                         op='inc'
                         ),
            # write ip to packet inner IP.src
            # [12] - offset to inner IPv4 address in the IPv4 header
            STLVmWrFlowVar(fv_name='inner_ip_src',
                           pkt_offset=len(OUTER_H) + 12),

            STLVmFixIpv4(offset="IP"),
            STLVmFixIpv4(offset=len(OUTER_H))
        ]
        return flow_config

    UL_flows_LEARN = generate_flow_config('10.0.0.0',        # ip_src_min
                                          '10.15.255.255',   # ip_src_max
                                          '192.0.0.0',       # inner_ip_src_min
                                          '192.15.255.255')  # inner_ip_src_max

    UL_flows = generate_flow_config('10.0.0.0',              # ip_src_min
                                    '10.14.255.255',         # ip_src_max
                                    '192.14.0.0',            # inner_ip_src_min
                                    '192.14.255.255')        # inner_ip_src_max

    ul_stream_LEARN = create_stream(UL_flows_LEARN, uplink_pkt, STLTXSingleBurst(
        pps=70000, total_pkts=1000000))

    ul_stream = create_stream(UL_flows, uplink_pkt, STLTXMultiBurst(
        pps=20000000, pkts_per_burst=100000000, count=11, ibg=0))

    print('\nTest execution 1st phase(learning) in progress. Press CTRL-C to stop...\n')
    run_traffic_stream(trex_client, ul_stream_LEARN, [N3_PORT])
    print('Learn Finished\n')

    print('Test execution 2nd phase(transmitting) in progress. Press CTRL-C to stop...\n')
    run_traffic_stream(trex_client, ul_stream, [N3_PORT])
    print('Transmit Finished\n')

except STLError as e:
    print(e)

finally:
    trex_client.disconnect()
