//
// utils.c
//

/*
 * %NT_SOFTWARE_LICENSE%
 */
#include <assert.h>
#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

#include "utils.h"

/*
 * UNIX Socket handling helper routines
 */

static int unix_build_address(const char *path, struct sockaddr_un *addr)
{
  if (addr == NULL || path == NULL) {
    return -1;
  }
  memset(addr, 0, sizeof(struct sockaddr_un));
  addr->sun_family = AF_UNIX;
  if (strlen(path) < sizeof(addr->sun_path)) {
    strncpy(addr->sun_path, path, sizeof(addr->sun_path) - 1);
    return 0;
  }
  return -1;
}

int unix_connect(const char *path)
{
  struct sockaddr_un addr;

  if (unix_build_address(path, &addr) == -1) {
    return -1;
  }

  int sd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (sd == -1) {
    return -1;
  }

  if (connect(sd, (struct sockaddr *)&addr, sizeof(struct sockaddr_un)) == -1) {
    close(sd);
    return -1;
  }
  return sd;
}

int read_data(int fd, size_t len, uint8_t *data, size_t *recv_len, int timeout)
{
  struct pollfd pfd;
  ssize_t ret = 0;

  pfd.fd = fd;
  pfd.events = POLLIN;
  pfd.revents = 0;

  ret = poll(&pfd, 1, timeout);
  if (ret < 0) {
    if (errno == EINTR) {
      return TRYAGAIN; // caught signal before timeout
    }
    if (errno == EINVAL) {
      return INVALID_PARAMETER; // timeout is negative
    }
    if (errno == EFAULT) {
      return INVALID_PARAMETER; // fds argument is illegal
    }
    // else
    assert(0);
    return -1;
  }
  if (ret == 0) {
    return TIMEOUT;
  }

  if (pfd.revents == 0) {
    assert(ret == 1);
    assert(0); // revents cannot be zero when NtSocket_Poll returns 1
    return TRYAGAIN;
  }

  if ((pfd.revents & POLLIN) && ((pfd.revents & (POLLERR | POLLNVAL)) == 0)) {
    ret = recv(pfd.fd, data, len, 0);
    if (ret < 0) {
      int lerrno = errno;
      if ((lerrno == EWOULDBLOCK) || (lerrno == EAGAIN)) {
        // we have data but if the very first read turns out to return EWOULDBLOCK or EAGAIN it means that the remote end has dropped the connection
        fprintf(stderr, "The socket with fd %d has been closed by remote end. %d [%s]\n", pfd.fd, lerrno, strerror(lerrno));
        return CONNECTION_CLOSED;
      }
      if (lerrno != EINTR) {
        fprintf(stderr, "recv() from fd %d received errno %d [%s]\n", pfd.fd, lerrno, strerror(lerrno));
        return CONNECTION_INVALID;
      }
      // EINTR
      return TRYAGAIN;
    }
    if (ret == 0) {
      if (pfd.revents & POLLHUP) {
        // this means that we have read all data and the remote end has HUP
        // fprintf(stderr, "The remote end has terminated the session\n");
        return CONNECTION_CLOSED;
      }
      return TRYAGAIN;
    }

    // ret can only be positive at this point
    *recv_len = (size_t)ret;
    return 0;
  }

  if ((pfd.revents & POLLHUP) == POLLHUP) {
    // this means that the remote end has HUP
    // fprintf(stderr, "The remote end has terminated the session");
    return CONNECTION_CLOSED;
  }

  fprintf(stderr, "poll() returned 0x%x. Invalidating the connection\n", pfd.revents);
  return CONNECTION_INVALID;
}

int read_all(int clfd, uint8_t *data, size_t length)
{
  size_t recv_len = 0;
  size_t left = length;
  size_t pos = 0;
  while (left > 0) {
    int ret = read_data(clfd, left, &data[pos], &recv_len, -1);
    if (ret == 0) {
      pos += recv_len;
      left -= recv_len;
    } else {
      if ((ret == CONNECTION_CLOSED) || (ret == TIMEOUT)) {
        // silently return status
        return ret;
      }
      if (ret != TRYAGAIN) {
        fprintf(stderr, "Failed getting all data. Error code: 0x%X\n", ret);
        return ret;
      }
    }
    // try again
  }
  return 0;
}

int write_all(int fd, const uint8_t *data, size_t size)
{
  size_t len = 0;
  while (len < size) {
    ssize_t res = write(fd, (const void *)&data[len], size - len);
    if (res < 0) {
      fprintf(stderr, "write to socket failed!\n");
      return -1;
    }
    len += res;
  }
  return 0;
}
